SystemOrganization addCategory: #'Traits-Composition'!
SystemOrganization addCategory: #'Traits-Kernel'!
SystemOrganization addCategory: #'Traits-Kernel-Traits'!
SystemOrganization addCategory: #'Traits-LocalSends'!
SystemOrganization addCategory: #'Traits-Requires'!
SystemOrganization addCategory: #'Traits-Tests'!

Trait named: #TApplyingOnClassSide
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TApplyingOnClassSide methodsFor: 'composition' stamp: 'al 7/30/2004 09:07'!
assertConsistantCompositionsForNew: aTraitComposition
	"Applying or modifying a trait composition on the class side
	of a behavior has some restrictions."

	| baseTraits notAddable message |
	baseTraits := aTraitComposition traits select: [:each | each isBaseTrait].
	baseTraits isEmpty ifFalse: [
		notAddable := (baseTraits reject: [:each | each classSide methodDict isEmpty]).
		notAddable isEmpty ifFalse: [
			message := String streamContents: [:stream |
				stream nextPutAll: 'You can not add the base trait(s)'; cr.
				notAddable
					do: [:each | stream nextPutAll: each name]
					separatedBy: [ stream nextPutAll: ', '].
				stream cr; nextPutAll: 'to this composition because it/they define(s) methods on the class side.'].
		^TraitCompositionException signal: message]].
		
	(self instanceSide traitComposition traits asSet =
			(aTraitComposition traits
				select: [:each | each isClassTrait]
				thenCollect: [:each | each baseTrait]) asSet) ifFalse: [
				^TraitCompositionException signal: 'You can not add or remove class side traits on
				the class side of a composition. (But you can specify aliases or exclusions
				for existing traits or add a trait which does not have any methods on the class side.)']! !

!TApplyingOnClassSide methodsFor: 'composition' stamp: 'apb 8/24/2005 14:15'!
noteNewBaseTraitCompositionApplied: aTraitComposition
	"The argument is the new trait composition of my base trait - add
	the new traits or remove non existing traits on my class side composition.
	(Each class trait in my composition has its base trait on the instance side
	of the composition - manually added traits to the class side are always 
	base traits.)"
	
	| newComposition traitsFromInstanceSide |
	traitsFromInstanceSide := self traitComposition traits
		select: [:each | each isClassTrait]
		thenCollect: [:each | each baseTrait].
		
	newComposition := self traitComposition copyTraitExpression.
	(traitsFromInstanceSide copyWithoutAll: aTraitComposition traits) do: [:each |
		newComposition removeFromComposition: each classTrait].
	(aTraitComposition traits copyWithoutAll: traitsFromInstanceSide) do: [:each |
		newComposition add:  (each classTrait)].

	self setTraitComposition: newComposition! !

Trait named: #TBehaviorCategorization
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TBehaviorCategorization methodsFor: 'organization' stamp: 'al 3/18/2006 13:46'!
category
	"Answer the system organization category for the receiver. First check whether the
	category name stored in the ivar is still correct and only if this fails look it up
	(latter is much more expensive)"

	| result |
	self basicCategory ifNotNilDo: [ :symbol |
		((SystemOrganization listAtCategoryNamed: symbol) includes: self name)
			ifTrue: [ ^symbol ] ].
	self basicCategory: (result := SystemOrganization categoryOfElement: self name).
	^result! !

!TBehaviorCategorization methodsFor: 'organization' stamp: 'al 3/18/2006 13:35'!
category: aString 
	"Categorize the receiver under the system category, aString, removing it from 
	any previous categorization."

	| oldCategory |
	oldCategory := self basicCategory.
	aString isString
		ifTrue: [
			self basicCategory: aString asSymbol.
			SystemOrganization classify: self name under: self basicCategory ]
		ifFalse: [self errorCategoryName].
	SystemChangeNotifier uniqueInstance
		class: self recategorizedFrom: oldCategory to: self basicCategory! !

Trait named: #TClassAndTraitDescription
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'sd 4/26/2008 17:28'!
acceptsLoggingOfCompilation
	"Answer whether the receiver's method submisions and class defintions should be logged to the changes file and to the current change set.  The metaclass follows the rule of the class itself.  6/18/96 sw"
	"weird name is so that it will come lexically before #compile, so that a clean build can make it through.  7/7/96 sw"

	^ true! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 5/8/2004 20:36'!
addAndClassifySelector: selector withMethod: compiledMethod inProtocol: category notifying: requestor
	| priorMethodOrNil |
	priorMethodOrNil := self compiledMethodAt: selector ifAbsent: [nil].
	self addSelectorSilently: selector withMethod: compiledMethod.
	SystemChangeNotifier uniqueInstance 
		doSilently: [self organization classify: selector under: category].
	priorMethodOrNil isNil
		ifTrue: [SystemChangeNotifier uniqueInstance methodAdded: compiledMethod selector: selector inProtocol: category class: self requestor: requestor]
		ifFalse: [SystemChangeNotifier uniqueInstance methodChangedFrom: priorMethodOrNil to: compiledMethod selector: selector inClass: self requestor: requestor].! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 5/8/2004 20:33'!
addSelector: selector withMethod: compiledMethod notifying: requestor
	| priorMethodOrNil |
	priorMethodOrNil := self compiledMethodAt: selector ifAbsent: [nil].
	self addSelectorSilently: selector withMethod: compiledMethod.
	priorMethodOrNil isNil
		ifTrue: [SystemChangeNotifier uniqueInstance methodAdded: compiledMethod selector: selector inClass: self requestor: requestor]
		ifFalse: [SystemChangeNotifier uniqueInstance methodChangedFrom: priorMethodOrNil to: compiledMethod selector: selector inClass: self requestor: requestor].! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 09:51'!
addSelectorSilently: selector withMethod: compiledMethod
	super addSelectorSilently: selector withMethod: compiledMethod.
	self instanceSide noteAddedSelector: selector meta: self isMeta.! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'al 5/9/2004 14:08'!
applyChangesOfNewTraitCompositionReplacing: oldComposition
	| changedSelectors |
	changedSelectors := super applyChangesOfNewTraitCompositionReplacing: oldComposition.
	self noteRecategorizedSelectors: changedSelectors oldComposition: oldComposition.
	^ changedSelectors.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 17:13'!
classComment: aString
	"Store the comment, aString or Text or RemoteString, associated with the class we are orgainzing.  Empty string gets stored only if had a non-empty one before."
	^ self classComment: aString stamp: '<historical>'! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 12/1/2005 15:43'!
classComment: aString stamp: aStamp
	"Store the comment, aString or Text or RemoteString, associated with the class we are organizing.  Empty string gets stored only if had a non-empty one before."

	| ptr header file oldCommentRemoteStr |
	(aString isKindOf: RemoteString) ifTrue:
		[SystemChangeNotifier uniqueInstance classCommented: self.
		^ self organization classComment: aString stamp: aStamp].

	oldCommentRemoteStr := self organization commentRemoteStr.
	(aString size = 0) & (oldCommentRemoteStr == nil) ifTrue: [^ self organization classComment: nil].
		"never had a class comment, no need to write empty string out"

	ptr := oldCommentRemoteStr ifNil: [0] ifNotNil: [oldCommentRemoteStr sourcePointer].
	SourceFiles ifNotNil: [(file := SourceFiles at: 2) ifNotNil:
		[file setToEnd; cr; nextPut: $!!.	"directly"
		"Should be saying (file command: 'H3') for HTML, but ignoring it here"
		header := String streamContents: [:strm | strm nextPutAll: self name;
			nextPutAll: ' commentStamp: '.
			aStamp storeOn: strm.
			strm nextPutAll: ' prior: '; nextPutAll: ptr printString].
		file nextChunkPut: header]].
	self organization classComment: (RemoteString newString: aString onFileNumber: 2) stamp: aStamp.
	SystemChangeNotifier uniqueInstance classCommented: self.
! !

!TClassAndTraitDescription methodsFor: 'accessing comment' stamp: 'dvf 9/27/2005 16:55'!
comment
	"Answer the receiver's comment. (If missing, supply a template) "
	| aString |
	aString := self instanceSide organization classComment.
	aString isEmpty ifFalse: [^ aString].
	^self classCommentBlank! !

!TClassAndTraitDescription methodsFor: 'accessing comment' stamp: 'al 5/9/2004 17:05'!
comment: aStringOrText
	"Set the receiver's comment to be the argument, aStringOrText."

	self instanceSide classComment: aStringOrText.! !

!TClassAndTraitDescription methodsFor: 'accessing comment' stamp: 'al 5/9/2004 17:07'!
comment: aStringOrText stamp: aStamp
	"Set the receiver's comment to be the argument, aStringOrText."

	self instanceSide classComment: aStringOrText stamp: aStamp.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 17:14'!
commentFollows 
	"Answer a ClassCommentReader who will scan in the comment."

	^ ClassCommentReader new setClass: self category: #Comment

	"False commentFollows inspect"! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 11/28/2005 22:18'!
commentStamp: changeStamp
	self organization commentStamp: changeStamp.
	^ self commentStamp: changeStamp prior: 0! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 17:15'!
commentStamp: changeStamp prior: indexAndOffset
	"Prior source link ignored when filing in."

	^ ClassCommentReader new setClass: self
				category: #Comment
				changeStamp: changeStamp! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 11/28/2005 22:18'!
compile: code classified: heading 
	"Compile the argument, code, as source code in the context of the 
	receiver and install the result in the receiver's method dictionary under 
	the classification indicated by the second argument, heading. nil is to be 
	notified if an error occurs. The argument code is either a string or an 
	object that converts to a string or a PositionableStream on an object that 
	converts to a string."

	^self
		compile: code
		classified: heading
		notifying: nil! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:00'!
compile: text classified: category notifying: requestor
	| stamp |
	stamp := self acceptsLoggingOfCompilation ifTrue: [Utilities changeStamp] ifFalse: [nil].
	^ self compile: text classified: category
		withStamp: stamp notifying: requestor! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:00'!
compile: text classified: category withStamp: changeStamp notifying: requestor
	^ self compile: text classified: category withStamp: changeStamp notifying: requestor logSource: self acceptsLoggingOfCompilation! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 11/28/2005 22:21'!
compile: text classified: category withStamp: changeStamp notifying: requestor logSource: logSource
	| methodAndNode |
	methodAndNode := self compile: text asString classified: category notifying: requestor
							trailer: self defaultMethodTrailer ifFail: [^nil].
	logSource ifTrue: [
		self logMethodSource: text forMethodWithNode: methodAndNode 
			inCategory: category withStamp: changeStamp notifying: requestor.
	].
	self addAndClassifySelector: methodAndNode selector withMethod: methodAndNode 
		method inProtocol: category notifying: requestor.
	self instanceSide noteCompilationOf: methodAndNode selector meta: self isClassSide.
	^ methodAndNode selector! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:12'!
compile: code notifying: requestor 
	"Refer to the comment in Behavior|compile:notifying:." 

	^self compile: code
		 classified: ClassOrganizer default
		 notifying: requestor! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:12'!
compileSilently: code classified: category
	"Compile the code and classify the resulting method in the given category, leaving no trail in the system log, nor in any change set, nor in the 'recent submissions' list. This should only be used when you know for sure that the compilation will succeed."

	^ self compileSilently: code classified: category notifying: nil.! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:13'!
compileSilently: code classified: category notifying: requestor
	"Compile the code and classify the resulting method in the given category, leaving no trail in the system log, nor in any change set, nor in the 'recent submissions' list. This should only be used when you know for sure that the compilation will succeed."

	^ SystemChangeNotifier uniqueInstance 
		doSilently: [self compile: code classified: category withStamp: nil notifying: requestor logSource: false].! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:53'!
copy: sel from: class 
	"Install the method associated with the first argument, sel, a message 
	selector, found in the method dictionary of the second argument, class, 
	as one of the receiver's methods. Classify the message under -As yet not 
	classified-."

	self copy: sel
		from: class
		classified: nil! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:54'!
copy: sel from: class classified: cat 
	"Install the method associated with the first arugment, sel, a message 
	selector, found in the method dictionary of the second argument, class, 
	as one of the receiver's methods. Classify the message under the third 
	argument, cat."

	| code category |
	"Useful when modifying an existing class"
	code := class sourceMethodAt: sel.
	code == nil
		ifFalse: 
			[cat == nil
				ifTrue: [category := class organization categoryOfElement: sel]
				ifFalse: [category := cat].
			(self methodDict includesKey: sel)
				ifTrue: [code asString = (self sourceMethodAt: sel) asString 
							ifFalse: [self error: self name 
										, ' ' 
										, sel 
										, ' will be redefined if you proceed.']].
			self compile: code classified: category]! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:54'!
copyAll: selArray from: class 
	"Install all the methods found in the method dictionary of the second 
	argument, class, as the receiver's methods. Classify the messages under 
	-As yet not classified-."

	self copyAll: selArray
		from: class
		classified: nil! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:54'!
copyAll: selArray from: class classified: cat 
	"Install all the methods found in the method dictionary of the second 
	argument, class, as the receiver's methods. Classify the messages under 
	the third argument, cat."

	selArray do: 
		[:s | self copy: s
				from: class
				classified: cat]! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:54'!
copyAllCategoriesFrom: aClass 
	"Specify that the categories of messages for the receiver include all of 
	those found in the class, aClass. Install each of the messages found in 
	these categories into the method dictionary of the receiver, classified 
	under the appropriate categories."

	aClass organization categories do: [:cat | self copyCategory: cat from: aClass]! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:54'!
copyCategory: cat from: class 
	"Specify that one of the categories of messages for the receiver is cat, as 
	found in the class, class. Copy each message found in this category."

	self copyCategory: cat
		from: class
		classified: cat! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:55'!
copyCategory: cat from: aClass classified: newCat 
	"Specify that one of the categories of messages for the receiver is the 
	third argument, newCat. Copy each message found in the category cat in 
	class aClass into this new category."

	self copyAll: (aClass organization listAtCategoryNamed: cat)
		from: aClass
		classified: newCat! !

!TClassAndTraitDescription methodsFor: 'copying' stamp: 'al 5/8/2004 18:55'!
copyMethodDictionaryFrom: donorClass
	"Copy the method dictionary of the donor class over to the receiver"

	self methodDict: donorClass copyOfMethodDictionary.
	self organization: donorClass organization deepCopy.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:34'!
definition
	"Answer a String that defines the receiver in good old ST-80."

	^ self definitionST80! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 6/29/2004 21:37'!
doneCompiling
	"A ClassBuilder has finished the compilation of the receiver.
	This message is a notification for a class that needs to do some
	cleanup / reinitialization after it has been recompiled."! !

!TClassAndTraitDescription methodsFor: 'private' stamp: 'al 5/8/2004 19:24'!
errorCategoryName
	self error: 'Category name must be a String'! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'sd 4/25/2008 15:37'!
fileOutCategory: catName 

	| internalStream |
	internalStream := WriteStream on: (String new: 1000).
	internalStream header; timeStamp.
	self fileOutCategory: catName on: internalStream moveSource: false toFile: 0.
	internalStream trailer.

	^ FileStream writeSourceCodeFrom: internalStream baseName: (self name , '-' , catName) isSt: true.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:51'!
fileOutCategory: aSymbol on: aFileStream moveSource: moveSource toFile: fileIndex 
	"File a description of the receiver's category, aString, onto aFileStream. If 
	moveSource, is true, then set the method source pointer to the new file position.
	Note when this method is called with moveSource=true, it is condensing the
	.sources file, and should only write one preamble per method category."

	| selectors |

	aFileStream cr.
	selectors := (aSymbol asString = ClassOrganizer allCategory)
				ifTrue: [ self organization allMethodSelectors ]
				ifFalse: [ self organization listAtCategoryNamed: aSymbol ].

	selectors := selectors select: [:each | self includesLocalSelector: each].
	
	"Overridden to preserve author stamps in sources file regardless"
	selectors do: [:sel |
		self printMethodChunk: sel 
			withPreamble: true
			on: aFileStream 
			moveSource: moveSource 
			toFile: fileIndex].
	^ self! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:51'!
fileOutChangedMessages: aSet on: aFileStream 
	"File a description of the messages of the receiver that have been 
	changed (i.e., are entered into the argument, aSet) onto aFileStream."

	self fileOutChangedMessages: aSet
		on: aFileStream
		moveSource: false
		toFile: 0! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:52'!
fileOutChangedMessages: aSet on: aFileStream moveSource: moveSource toFile: fileIndex 
	"File a description of the messages of this class that have been 
	changed (i.e., are entered into the argument, aSet) onto aFileStream.  If 
	moveSource, is true, then set the method source pointer to the new file position.
	Note when this method is called with moveSource=true, it is condensing the
	.changes file, and should only write a preamble for every method."
	| org sels |
	(org := self organization) categories do: 
		[:cat | 
		sels := (org listAtCategoryNamed: cat) select: [:sel | aSet includes: sel].
		sels do:
			[:sel |  self printMethodChunk: sel withPreamble: true on: aFileStream
							moveSource: moveSource toFile: fileIndex]]! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'sd 4/25/2008 15:36'!
fileOutMethod: selector
	"Write source code of a single method on a file.  Make up a name for the file."
	
	| internalStream |
	(selector == #Comment) ifTrue: [^ self inform: 'Sorry, cannot file out class comment in isolation.'].
	(self includesSelector: selector) ifFalse: [^ self error: 'Selector ', selector asString, ' not found'].
	internalStream := WriteStream on: (String new: 1000).
	internalStream header; timeStamp.
	self printMethodChunk: selector withPreamble: true
		on: internalStream moveSource: false toFile: 0.

	FileStream writeSourceCodeFrom: internalStream baseName: (self name , '-' , (selector copyReplaceAll: ':' with: '')) isSt: true.! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:52'!
fileOutOn: aFileStream 
	"File a description of the receiver on aFileStream."

	self fileOutOn: aFileStream
		moveSource: false
		toFile: 0! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'sd 4/25/2008 15:38'!
fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex
	"File a description of the receiver on aFileStream. If the boolean 
	argument, moveSource, is true, then set the trailing bytes to the position 
	of aFileStream and to fileIndex in order to indicate where to find the 
	source code."

	aFileStream nextChunkPut: self definition.

	self organization
		putCommentOnFile: aFileStream
		numbered: fileIndex
		moveSource: moveSource
		forClass: self.
	self organization categories do: 
		[:heading |
		self fileOutCategory: heading
			on: aFileStream
			moveSource: moveSource
			toFile: fileIndex]! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:53'!
fileOutOrganizationOn: aFileStream
	"File a description of the receiver's organization on aFileStream."

	aFileStream cr; nextPut: $!!.
	aFileStream nextChunkPut: self name, ' reorganize'; cr.
	aFileStream nextChunkPut: self organization printString; cr! !

!TClassAndTraitDescription methodsFor: 'accessing comment' stamp: 'al 5/9/2004 17:09'!
hasComment
	"return whether this class truly has a comment other than the default"
	| org |
	org := self instanceSide organization.
	^org classComment isEmptyOrNil not! !

!TClassAndTraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 5/9/2004 20:50'!
isClassSide
	^self == self classSide! !

!TClassAndTraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 5/9/2004 20:50'!
isInstanceSide
	^self isClassSide not! !

!TClassAndTraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 5/9/2004 20:55'!
isMeta
	^self isClassSide! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'marcus.denker 8/25/2008 12:04'!
localMethods
	"returns the methods of classes including the ones of the traits that the class uses" 
	 
	^ self methods select: [:each | self includesLocalSelector: each selector].! !

!TClassAndTraitDescription methodsFor: 'organization' stamp: 'dvf 8/19/2005 11:31'!
methodReferencesInCategory: aCategoryName
	^(self organization listAtCategoryNamed: aCategoryName)
		collect: [:ea | MethodReference new
						setClassSymbol: self theNonMetaClass name
						classIsMeta: self isMeta
						methodSymbol: ea
						stringVersion: '']
! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'marcus.denker 8/25/2008 12:03'!
methods
	"returns the methods of classes including the ones of the traits that the class uses" 
	 
	^ self methodDict values  ! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:53'!
methodsFor: categoryName 
	"Answer a ClassCategoryReader for compiling the messages in the category, categoryName, of the receiver."

	^ ClassCategoryReader new setClass: self category: categoryName asSymbol

	"(False methodsFor: 'logical operations') inspect"! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:54'!
methodsFor: aString priorSource: sourcePosition inFile: fileIndex
	"Prior source pointer ignored when filing in."
	^ self methodsFor: aString! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:54'!
methodsFor: categoryName stamp: changeStamp 
	^ self methodsFor: categoryName stamp: (Utilities fixStamp: changeStamp) prior: 0! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:54'!
methodsFor: categoryName stamp: changeStamp prior: indexAndOffset
	"Prior source link ignored when filing in."
	^ ClassCategoryReader new setClass: self
				category: categoryName asSymbol
				changeStamp: changeStamp

"Most importantly, return the new ClassCategoryReader, so a fileIn will let it seize control.  So method will be placed in the proper category.  See the transfer of control where ReadWriteStream fileIn calls scanFrom:"! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 7/9/2006 18:20'!
moveChangesTo: newFile 
	"Used in the process of condensing changes, this message requests that 
	the source code of all methods of the receiver that have been changed 
	should be moved to newFile."

	| changes |
	changes := self methodDict keys select: [:sel |
		(self compiledMethodAt: sel) fileIndex > 1 and: [
			(self includesLocalSelector: sel) or: [
				(self compiledMethodAt: sel) sendsToSuper]]].
	self
		fileOutChangedMessages: changes
		on: newFile
		moveSource: true
		toFile: 2! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 5/9/2004 16:36'!
noteAddedSelector: aSelector meta: isMeta
	"A hook allowing some classes to react to adding of certain selectors"! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'dvf 9/27/2005 17:13'!
noteCompilationOf: aSelector meta: isMeta
	"A hook allowing some classes to react to recompilation of certain selectors"! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'al 5/9/2004 14:00'!
noteRecategorizedSelector: aSymbol from: oldCategoryOrNil to: newCategoryOrNil
	| changedCategories |
	changedCategories := self updateOrganizationSelector: aSymbol oldCategory: oldCategoryOrNil newCategory: newCategoryOrNil.
	changedCategories do: [:each |
		(self organization isEmptyCategoryNamed: each) ifTrue: [self organization removeCategory: each]]! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'al 5/9/2004 14:01'!
noteRecategorizedSelectors: aCollection oldComposition: aTraitComposition
	| oldCategory newCategory |
	aCollection do: [:each | 
		oldCategory := self organization categoryOfElement: each.
		newCategory := (self traitComposition methodDescriptionForSelector: each) effectiveMethodCategory.
		self noteRecategorizedSelector: each from: oldCategory to: newCategory]! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'al 5/9/2004 14:02'!
notifyOfRecategorizedSelector: element from: oldCategory to: newCategory
	SystemChangeNotifier uniqueInstance selector: element recategorizedFrom: oldCategory to: newCategory inClass: self! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:54'!
printCategoryChunk: categoryName on: aFileStream
	^ self printCategoryChunk: categoryName withStamp: '' on: aFileStream! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:55'!
printCategoryChunk: category on: aFileStream priorMethod: priorMethod
	^ self printCategoryChunk: category on: aFileStream
		withStamp: Utilities changeStamp priorMethod: priorMethod! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'sd 4/25/2008 15:39'!
printCategoryChunk: category on: aFileStream withStamp: changeStamp priorMethod: priorMethod 
	"Print a method category preamble.  This must have a category name.
	It may have an author/date stamp, and it may have a prior source link.
	If it has a prior source link, it MUST have a stamp, even if it is empty."

"The current design is that changeStamps and prior source links are preserved in the changes file.  All fileOuts include changeStamps.  Condensing sources, however, eliminates all stamps (and links, natch)."

	aFileStream cr; nextPut: $!!.
	aFileStream nextChunkPut: (String streamContents:
		[:strm |
		strm nextPutAll: self name; nextPutAll: ' methodsFor: '; print: category asString.
		(changeStamp ~~ nil and:
			[changeStamp size > 0 or: [priorMethod ~~ nil]]) ifTrue:
			[strm nextPutAll: ' stamp: '; print: changeStamp].
		priorMethod ~~ nil ifTrue:
			[strm nextPutAll: ' prior: '; print: priorMethod sourcePointer]]).
	! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:55'!
printCategoryChunk: categoryName withStamp: changeStamp on: aFileStream
	^ self printCategoryChunk: categoryName on: aFileStream withStamp: changeStamp
		priorMethod: nil! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'md 2/22/2006 16:25'!
printMethodChunk: selector withPreamble: doPreamble on: outStream
		moveSource: moveSource toFile: fileIndex
	"Copy the source code for the method associated with selector onto the fileStream.  If moveSource true, then also set the source code pointer of the method."
	| preamble method oldPos newPos sourceFile endPos |
	doPreamble 
		ifTrue: [preamble := self name , ' methodsFor: ' ,
					(self organization categoryOfElement: selector) asString printString]
		ifFalse: [preamble := ''].
	method := self methodDict at: selector ifAbsent:
		[outStream nextPutAll: selector; cr.
		outStream tab; nextPutAll: '** ERROR!!  THIS SCRIPT IS MISSING ** ' translated; cr; cr.
		outStream nextPutAll: '  '.
		^ outStream].

	((method fileIndex = 0
		or: [(SourceFiles at: method fileIndex) == nil])
		or: [(oldPos := method filePosition) = 0])
		ifTrue:
		["The source code is not accessible.  We must decompile..."
		preamble size > 0 ifTrue: [outStream cr; nextPut: $!!; nextChunkPut: preamble; cr].
		outStream nextChunkPut: method decompileString]
		ifFalse:
		[sourceFile := SourceFiles at: method fileIndex.
		preamble size > 0
			ifTrue:    "Copy the preamble"
				[outStream copyPreamble: preamble from: sourceFile at: oldPos]
			ifFalse:
				[sourceFile position: oldPos].
		"Copy the method chunk"
		newPos := outStream position.
		outStream copyMethodChunkFrom: sourceFile.
		sourceFile skipSeparators.      "The following chunk may have ]style["
		sourceFile peek == $] ifTrue: [
			outStream cr; copyMethodChunkFrom: sourceFile].
		moveSource ifTrue:    "Set the new method source pointer"
			[endPos := outStream position.
			method checkOKToAdd: endPos - newPos at: newPos.
			method setSourcePosition: newPos inFile: fileIndex]].
	preamble size > 0 ifTrue: [outStream nextChunkPut: ' '].
	^ outStream cr! !

!TClassAndTraitDescription methodsFor: 'printing' stamp: 'al 5/9/2004 16:50'!
printOn: aStream 
	aStream nextPutAll: self name! !

!TClassAndTraitDescription methodsFor: 'printing' stamp: 'al 5/9/2004 16:50'!
printOnStream: aStream 
	aStream print: self name! !

!TClassAndTraitDescription methodsFor: 'filein/out' stamp: 'al 5/9/2004 18:56'!
putClassCommentToCondensedChangesFile: aFileStream
	"Called when condensing changes.  If the receiver has a class comment, and if that class comment does not reside in the .sources file, then write it to the given filestream, with the resulting RemoteString being reachable from the source file #2.  Note that any existing backpointer into the .sources file is lost by this process -- a situation that maybe should be fixed someday."

	| header aStamp aCommentRemoteStr |
	self isMeta ifTrue: [^ self].  "bulletproofing only"
	((aCommentRemoteStr := self organization commentRemoteStr) isNil or:
		[aCommentRemoteStr sourceFileNumber == 1]) ifTrue: [^ self].

	aFileStream cr; nextPut: $!!.
	header := String streamContents: [:strm | strm nextPutAll: self name;
		nextPutAll: ' commentStamp: '.
		(aStamp := self organization commentStamp ifNil: ['<historical>']) storeOn: strm.
		strm nextPutAll: ' prior: 0'].
	aFileStream nextChunkPut: header.
	aFileStream cr.
	self organization classComment: (RemoteString newString: self organization classComment onFileNumber: 2 toFile: aFileStream) stamp: aStamp! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/9/2004 19:17'!
reformatAll
	"Reformat all methods in this class.
	Leaves old code accessible to version browsing"
	self selectorsDo: [:sel | self reformatMethodAt: sel]! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'dvf 9/27/2005 17:18'!
reformatMethodAt: selector
	| newCodeString method |
	newCodeString := self prettyPrinterClass 
				format: (self sourceCodeAt: selector)
				in: self
				notifying: nil
				decorated: false.
	method := self compiledMethodAt: selector.
	method 
		putSource: newCodeString
		fromParseNode: nil
		class: self
		category: (self organization categoryOfElement: selector)
		inFile: 2
		priorMethod: method
! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 5/8/2004 19:50'!
removeCategory: aString 
	"Remove each of the messages categorized under aString in the method 
	dictionary of the receiver. Then remove the category aString."
	| categoryName |
	categoryName := aString asSymbol.
	(self organization listAtCategoryNamed: categoryName) do:
		[:sel | self removeSelector: sel].
	self organization removeCategory: categoryName! !

!TClassAndTraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 5/8/2004 20:49'!
removeSelector: selector 
	"Remove the message whose selector is given from the method 
	dictionary of the receiver, if it is there. Answer nil otherwise."
	
	| priorMethod priorProtocol | 
	priorMethod := self compiledMethodAt: selector ifAbsent: [^ nil].
	priorProtocol := self whichCategoryIncludesSelector: selector.
	super removeSelector: selector.
	SystemChangeNotifier uniqueInstance 
		doSilently: [self updateOrganizationSelector: selector oldCategory: priorProtocol newCategory: nil].
	SystemChangeNotifier uniqueInstance 
			methodRemoved: priorMethod selector: selector inProtocol: priorProtocol class: self.! !

!TClassAndTraitDescription methodsFor: 'organization' stamp: 'al 5/9/2004 19:19'!
reorganize
	"During fileIn, !!Rectangle reorganize!! allows Rectangle to seize control and treat the next chunk as its organization.  See the transfer of control where ReadWriteStream fileIn calls scanFrom:"

	^self organization! !

!TClassAndTraitDescription methodsFor: 'printing' stamp: 'al 5/9/2004 16:50'!
storeOn: aStream
	"Classes and Metaclasses have global names."

	aStream nextPutAll: self name! !

!TClassAndTraitDescription methodsFor: 'organization updating' stamp: 'al 3/27/2006 10:21'!
updateOrganizationSelector: aSymbol oldCategory: oldCategoryOrNil newCategory: newCategoryOrNil
	| currentCategory effectiveCategory sel changedCategories composition |
	changedCategories := IdentitySet new.
	composition := self hasTraitComposition
		ifTrue: [self traitComposition]
		ifFalse: [TraitComposition new].
	(composition methodDescriptionsForSelector: aSymbol) do: [:each |
		sel := each selector.
		(self includesLocalSelector: sel) ifFalse: [
			currentCategory := self organization categoryOfElement: sel.
			effectiveCategory := each effectiveMethodCategoryCurrent: currentCategory new: newCategoryOrNil.
			effectiveCategory isNil ifTrue: [
				currentCategory ifNotNil: [changedCategories add: currentCategory].
				self organization removeElement: sel.
			] ifFalse: [
				((currentCategory isNil or: [currentCategory == ClassOrganizer ambiguous or: [currentCategory == oldCategoryOrNil]]) and: [currentCategory ~~ effectiveCategory]) ifTrue: [
					currentCategory ifNotNil: [changedCategories add: currentCategory].
					self organization 
						classify: sel 
						under: effectiveCategory
						suppressIfDefault: false]]]].
	^ changedCategories! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:13'!
wantsChangeSetLogging
	"Answer whether code submitted for the receiver should be remembered by the changeSet mechanism.  7/12/96 sw"

	^ true! !

!TClassAndTraitDescription methodsFor: 'compiling' stamp: 'al 5/8/2004 20:13'!
wantsRecompilationProgressReported
	"Answer whether the receiver would like progress of its recompilation reported interactively to the user."

	^ true! !

!TClassAndTraitDescription methodsFor: 'organization' stamp: 'al 5/8/2004 19:15'!
whichCategoryIncludesSelector: aSelector 
	"Answer the category of the argument, aSelector, in the organization of 
	the receiver, or answer nil if the receiver does not inlcude this selector."

	(self includesSelector: aSelector)
		ifTrue: [^ self organization categoryOfElement: aSelector]
		ifFalse: [^nil]! !

!TClassAndTraitDescription methodsFor: 'organization' stamp: 'al 5/8/2004 19:29'!
zapOrganization
	"Remove the organization of this class by message categories.
	This is typically done to save space in small systems.  Classes and methods
	created or filed in subsequently will, nonetheless, be organized"

	self organization: nil.
	self isClassSide ifFalse: [self classSide zapOrganization]! !

Trait named: #TComposingDescription
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TComposingDescription methodsFor: 'composition' stamp: 'apb 8/22/2005 17:31'!
+ aTraitOrTraitComposition
	"Use double dispatch to avoid having nested composition in cases where
	parenthesis are used, such as T1 + (T2 + T3)"
	
	^aTraitOrTraitComposition addOnTheLeft: self! !

!TComposingDescription methodsFor: 'composition' stamp: 'apb 8/22/2005 17:12'!
- anArrayOfSelectors
	^TraitExclusion
		with: self
		exclusions: anArrayOfSelectors! !

!TComposingDescription methodsFor: 'composition' stamp: 'apb 8/22/2005 17:02'!
@ anArrayOfAssociations 
	^ TraitAlias with: self aliases: anArrayOfAssociations! !

!TComposingDescription methodsFor: 'private' stamp: 'apb 8/22/2005 17:53'!
addCompositionOnLeft: aTraitComposition
	^ aTraitComposition add: self! !

!TComposingDescription methodsFor: 'private' stamp: 'apb 8/22/2005 17:40'!
addOnTheLeft: aTraitExpression
	^TraitComposition with: aTraitExpression with: self! !

!TComposingDescription methodsFor: 'converting' stamp: 'apb 8/22/2005 15:47'!
asTraitComposition
	^TraitComposition with: self! !

Trait named: #TPureBehavior
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 8/2/2004 16:26'!
>> selector 
	"Answer the compiled method associated with the argument, selector (a 
	Symbol), a message selector in the receiver's method dictionary. If the 
	selector is not in the dictionary, create an error notification."

	^self compiledMethodAt: selector
! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:51'!
addExclusionOf: aSymbol to: aTrait
	self setTraitComposition: (
		self traitComposition copyWithExclusionOf: aSymbol to: aTrait)! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 7/30/2004 22:43'!
addSelector: selector withMethod: compiledMethod 
	^ self addSelector: selector withMethod: compiledMethod notifying: nil! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 7/30/2004 22:43'!
addSelector: selector withMethod: compiledMethod notifying: requestor
	^ self addSelectorSilently: selector withMethod: compiledMethod! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 9/16/2005 14:31'!
addSelectorSilently: selector withMethod: compiledMethod
	self methodDictAddSelectorSilently: selector withMethod: compiledMethod.
	self registerLocalSelector: selector! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 1/9/2006 17:58'!
addToComposition: aTrait
	self setTraitComposition: (self traitComposition copyTraitExpression
		add: aTrait;
		yourself)! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 8/16/2004 15:36'!
addTraitSelector: aSymbol withMethod: aCompiledMethod
	"Add aMethod with selector aSymbol to my
	methodDict. aMethod must not be defined locally.
	Note that I am overridden by ClassDescription
	to do a recompilation of the method if it has supersends."

	self assert: [(self includesLocalSelector: aSymbol) not].
	self ensureLocalSelectors.
	self basicAddSelector: aSymbol withMethod: aCompiledMethod.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:12'!
allSelectors
	self explicitRequirement! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:39'!
applyChangesOfNewTraitCompositionReplacing: oldComposition
	| changedSelectors |
	changedSelectors := self traitComposition
		changedSelectorsComparedTo: oldComposition.
	changedSelectors isEmpty ifFalse: [
		self noteChangedSelectors: changedSelectors].
	self traitComposition isEmpty ifTrue: [
		self purgeLocalSelectors].
	^changedSelectors! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'marcus.denker 9/17/2008 19:07'!
basicAddSelector: selector withMethod: compiledMethod 
	"Add the message selector with the corresponding compiled method to the 
	receiver's method dictionary.
	Do this without sending system change notifications"

	| oldMethodOrNil |
	oldMethodOrNil := self lookupSelector: selector.
	self methodDict at: selector put: compiledMethod.
	compiledMethod methodClass: self.
	compiledMethod selector: selector.

	"Now flush Squeak's method cache, either by selector or by method"
	oldMethodOrNil ifNotNil: [oldMethodOrNil flushCache].
	selector flushCache.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:13'!
basicLocalSelectors
	self explicitRequirement! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:13'!
basicLocalSelectors: aSetOrNil
	self explicitRequirement! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 7/30/2004 22:43'!
basicRemoveSelector: selector 
	"Assuming that the argument, selector (a Symbol), is a message selector 
	in my method dictionary, remove it and its method."

	| oldMethod |
	oldMethod := self methodDict at: selector ifAbsent: [^ self].
	self methodDict removeKey: selector.

	"Now flush Squeak's method cache, either by selector or by method"
	oldMethod flushCache.
	selector flushCache! !

!TPureBehavior methodsFor: 'compiling' stamp: 'md 3/5/2006 23:49'!
binding
	^ nil -> self! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 11/28/2005 11:28'!
bindingOf: varName
	
	"Answer the binding of some variable resolved in the scope of the receiver"
	| aSymbol binding |
	aSymbol := varName asSymbol.

	"Look in declared environment."
	binding := self environment bindingOf: aSymbol.
	^binding! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 9/16/2005 14:16'!
canUnderstand: selector
	"Answer whether the receiver can respond to the message whose selector 
	is the argument."

	^ self classAndMethodFor: selector do: [:c :m | m isProvided] ifAbsent: [false].! !

!TPureBehavior methodsFor: 'testing' stamp: 'al 9/16/2005 14:23'!
canZapMethodDictionary
	"Return true if it is safe to zap the method dictionary on #obsolete"
	^true! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 11/28/2005 21:48'!
changeRecordsAt: selector
	"Return a list of ChangeRecords for all versions of the method at selector. Source code can be retrieved by sending string to any one.  Return nil if the method is absent."

	"(Pen changeRecordsAt: #go:) collect: [:cRec | cRec string]"
	^ChangeSet 
		scanVersionsOf: (self compiledMethodAt: selector ifAbsent: [^ nil])
		class: self meta: self isMeta
		category: (self whichCategoryIncludesSelector: selector)
		selector: selector.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:14'!
classAndMethodFor: aSymbol do: binaryBlock ifAbsent: absentBlock
	"Looks up the selector aSymbol in this class/trait. If it is found, binaryBlock is evaluated
	with this class/trait and the associated method. Otherwise absentBlock is evaluated.
	Note that this implementation is very simple because PureBehavior does not know
	about inheritance (cf. implementation in Behavior)"
	
	^ binaryBlock value: self value: (self compiledMethodAt: aSymbol ifAbsent: [^ absentBlock value]).! !

!TPureBehavior methodsFor: 'send caches' stamp: 'al 9/16/2005 14:41'!
clearSendCaches
	LocalSends current clearOut: self! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 8/2/2004 19:34'!
compile: code 
	"Compile the argument, code, as source code in the context of the 
	receiver. Create an error notification if the code can not be compiled. 
	The argument is either a string or an object that converts to a string or a 
	PositionableStream on an object that converts to a string."

	^self compile: code notifying: nil! !

!TPureBehavior methodsFor: 'compiling' stamp: 'md 2/28/2006 10:04'!
compile: code classified: category notifying: requestor trailer: bytes ifFail: failBlock
	"Compile code without logging the source in the changes file"

	| methodNode |
	methodNode  := self compilerClass new
				compile: code
				in: self
				classified: category 
				notifying: requestor
				ifFail: failBlock.
	^ CompiledMethodWithNode generateMethodFromNode: methodNode trailer: bytes.! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 11/28/2005 21:52'!
compile: code notifying: requestor 
	"Compile the argument, code, as source code in the context of the 
	receiver and insEtall the result in the receiver's method dictionary. The 
	second argument, requestor, is to be notified if an error occurs. The 
	argument code is either a string or an object that converts to a string or 
	a PositionableStream. This method also saves the source code."
	
	| methodAndNode |
	methodAndNode  := self
		compile: code "a Text"
		classified: nil
		notifying: requestor
		trailer: self defaultMethodTrailer
		ifFail: [^nil].
	methodAndNode method putSource: code fromParseNode: methodAndNode node inFile: 2
			withPreamble: [:f | f cr; nextPut: $!!; nextChunkPut: 'Behavior method'; cr].
	self addSelector: methodAndNode selector withMethod: methodAndNode method notifying: requestor.
	^ methodAndNode selector! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 7/30/2004 22:02'!
compileAll
	^ self compileAllFrom: self! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 7/30/2004 22:02'!
compileAllFrom: oldClass
	"Compile all the methods in the receiver's method dictionary.
	This validates sourceCode and variable references and forces
	all methods to use the current bytecode set"
	"ar 7/10/1999: Use oldClass selectors not self selectors"
	
	oldClass selectorsDo: [:sel | self recompile: sel from: oldClass].
	self environment currentProjectDo: [:proj | proj compileAllIsolated: self from: oldClass].! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 8/2/2004 16:29'!
compiledMethodAt: selector 
	"Answer the compiled method associated with the argument, selector (a 
	Symbol), a message selector in the receiver's method dictionary. If the 
	selector is not in the dictionary, create an error notification."

	^ self methodDict at: selector! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 20:55'!
compiledMethodAt: selector ifAbsent: aBlock
	"Answer the compiled method associated with the argument, selector (a Symbol), a message selector in the receiver's method dictionary. If the selector is not in the dictionary, return the value of aBlock"

	^ self methodDict at: selector ifAbsent: [aBlock value]! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 11/28/2005 11:28'!
compilerClass
	"Answer a compiler class appropriate for source methods of this class."

	^Compiler! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 20:55'!
compress
	"Compact the method dictionary of the receiver."

	self methodDict rehash! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:01'!
compressedSourceCodeAt: selector
	"(Paragraph compressedSourceCodeAt: #displayLines:affectedRectangle:) size 721 1921
	Paragraph selectors inject: 0 into: [:tot :sel | tot + (Paragraph compressedSourceCodeAt: sel) size] 13606 31450"
	| rawText parse |
	rawText := (self sourceCodeAt: selector) asString.
	parse := self compilerClass new parse: rawText in: self notifying: nil.
	^ rawText compressWithTable:
		((selector keywords ,
		parse tempNames ,
		self instVarNames ,
		#(self super ifTrue: ifFalse:) ,
		((0 to: 7) collect:
			[:i | String streamContents:
				[:s | s cr. i timesRepeat: [s tab]]]) ,
		(self compiledMethodAt: selector) literalStrings)
			asSortedCollection: [:a :b | a size > b size])! !

!TPureBehavior methodsFor: 'copying' stamp: 'al 9/16/2005 14:19'!
copy
	"Answer a copy of the receiver without a list of subclasses."

	| myCopy |
	myCopy := self shallowCopy.
	^myCopy methodDictionary: self copyOfMethodDictionary! !

!TPureBehavior methodsFor: 'copying' stamp: 'al 7/30/2004 22:51'!
copyOfMethodDictionary
	"Return a copy of the receiver's method dictionary"

	^ self methodDict copy! !

!TPureBehavior methodsFor: 'user interface' stamp: 'al 7/30/2004 22:38'!
crossReference
	"Answer an Array of arrays of size 2 whose first element is a message selector in the receiver's method dictionary and whose second element is a set of all message selectors in the method dictionary whose methods send a message with that selector. Subclasses are not included."

	^self selectors asSortedCollection asArray collect: [:x | 		Array 
			with: (String with: Character cr), x 
			with: (self whichSelectorsReferTo: x)]

	"Point crossReference."! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 11/28/2005 11:28'!
decompile: selector 
	"Find the compiled code associated with the argument, selector, as a 
	message selector in the receiver's method dictionary and decompile it. 
	Answer the resulting source code as a string. Create an error notification 
	if the selector is not in the receiver's method dictionary."

	^self decompilerClass new decompile: selector in: self! !

!TPureBehavior methodsFor: 'compiling' stamp: 'md 3/1/2006 20:01'!
decompilerClass
	"Answer a decompiler class appropriate for compiled methods of this class."

	^ self compilerClass decompilerClass! !

!TPureBehavior methodsFor: 'copying' stamp: 'al 9/16/2005 14:20'!
deepCopy
	"Classes should only be shallowCopied or made anew."

	^ self shallowCopy! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 8/2/2004 19:57'!
defaultMethodTrailer
	^ #(0 0 0 0)! !

!TPureBehavior methodsFor: 'printing' stamp: 'al 9/16/2005 14:21'!
defaultNameStemForInstances
	"Answer a basis for external names for default instances of the receiver.
	For classees, the class-name itself is a good one."

	^ self name! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 11/28/2005 11:46'!
deregisterLocalSelector: aSymbol
	self basicLocalSelectors notNil ifTrue: [
		self basicLocalSelectors remove: aSymbol ifAbsent: []]! !

!TPureBehavior methodsFor: 'initialization' stamp: 'al 12/1/2005 16:01'!
emptyMethodDictionary

	^ MethodDictionary new! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 11/28/2005 11:47'!
ensureLocalSelectors
	"Ensures that the instance variable localSelectors is effectively used to maintain
	the set of local selectors.
	This method must be called before any non-local selectors are added to the
	method dictionary!!"

	self basicLocalSelectors isNil 
		ifTrue: [self basicLocalSelectors: self selectors]! !

!TPureBehavior methodsFor: 'naming' stamp: 'al 9/16/2005 14:22'!
environment
	"Return the environment in which the receiver is visible"
	^Smalltalk! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 11/28/2005 11:28'!
evaluatorClass
	"Answer an evaluator class appropriate for evaluating expressions in the 
	context of this class."

	^Compiler! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 12/1/2005 15:40'!
firstCommentAt:  selector
	"Answer a string representing the first comment in the method associated with selector.  Return an empty string if the relevant source file is not available, or if the method's source code does not contain a comment.  Not smart enough to bypass quotes in string constants, but does map doubled quote into a single quote."

	|someComments|
	someComments := self commentsAt: selector.
	^someComments isEmpty ifTrue: [''] ifFalse: [someComments first]


"Behavior firstCommentAt: #firstCommentAt:"! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 11/28/2005 21:54'!
firstPrecodeCommentFor:  selector
	"If there is a comment in the source code at the given selector that preceeds the body of the method, return it here, else return nil"

	| parser source tree |
	"Behavior firstPrecodeCommentFor: #firstPrecodeCommentFor:"
	(#(Comment Definition Hierarchy) includes: selector)
		ifTrue:
			["Not really a selector"
			^ nil].
	source := self sourceCodeAt: selector asSymbol ifAbsent: [^ nil].
	parser := self parserClass new.
	tree := 
		parser
			parse: (ReadStream on: source)
			class: self
			noPattern: false
			context: nil
			notifying: nil
			ifFail: [^ nil].
	^ (tree comment ifNil: [^ nil]) first! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 10/22/2006 18:21'!
flattenDown: aTrait
	| selectors |
	self assert: [self hasTraitComposition and: [self traitComposition allTraits includes: aTrait]].
	selectors := (self traitComposition transformationOfTrait: aTrait) selectors.
	self basicLocalSelectors: self basicLocalSelectors , selectors.
	self removeFromComposition: aTrait.! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 10/22/2006 18:21'!
flattenDownAllTraits
	self traitComposition allTraits do: [:each | self flattenDown: each].
	self assert: [ self traitComposition isEmpty ].
	self traitComposition: nil.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:49'!
"popeye" formalHeaderPartsFor: "olive oil" aSelector
	"RELAX!!  The warning you may have just seen about possibly having a bad source file does not apply here, because this method *intends* to have its source code start with a comment.
	This method returns a collection giving the parts in the formal declaration for aSelector.  This parse is in support of schemes in which adjutant properties of a method can be declared via special comments secreted in the formal header
	The result will have
     	3 elements for a simple, argumentless selector.
		5 elements for a single-argument selector
		9 elements for a two-argument selector
		13 elements for a three-argument, selector
		etc...

	The syntactic elements are:

		1		comment preceding initial selector fragment

		2		first selector fragment
		3		comment following first selector fragment  (nil if selector has no arguments)

        ----------------------  (ends here for, e.g., #copy)

		4		first formal argument
		5		comment following first formal argument (nil if selector has only one argument)

        ----------------------  (ends here for, e.g., #copyFrom:)

		6		second keyword
		7		comment following second keyword
		8		second formal argument
		9		comment following second formal argument (nil if selector has only two arguments)

         ----------------------  (ends here for, e.g., #copyFrom:to:)

	Any nil element signifies an absent comment.
	NOTE: The comment following the final formal argument is *not* successfully retrieved by this method in its current form, though it can be obtained, if needed, by other means (e.g. calling #firstPrecodeCommentFor:).  Thus, the *final* element in the structure returned by this method is always going to be nil."

	^ Scanner new scanMessageParts: (self methodHeaderFor: aSelector)

"
	Behavior class formalHeaderPartsFor: #formalHeaderPartsFor:
"


	! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:49'!
formalParametersAt: aSelector
	"Return the names of the arguments used in this method."

	| source parser message list params |
	source := self sourceCodeAt: aSelector ifAbsent: [^ #()].	"for now"
	(parser := self parserClass new) parseSelector: source.
	message := source copyFrom: 1 to: (parser endOfLastToken min: source size).
	list := message string findTokens: Character separators.
	params := OrderedCollection new.
	list withIndexDo: [:token :ind | ind even ifTrue: [params addLast: token]].
	^ params! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 7/30/2004 22:35'!
fullyImplementsVocabulary: aVocabulary
	"Answer whether instances of the receiver respond to all the messages in aVocabulary"

	(aVocabulary encompassesAPriori: self) ifTrue: [^ true].
	aVocabulary allSelectorsInVocabulary do:
		[:aSelector | (self canUnderstand: aSelector) ifFalse: [^ false]].
	^ true! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'marcus.denker 8/25/2008 09:23'!
hasMethods
	"Answer whether the receiver has any methods in its method dictionary."

	^ self methodDict notEmpty! !

!TPureBehavior methodsFor: 'send caches' stamp: 'al 9/16/2005 14:41'!
hasRequiredSelectors
	^ self requiredSelectors notEmpty! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:29'!
hasTraitComposition
	self explicitRequirement! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 7/30/2004 22:35'!
implementsVocabulary: aVocabulary
	"Answer whether instances of the receiver respond to the messages in aVocabulary."

	(aVocabulary isKindOf: FullVocabulary orOf: ScreenedVocabulary) ifTrue: [^ true].
	^ self fullyImplementsVocabulary: aVocabulary! !

!TPureBehavior methodsFor: 'testing' stamp: 'al 3/1/2006 23:02'!
includesBehavior: aBehavior
	^self == aBehavior! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 9/16/2005 14:16'!
includesLocalSelector: aSymbol
	^self basicLocalSelectors isNil
		ifTrue: [self includesSelector: aSymbol]
		ifFalse: [self localSelectors includes: aSymbol]! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 7/30/2004 22:35'!
includesSelector: aSymbol 
	"Answer whether the message whose selector is the argument is in the 
	method dictionary of the receiver's class."

	^ self methodDict includesKey: aSymbol! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 7/30/2004 22:37'!
isAliasSelector: aSymbol
	"Return true if the selector aSymbol is an alias defined
	in my or in another composition somewhere deeper in 
	the tree of traits compositions."

	^(self includesLocalSelector: aSymbol) not
		and: [self hasTraitComposition]
		and: [self traitComposition isAliasSelector: aSymbol]! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 9/16/2005 14:16'!
isDisabledSelector: selector
	^ self classAndMethodFor: selector do: [:c :m | m isDisabled] ifAbsent: [false]! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 7/30/2004 22:37'!
isLocalAliasSelector: aSymbol
	"Return true if the selector aSymbol is an alias defined
	in my trait composition."

	^(self includesLocalSelector: aSymbol) not
		and: [self hasTraitComposition]
		and: [self traitComposition isLocalAliasSelector: aSymbol]! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 9/16/2005 14:16'!
isProvidedSelector: selector
	^ ProvidedSelectors current isSelector: selector providedIn: self
! !

!TPureBehavior methodsFor: 'printing' stamp: 'al 12/1/2005 15:25'!
literalScannedAs: scannedLiteral notifying: requestor
	"Postprocesses a literal scanned by Scanner scanToken (esp. xLitQuote).
	If scannedLiteral is not an association, answer it.
	Else, if it is of the form:
		nil->#NameOfMetaclass
	answer nil->theMetaclass, if any has that name, else report an error.
	Else, if it is of the form:
		#NameOfGlobalVariable->anythiEng
	answer the global, class, or pool association with that nameE, if any, else
	add it to Undeclared a answer the new Association."

	| key value |
	(scannedLiteral isVariableBinding)
		ifFalse: [^ scannedLiteral].
	key := scannedLiteral key.
	value := scannedLiteral value.
	key isNil 
		ifTrue: "###<metaclass soleInstance name>"
			[(self bindingOf: value) ifNotNilDo:[:assoc|
				 (assoc value isKindOf: Behavior)
					ifTrue: [^ nil->assoc value class]].
			 requestor notify: 'No such metaclass'.
			 ^false].
	(key isSymbol)
		ifTrue: "##<global var name>"
			[(self bindingOf: key) ifNotNilDo:[:assoc | ^assoc].
			Undeclared at: key put: nil.
			 ^Undeclared bindingOf: key].
	requestor notify: '## must be followed by a non-local variable name'.
	^false

"	Form literalScannedAs: 14 notifying: nil 14
	Form literalScannedAs: #OneBitForm notiEfying: nil  OneBitForm
	Form literalScannedAs: ##OneBitForm notifying: nil  OneBitForm->a Form
	Form literalScannedAs: ##Form notifying: nil   Form->Form
	Form literalScannedAs: ###Form notifying: nil   nilE->Form class
"! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 11/28/2005 11:47'!
localSelectors
	"Return a set of selectors defined locally.
	The instance variable is lazily initialized. If it is nil then there
	are no non-local selectors"

	^ self basicLocalSelectors isNil
		ifTrue: [self selectors]
		ifFalse: [self basicLocalSelectors].! !

!TPureBehavior methodsFor: 'printing' stamp: 'al 9/16/2005 14:21'!
longPrintOn: aStream
	"Append to the argument, aStream, the names and values of all of the receiver's instance variables.  But, not useful for a class with a method dictionary."

	aStream nextPutAll: '<<too complex to show>>'; cr.! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:50'!
lookupSelector: selector
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:50'!
methodDict
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:50'!
methodDict: aDictionary
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 11/28/2005 11:31'!
methodDictAddSelectorSilently: selector withMethod: compiledMethod
	self basicAddSelector: selector withMethod: compiledMethod! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:02'!
methodDictionary
	"Convenience"
	^self methodDict! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/30/2005 18:24'!
methodDictionary: aDictionary
	self methodDict: aDictionary! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:50'!
methodHeaderFor: selector 
	"Answer the string corresponding to the method header for the given selector"

	| sourceString parser |
	sourceString := self ultimateSourceCodeAt: selector ifAbsent: [self standardMethodHeaderFor: selector].
	(parser := self parserClass new) parseSelector: sourceString.
	^ sourceString asString copyFrom: 1 to: (parser endOfLastToken min: sourceString size)

	"Behavior methodHeaderFor: #methodHeaderFor: "
! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:02'!
methodsDo: aBlock
	"Evaluate aBlock for all the compiled methods in my method dictionary."

	^ self methodDict valuesDo: aBlock! !

!TPureBehavior methodsFor: 'naming' stamp: 'al 9/16/2005 14:21'!
name
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 8/6/2004 13:09'!
noteChangedSelectors: aCollection
	"Start update of my methodDict (after changes to traits in traitComposition
	or after a local method was removed from my methodDict). The argument 
	is a collection of method selectors that may have been changed. Most of the time
	aCollection only holds one selector. But when there are aliases involved 
	there may be several method changes that have to be propagated to users."

	| affectedSelectors |
	affectedSelectors := IdentitySet new.
	aCollection do: [:selector |
		affectedSelectors addAll: (self updateMethodDictionarySelector: selector)].
	self notifyUsersOfChangedSelectors: affectedSelectors.
	^ affectedSelectors! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:40'!
notifyUsersOfChangedSelector: aSelector
	self notifyUsersOfChangedSelectors: (Array with: aSelector)! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:40'!
notifyUsersOfChangedSelectors: aCollection! !

!TPureBehavior methodsFor: 'initialization' stamp: 'al 12/1/2005 15:24'!
obsolete
	"Invalidate and recycle local methods,
	e.g., zap the method dictionary if can be done safely."
	self canZapMethodDictionary
		ifTrue: [self methodDict: self emptyMethodDictionary].
	self hasTraitComposition ifTrue: [
		self traitComposition traits do: [:each |
			each removeUser: self]]! !

!TPureBehavior methodsFor: 'newcompiler' stamp: 'md 3/1/2006 21:23'!
parseScope

	^ Smalltalk at: #ClassScope ifPresent: [:class | class new class: self]! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 11/28/2005 11:28'!
parserClass
	"Answer a parser class to use for parsing method headers."

	^self compilerClass parserClass! !

!TPureBehavior methodsFor: 'printing' stamp: 'al 12/7/2005 21:05'!
prettyPrinterClass
	^self compilerClass! !

!TPureBehavior methodsFor: '*traits' stamp: 'al 9/16/2005 14:48'!
providedSelectors
	^ProvidedSelectors current for: self! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 11/28/2005 11:47'!
purgeLocalSelectors
	self basicLocalSelectors: nil! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 7/30/2004 22:03'!
recompile: selector
	"Compile the method associated with selector in the receiver's method dictionary."
	^self recompile: selector from: self! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 7/30/2004 22:03'!
recompile: selector from: oldClass
	"Compile the method associated with selector in the receiver's method dictionary."
	"ar 7/10/1999: Use oldClass compiledMethodAt: not self compiledMethodAt:"
	| method trailer methodNode |
	method := oldClass compiledMethodAt: selector.
	trailer := method trailer.
	methodNode := self compilerClass new
				compile: (oldClass sourceCodeAt: selector)
				in: self
				notifying: nil
				ifFail: [^ self].   "Assume OK after proceed from SyntaxError"
	selector == methodNode selector ifFalse: [self error: 'selector changed!!'].
	self basicAddSelector: selector withMethod: (methodNode generate: trailer).
! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 7/30/2004 22:17'!
recompileChanges
	"Compile all the methods that are in the changes file.
	This validates sourceCode and variable references and forces
	methods to use the current bytecode set"

	self selectorsDo:
		[:sel | (self compiledMethodAt: sel) fileIndex > 1 ifTrue:
			[self recompile: sel from: self]]! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 7/30/2004 22:17'!
recompileNonResidentMethod: method atSelector: selector from: oldClass
	"Recompile the method supplied in the context of this class."

	| trailer methodNode |
	trailer := method trailer.
	methodNode := self compilerClass new
			compile: (method getSourceFor: selector in: oldClass)
			in: self
			notifying: nil
			ifFail: ["We're in deep doo-doo if this fails (syntax error).
				Presumably the user will correct something and proceed,
				thus installing the result in this methodDict.  We must
				retrieve that new method, and restore the original (or remove)
				and then return the method we retrieved."
				^ self error: 'see comment'].
	selector == methodNode selector ifFalse: [self error: 'selector changed!!'].
	^ methodNode generate: trailer
! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 11/28/2005 11:47'!
registerLocalSelector: aSymbol
	self basicLocalSelectors notNil ifTrue: [
		self basicLocalSelectors add: aSymbol]! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:54'!
removeAlias: aSymbol of: aTrait
	self setTraitComposition: (
		self traitComposition copyWithoutAlias: aSymbol of: aTrait)! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 10/22/2006 18:22'!
removeFromComposition: aTrait
	self setTraitComposition: (self traitComposition copyTraitExpression
		removeFromComposition: aTrait)! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 3/26/2006 21:37'!
removeSelector: aSelector 
	"Assuming that the argument, selector (a Symbol), is a message selector 
	in my method dictionary, remove it and its method.
	
	If the method to remove will be replaced by a method from my trait composition,
	the current method does not have to be removed because we mark it as non-local.
	If it is not identical to the actual method from the trait it will be replaced automatically
	by #noteChangedSelectors:.
	
	This is useful to avoid bootstrapping problems when moving methods to a trait
	(e.g., from TPureBehavior to TMethodDictionaryBehavior). Manual moving (implementing
	the method in the trait and then remove it from the class) does not work if the methods
	themselves are used for this process (such as compiledMethodAt:, includesLocalSelector: or
	addTraitSelector:withMethod:)"

	| changeFromLocalToTraitMethod |
	changeFromLocalToTraitMethod := (self includesLocalSelector: aSelector)
		and: [self hasTraitComposition]
		and: [self traitComposition includesMethod: aSelector].

	changeFromLocalToTraitMethod
		ifFalse: [self basicRemoveSelector: aSelector]
		ifTrue: [self ensureLocalSelectors].
	self deregisterLocalSelector: aSelector.
	self noteChangedSelectors: (Array with: aSelector)
	
! !

!TPureBehavior methodsFor: 'adding/removing methods' stamp: 'al 7/30/2004 22:50'!
removeSelectorSilently: selector 
	"Remove selector without sending system change notifications"

	^ SystemChangeNotifier uniqueInstance doSilently: [self removeSelector: selector].! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:46'!
removeTraitSelector: aSymbol
	self assert: [(self includesLocalSelector: aSymbol) not].
	self basicRemoveSelector: aSymbol! !

!TPureBehavior methodsFor: 'send caches' stamp: 'al 9/16/2005 14:41'!
requirements
	^ self requiredSelectorsCache 
		ifNil: [#()] 
		ifNotNilDo: [:rsc | rsc requirements]! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:03'!
selectors
	"Answer a Set of all the message selectors specified in the receiver's 
	method dictionary."

	^ self methodDict keys! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:03'!
selectorsAndMethodsDo: aBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysAndValuesDo: aBlock! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:03'!
selectorsDo: selectorBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysDo: selectorBlock! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:03'!
selectorsWithArgs: numberOfArgs
	"Return all selectors defined in this class that take this number of arguments.  Could use String.keywords.  Could see how compiler does this."

	| list num |
	list := OrderedCollection new.
	self selectorsDo: [:aSel | 
		num := aSel count: [:char | char == $:].
		num = 0 ifTrue: [aSel last isLetter ifFalse: [num := 1]].
		num = numberOfArgs ifTrue: [list add: aSel]].
	^ list! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:30'!
selfSentSelectorsFromSelectors: interestingSelectors 
	| m result info |
	result := IdentitySet new.
	interestingSelectors collect: 
			[:sel | 
			m := self compiledMethodAt: sel ifAbsent: [].
			m ifNotNil: 
					[info := (SendInfo on: m) collectSends.
					info selfSentSelectors do: [:sentSelector | result add: sentSelector]]].
	^result! !

!TPureBehavior methodsFor: 'send caches' stamp: 'al 9/16/2005 14:42'!
sendCaches
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'send caches' stamp: 'al 9/16/2005 14:42'!
sendCaches: aSendCaches
	^ self explicitRequirement! !

!TPureBehavior methodsFor: 'send caches' stamp: 'al 9/16/2005 14:42'!
setRequiredStatusOf: selector to: aBoolean
	aBoolean 
		ifTrue: [self requiredSelectorsCache addRequirement: selector]
		ifFalse: [self requiredSelectorsCache removeRequirement: selector].! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:48'!
setTraitComposition: aTraitComposition
	| oldComposition |
	(self hasTraitComposition not and: [aTraitComposition isEmpty]) ifTrue: [^self].
	aTraitComposition assertValidUser: self.

	oldComposition := self traitComposition.
	self traitComposition: aTraitComposition.
	self applyChangesOfNewTraitCompositionReplacing: oldComposition.
	
	oldComposition traits do: [:each | each removeUser: self].
	aTraitComposition traits do: [:each | each addUser: self]! !

!TPureBehavior methodsFor: 'traits' stamp: 'apb 8/22/2005 18:28'!
setTraitCompositionFrom: aTraitExpression
	^ self setTraitComposition: aTraitExpression asTraitComposition! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:04'!
sourceCodeAt: selector

	^ (self methodDict at: selector) getSourceFor: selector in: self! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:04'!
sourceCodeAt: selector ifAbsent: aBlock

	^ (self methodDict at: selector ifAbsent: [^ aBlock value]) getSourceFor: selector in: self! !

!TPureBehavior methodsFor: 'compiling' stamp: 'al 9/16/2005 14:17'!
sourceCodeTemplate
	"Answer an expression to be edited and evaluated in order to define 
	methods in this class or trait."

	^'message selector and argument names
	"comment stating purpose of message"

	| temporary variable names |
	statements'! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:04'!
sourceMethodAt: selector 
	"Answer the paragraph corresponding to the source code for the 
	argument."

	^(self sourceCodeAt: selector) asText makeSelectorBoldIn: self! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/30/2004 21:05'!
sourceMethodAt: selector ifAbsent: aBlock
	"Answer the paragraph corresponding to the source code for the 
	argument."

	^ (self sourceCodeAt: selector ifAbsent: [^ aBlock value]) asText makeSelectorBoldIn: self! !

!TPureBehavior methodsFor: 'private' stamp: 'al 12/1/2005 15:20'!
spaceUsed
	"Answer a rough estimate of number of bytes used by this class and its metaclass. Does not include space used by class variables."

	| space method |
	space := 0.
	self selectorsDo: [:sel |
		space := space + 16.  "dict and org'n space"
		method := self compiledMethodAt: sel.
		space := space + (method size + 6 "hdr + avg pad").
		method literals do: [:lit |
			(lit isMemberOf: Array) ifTrue: [space := space + ((lit size + 1) * 4)].
			(lit isMemberOf: Float) ifTrue: [space := space + 12].
			(lit isMemberOf: ByteString) ifTrue: [space := space + (lit size + 6)].
			(lit isMemberOf: LargeNegativeInteger) ifTrue: [space := space + ((lit size + 1) * 4)].
			(lit isMemberOf: LargePositiveInteger) ifTrue: [space := space + ((lit size + 1) * 4)]]].
		^ space! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:51'!
standardMethodHeaderFor: aSelector
	| args |
	args := (1 to: aSelector numArgs)	collect:[:i| 'arg', i printString].
	args size = 0 ifTrue:[^aSelector asString].
	args size = 1 ifTrue:[^aSelector,' arg1'].
	^String streamContents:[:s|
		(aSelector findTokens:':') with: args do:[:tok :arg|
			s nextPutAll: tok; nextPutAll:': '; nextPutAll: arg; nextPutAll:' '.
		].
	].
! !

!TPureBehavior methodsFor: 'printing' stamp: 'al 12/1/2005 15:37'!
storeLiteral: aCodeLiteral on: aStream
	"Store aCodeLiteral on aStream, changing an Association to ##GlobalName
	 or ###MetaclassSoleInstanceName format if appropriate"
	| key value |
	(aCodeLiteral isVariableBinding)
		ifFalse:
			[aCodeLiteral storeOn: aStream.
			 ^self].
	key := aCodeLiteral key.
	(key isNil and: [(value := aCodeLiteral value) isMemberOf: Metaclass])
		ifTrue:
			[aStream nextPutAll: '###'; nextPutAll: value soleInstance name.
			 ^self].
	(key isSymbol and: [(self bindingOf: key) notNil])
		ifTrue:
			[aStream nextPutAll: '##'; nextPutAll: key.
			 ^self].
	aCodeLiteral storeOn: aStream! !

!TPureBehavior methodsFor: 'send caches' stamp: 'al 9/16/2005 14:42'!
superRequirements
	^ self requiredSelectorsCache superRequirements! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'marcus.denker 9/29/2008 08:46'!
thoroughWhichSelectorsReferTo: literal special: specialFlag byte: specialByte
	"Answer a set of selectors whose methods access the argument as a 
	literal. Dives into the compact literal notation, making it slow but 
	thorough "

	| selectors |
	selectors := IdentitySet new.
	self selectorsAndMethodsDo: [:sel :method |
		((method hasLiteralThorough: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue: [selectors add: sel]].
	^ selectors! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:30'!
traitComposition
	"Return my trait composition. Manipulating the composition does not
	effect changes automatically. Use #setTraitComposition: to do this but
	note, that you have to make a copy of the old trait composition before
	changing it because only the difference between the new and the old
	composition is updated."
	
	^self explicitRequirement ! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:53'!
traitComposition: aTraitComposition
	^self explicitRequirement ! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:31'!
traitCompositionIncludes: aTrait
	^self == aTrait or: 
		[self hasTraitComposition and: 
			[self traitComposition allTraits includes: aTrait]]! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 3/26/2006 21:29'!
traitCompositionString
	^self hasTraitComposition
		ifTrue: [self traitComposition asString]
		ifFalse: ['{}']! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:55'!
traitOrClassOfSelector: aSymbol
	"Return the trait or the class which originally defines the method aSymbol
	or return self if locally defined or if it is a conflict marker method.
	This is primarly used by Debugger to determin the behavior in which a recompiled
	method should be put. If a conflict method is recompiled it should be put into
	the class, thus return self. Also see TraitComposition>>traitProvidingSelector:"
	
	((self includesLocalSelector: aSymbol) or: [
		self hasTraitComposition not]) ifTrue: [^self].
	^(self traitComposition traitProvidingSelector: aSymbol) ifNil: [self]! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 9/16/2005 14:53'!
traitTransformations 
	^ self traitComposition transformations ! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 7/30/2004 21:55'!
traitsProvidingSelector: aSymbol
	| result |
	result := OrderedCollection new.
	self hasTraitComposition ifFalse: [^result].
	(self traitComposition methodDescriptionsForSelector: aSymbol)
		do: [:methodDescription | methodDescription selector = aSymbol ifTrue: [
			result addAll: (methodDescription locatedMethods
				collect: [:each | each location])]].
	^result! !

!TPureBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/16/2005 14:51'!
ultimateSourceCodeAt: selector ifAbsent: aBlock
	"Return the source code at selector"
	
	^self
		sourceCodeAt: selector
		ifAbsent: aBlock! !

!TPureBehavior methodsFor: 'traits' stamp: 'al 3/26/2006 21:36'!
updateMethodDictionarySelector: aSymbol
	"A method with selector aSymbol in myself or my traitComposition has been changed.
	Do the appropriate update to my methodDict (remove or update method) and
	return all affected selectors of me so that my useres get notified."

	| effectiveMethod modifiedSelectors descriptions selector |
	modifiedSelectors := IdentitySet new.
	descriptions := self hasTraitComposition
		ifTrue: [ self traitComposition methodDescriptionsForSelector: aSymbol ]
		ifFalse: [ #() ].
	descriptions do: [:methodDescription |
		selector := methodDescription selector.
		(self includesLocalSelector: selector) ifFalse: [
			methodDescription isEmpty
				ifTrue: [
					self removeTraitSelector: selector.
					modifiedSelectors add: selector]
				ifFalse: [
					effectiveMethod := methodDescription effectiveMethod.
					(self compiledMethodAt: selector ifAbsent: [nil]) ~~ effectiveMethod ifTrue: [
						self addTraitSelector: selector withMethod: effectiveMethod.
						modifiedSelectors add: selector]]]].
	^modifiedSelectors! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'al 7/30/2004 22:37'!
whichSelectorsReferTo: literal 
	"Answer a Set of selectors whose methods access the argument as a
literal."

	| special byte |
	special := self environment hasSpecialSelector: literal ifTrueSetByte: [:b |
byte := b].
	^self whichSelectorsReferTo: literal special: special byte: byte

	"Rectangle whichSelectorsReferTo: #+."! !

!TPureBehavior methodsFor: 'testing method dictionary' stamp: 'md 2/15/2006 11:01'!
whichSelectorsReferTo: literal special: specialFlag byte: specialByte
	"Answer a set of selectors whose methods access the argument as a literal."

	| who |
	who := IdentitySet new.
	self selectorsAndMethodsDo: 
		[:sel :method |
		((method hasLiteral: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue:
				[((literal isVariableBinding) not
					or: [method literals allButLast includes: literal])
						ifTrue: [who add: sel]]].
	^ who! !

!TPureBehavior methodsFor: 'accessing class hierarchy' stamp: 'al 12/31/2005 13:40'!
withAllSuperclasses
	"Answer an OrderedCollection of the receiver and the receiver's 
	superclasses. The first element is the receiver, 
	followed by its superclass; the last element is Object."

	| temp |
	temp := self allSuperclasses.
	temp addFirst: self.
	^ temp! !

Trait named: #TTransformationCompatibility
	uses: {}
	category: 'Traits-Kernel-Traits'!

!TTransformationCompatibility commentStamp: 'apb 8/22/2005 15:24' prior: 0!
methods that make the user of this trait protocol-compatible with TraitTransformation.
When used in Trait and ClassTrait, this eliminates the need for TraitHolders.
!

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:35'!
aliasesForSelector: aSelector 
	^ OrderedCollection new
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:36'!
allAliasesDict
	^IdentityDictionary new
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:36'!
changedSelectorsComparedTo: aTraitTransformation
	| selectors otherSelectors changedSelectors aliases otherAliases |
	selectors := self allSelectors asIdentitySet.
	otherSelectors := aTraitTransformation allSelectors asIdentitySet.
	changedSelectors := IdentitySet withAll: (
		(selectors difference: otherSelectors) union: (otherSelectors difference: selectors)).
	aliases := self allAliasesDict.
	otherAliases := aTraitTransformation allAliasesDict.
	aliases keysAndValuesDo: [:key :value |
		(value ~~ (otherAliases at: key ifAbsent: [nil])) ifTrue: [changedSelectors add: key]].
	otherAliases keysAndValuesDo: [:key :value |
		(value ~~ (aliases at: key ifAbsent: [nil])) ifTrue: [changedSelectors add: key]].
	^ changedSelectors.
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:36'!
collectMethodsFor: aSelector into: methodDescription
	(self includesSelector: aSelector) ifTrue: [ 
		methodDescription addLocatedMethod: (
			LocatedMethod
				location: self
				selector: aSelector)]
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'dvf 9/22/2005 18:15'!
subject
	"for compatibility with TraitTransformations"
	^ self
! !

!TTransformationCompatibility methodsFor: 'enquiries' stamp: 'apb 8/22/2005 15:36'!
trait
	"for compatibility with TraitTransformations"
	^ self
! !

TestResource subclass: #TraitsResource
	instanceVariableNames: 'createdClassesAndTraits t1 t2 t3 t4 t5 t6 c1 c2 c3 c4 c5 c6 c7 c8 dirty'
	classVariableNames: 'SetUpCount'
	poolDictionaries: ''
	category: 'Traits-Tests'!

!TraitsResource class methodsFor: 'as yet unclassified' stamp: 'dvf 8/30/2005 16:57'!
resetIfDirty
	self current isDirty ifTrue: [self reset]! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c1
	^c1! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c1: anObject
	^c1 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c2
	^c2! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c2: anObject
	^c2 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c3
	^c3! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c3: anObject
	^c3 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c4
	^c4! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c4: anObject
	^c4 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c5
	^c5! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c5: anObject
	^c5 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c6
	^c6! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c6: anObject
	^c6 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c7
	^c7! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:00'!
c7: anObject
	^c7 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:01'!
c8
	^c8! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 14:01'!
c8: anObject
	^c8 := anObject! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'dvf 8/30/2005 11:47'!
categoryName
	^self class category! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'al 10/10/2005 10:37'!
codeChangedEvent: anEvent

	(anEvent isDoIt not
		and: [anEvent itemClass notNil]
		and: [self createdClassesAndTraits includes: anEvent itemClass instanceSide]) ifTrue: [self setDirty] ! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'dvf 8/30/2005 11:47'!
createClassNamed: aSymbol superclass: aClass uses: aTraitComposition
	| class |
	class := aClass
		subclass: aSymbol
		uses: aTraitComposition
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: ''		
		category: self categoryName.
	self createdClassesAndTraits add: class.
	^class! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'dvf 8/30/2005 11:47'!
createTraitNamed: aSymbol uses: aTraitComposition
	| trait |
	trait := Trait
		named: aSymbol
		uses: aTraitComposition
		category: self categoryName.
	self createdClassesAndTraits add: trait.
	^trait! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'dvf 8/30/2005 11:47'!
createdClassesAndTraits
	createdClassesAndTraits ifNil: [
		createdClassesAndTraits := OrderedCollection new].
	^createdClassesAndTraits! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 15:19'!
isDirty
	^dirty! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 15:19'!
setDirty
	dirty := true! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'al 9/16/2005 11:52'!
setUp
	"Please note, that most tests rely on this setup of traits and
	classes - and that especially the order of the definitions matters."
	"SetUpCount := SetUpCount + 1."

	dirty := false.
	SystemChangeNotifier uniqueInstance doSilently: 
			[self t1: (self createTraitNamed: #T1
						uses: { }).
			self t1 comment: 'I am the trait T1'.
			self t2: (self createTraitNamed: #T2
						uses: { }).
			self t2 compile: 'm21 ^21' classified: #cat1.
			self t2 compile: 'm22 ^22' classified: #cat2.
			self t2 classSide compile: 'm2ClassSide: a ^a'.
			self t3: (self createTraitNamed: #T3
						uses: { }).
			self t3 compile: 'm31 ^31' classified: #cat1.
			self t3 compile: 'm32 ^32' classified: #cat2.
			self t3 compile: 'm33 ^33' classified: #cat3.
			self t4: (self createTraitNamed: #T4
						uses: { (self t1). (self t2) }).
			self t4 compile: 'm11 ^41' classified: #catX.	"overrides T1>>m11"
			self t4 compile: 'm42 ^42' classified: #cat2.
			self t5: (self createTraitNamed: #T5 uses: self t1 + self t2).
			self t5 compile: 'm51 ^super foo' classified: #cat1.
			self t5 compile: 'm52 ^ self class bar' classified: #cat1.
			self t5 compile: 'm53 ^ self class bar' classified: #cat1.
			self t6: (self createTraitNamed: #T6
						uses: (self t1 + self t2) @ { (#m22Alias -> #m22) }).
			self c1: (self 
						createClassNamed: #C1
						superclass: Object
						uses: { }).
			self c1 compile: 'foo ^true' classified: #accessing.
			self t1 compile: 'm11 ^11' classified: #cat1.
			self t1 compile: 'm12 ^12' classified: #cat2.
			self t1 compile: 'm13 ^self m12' classified: #cat3.
			self c2: (self 
						createClassNamed: #C2
						superclass: self c1
						uses: self t5 - { #m11 }).
			self c2 compile: 'foo ^false' classified: #private.
			self c2 compile: 'bar ^self foo' classified: #private.
			self setUpTrivialRequiresFixture.
			self setUpTwoLevelRequiresFixture.
			self setUpTranslatingRequiresFixture].
	SystemChangeNotifier uniqueInstance notify: self ofAllSystemChangesUsing: #codeChangedEvent:! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'noha 6/11/2008 18:41'!
setUpTranslatingRequiresFixture
	self c6: (self 
				createClassNamed: #C6
				superclass: ProtoObject
				uses: { }).
	ProtoObject removeSubclass: self c6.
	self c6 superclass: nil.
	self c7: (self 
				createClassNamed: #C7
				superclass: self c6
				uses: { }).
	self c8: (self 
				createClassNamed: #C8
				superclass: self c7
				uses: { }).
	self c6 compile: 'foo ^self x' classified: #accessing.
	self c7 compile: 'foo ^3' classified: #accessing.
	self c7 compile: 'bar ^super foo' classified: #accessing.
	self c8 compile: 'bar ^self blah' classified: #accessing! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'noha 6/11/2008 18:42'!
setUpTrivialRequiresFixture
	self c3: (self 
				createClassNamed: #C3
				superclass: ProtoObject
				uses: { }).
	ProtoObject removeSubclass: self c3.
	self c3 superclass: nil.
	self c3 compile: 'foo ^self bla' classified: #accessing! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'noha 6/11/2008 18:42'!
setUpTwoLevelRequiresFixture
	self c4: (self 
				createClassNamed: #C4
				superclass: ProtoObject
				uses: { }).
	ProtoObject removeSubclass: self c4.
	self c4 superclass: nil.
	self c5: (self 
				createClassNamed: #C5
				superclass: self c4
				uses: { }).
	self c4 compile: 'foo ^self blew' classified: #accessing.
	self c5 compile: 'foo ^self blah' classified: #accessing! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:58'!
t1
	^t1! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:58'!
t1: anObject
	^t1 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t2
	^t2! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t2: anObject
	^t2 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t3
	^t3! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t3: anObject
	^t3 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t4
	^t4! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t4: anObject
	^t4 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t5
	^t5! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t5: anObject
	^t5 := anObject! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t6
	^t6! !

!TraitsResource methodsFor: 'accessing' stamp: 'dvf 8/30/2005 13:59'!
t6: anObject
	^t6 := anObject! !

!TraitsResource methodsFor: 'as yet unclassified' stamp: 'dvf 8/30/2005 16:01'!
tearDown
	| behaviorName |
	SystemChangeNotifier uniqueInstance noMoreNotificationsFor: self.
	self createdClassesAndTraits do: 
			[:aClassOrTrait | 
			behaviorName := aClassOrTrait name.
			Smalltalk at: behaviorName
				ifPresent: [:classOrTrait | classOrTrait removeFromSystem].
			ChangeSet current removeClassChanges: behaviorName].
	createdClassesAndTraits := self t1: (self 
						t2: (self t3: (self 
										t4: (self t5: (self 
														t6: (self c1: (self 
																		c2: (self c3: (self c4: (self c5: (self c6: (self c7: (self c8: nil)))))))))))))! !

InstructionStream subclass: #SendInfo
	instanceVariableNames: 'stack savedStacks selfSentSelectors superSentSelectors classSentSelectors isStartOfBlock numBlockArgs nr1 nr2 nr3 nr4 nr5'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

!SendInfo commentStamp: '<historical>' prior: 0!
SendInfo objects perform an abstract interpretation of a compiled method to ascertain the messages that are self-sent, super-sent and class-sent.

The idea is to simulate the execution of the instruction stream in the compiled method, keeping track of whether the values on the stack are self or some other value.  IN one place we have to keep track of the small integers that are on the stack,  becuase they determine how many elements are to be popped off of the stack.

Everything is fairly straighforward except for jumps.

Conditional forward jumps, as generated by ifTrue: and ifFalse:, are fairly easy.  At the site of the conditional jump, we copy the stack, and use one copy on each path.  When the paths eventually merge, the stacks should be the same size (if they aren't it's a compiler error!!).  We build a new stack that has #self on it every place where either of the old stacks had #self.  Thus, expressions like

	(aBoolean ifTrue: self ifFalse: other) foo: 3

that might send foo: 3 to self are recognized.

For unconditional jumps, we save the stack for use at the join point, and continue execution at the instruciton after the jump with an empty stack, which will be immediately overwritten by the stack that comes with the arriving execution.

The bottlenecks in this algorithm turned out to be detecting join points and simulating the stack.  Using an OrderedCollection for a stack caused a lot of redundant work, especially when emptying the stack.  Using a dictionary to detect join points turned out to be very slow, because of the possibility of having to scan through the hash table.

QuickIntegerDictionary and QuickStack provide the same core functionality, but in much more efficient forms.


Use SendInfo as follows:

				(SendInfo on: aCompiledMethod) collectSends

aSendInfo is both an InstructionStream and an InstructionStream client.

Structure:
 stack --	the simulated execution stack
 savedStacks -- The dictionary on which the extra stacks waiting to be merged in are saved.

  sentSelectors		-- an Identity Set accumulating the set of sent selectors.
  superSentSelectors	-- an Identity Set accumulating the set of super sent selectors.

  classSentSelectors -- an Identity Set accumulating the set of selectors sent to self class.
  isStartOfBlock -- a flag indicating that we have found the start of a block, and that the next jump will skip over it.
  numBlockArgs --  
nr1' 'nr2' 'nr3' 'nr4')
!

!SendInfo classSide methodsFor: 'accessing' stamp: 'NS 5/23/2005 10:53'!
neverRequiredSelectors
	| nrs |
	nrs := Array new: 5.
	nrs at: 1 put: CompiledMethod conflictMarker.
	nrs at: 2 put: CompiledMethod disabledMarker.
	nrs at: 3 put: CompiledMethod explicitRequirementMarker.
	nrs at: 4 put: CompiledMethod implicitRequirementMarker.
	nrs at: 5 put: CompiledMethod subclassResponsibilityMarker.
	^ nrs.
! !

!SendInfo methodsFor: 'private' stamp: 'NS 5/23/2005 11:01'!
addClassSentSelector: aSymbol
	classSentSelectors ifNil: [classSentSelectors := IdentitySet new].
	classSentSelectors add: aSymbol.! !

!SendInfo methodsFor: 'private' stamp: 'NS 5/23/2005 11:15'!
addSelfSentSelector: aSymbol
	selfSentSelectors ifNil: [selfSentSelectors := IdentitySet new].
	selfSentSelectors add: aSymbol.! !

!SendInfo methodsFor: 'private' stamp: 'NS 5/23/2005 11:15'!
addSuperSentSelector: aSymbol
	superSentSelectors ifNil: [superSentSelectors := IdentitySet new].
	superSentSelectors add: aSymbol.! !

!SendInfo methodsFor: 'private' stamp: 'apb 8/30/2003 10:36'!
assert: aBlock because: aMessage
	"Throw an assertion error if aBlock does not evaluates to true."

	aBlock value ifFalse: [AssertionFailure signal: aMessage]! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 8/29/2003 02:13'!
atMergePoint
	^ savedStacks includesKey: pc! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 8/28/2003 16:44'!
blockReturn
	"we could empty the stack, but what's the point?"! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/29/2003 13:40'!
blockReturnTop
	"Return from a block with Top Of Stack as result.
	The following instruction will be branched to from somewhere, and will
	therefore trigger a stackMerge, so it is important that the stack be emptied."

	self pop.
	self emptyStack.! !

!SendInfo methodsFor: 'accessing' stamp: 'NS 5/23/2005 11:04'!
classSentSelectors
	^  classSentSelectors ifNil: [#()] ifNotNil: [classSentSelectors].! !

!SendInfo methodsFor: 'initialization' stamp: 'apb 8/28/2003 17:32'!
collectSends
	| end |
	end := self method endPC.
	[pc <= end]
		whileTrue: [self interpretNextInstructionFor: self]! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/28/2003 19:28'!
doDup
	"Simulate the action of a 'duplicate top of stack' bytecode."

	self push: self top! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'NS 5/23/2005 11:14'!
doPop

	stack isEmpty ifFalse: [self pop]! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 9/7/2004 10:24'!
emptyStack
	stack becomeEmpty! !

!SendInfo methodsFor: 'accessing' stamp: 'apb 8/28/2003 16:52'!
home
	"Answer the context in which the receiver was defined."

	^ sender! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/31/2003 12:41'!
interpretNextInstructionFor: client 
	self atMergePoint
		ifTrue: [self mergeStacks].
	super interpretNextInstructionFor: client! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 9/7/2004 10:42'!
jump: distance 
	"Simulate the action of a 'unconditional jump' bytecode whose  
	offset is the argument, distance."
	distance < 0
		ifTrue: [^ self].
	distance = 0
		ifTrue: [self error: 'bad compiler!!'].
	savedStacks at: (self pc + distance) put: stack.
	"We empty the stack to signify that execution cannot 'fall through' to the
	next statement.  Note that since we just stored the current stack, not a copy, in
	the savedStacks dictionary, here we need to allocate a new stack."
	self newEmptyStack.  
	isStartOfBlock
		ifTrue: [isStartOfBlock := false.
			numBlockArgs	timesRepeat: [self push: #stuff]]! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/31/2003 21:50'!
jump: distance if: aBooleanConstant 
	"Simulate the action of a 'conditional jump' bytecode whose offset is 
	distance, and whose condition is aBooleanConstant."

	| destination |
	distance < 0 ifTrue:[^ self].
	distance = 0 ifTrue:[self error: 'bad compiler!!'].
	destination := self pc + distance.
	"remove the condition from the stack."
	self pop.
	savedStacks at: destination put: stack copy.
! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 8/31/2003 21:30'!
mergeStacks
	| otherStack |
	otherStack := savedStacks at: pc.
	savedStacks removeKey: pc.
	stack isEmpty ifTrue: [
		"This happens at the end of a block, or a short circuit conditional.  
		In these cases, it is not possible for execution to 'fall through' to 
		the merge point.  In other words, this is not a real merge point at all, 
		and we just continue execution with the saved stack."
		^ stack := otherStack ]. 
	"self assert: [stack size = otherStack size].  This assertion was true for every
	method in every subclass of Object, so I think that we can safely omit it!!"
	1 to: stack size
		do: [:i | ((stack at: i) ~~ #self
					and: [(otherStack at: i) == #self])
				ifTrue: [stack at: i put: #self]]! !

!SendInfo methodsFor: 'initialization' stamp: 'NS 2/15/2005 15:58'!
method: method pc: initialPC
	super method: method pc: initialPC.
	self prepareState.! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/31/2003 21:25'!
methodReturnConstant: aConstant
	"Simulate the action of a 'return receiver' bytecode. This corresponds to 
	the source expression '^aConstant'."

	self push: aConstant.
	self emptyStack! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/31/2003 21:25'!
methodReturnReceiver
	"Simulate the action of a 'return receiver' bytecode. This corresponds to 
	the source expression '^self'."

	self push: #self.	
	self emptyStack ! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/31/2003 21:25'!
methodReturnTop
	"Simulate the action of a 'return receiver' bytecode. This corresponds to 
	the source expression '^ <result of the last evaluation>'."

	self emptyStack ! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 9/7/2004 10:41'!
newEmptyStack
	stack := QuickStack new! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 8/28/2003 17:41'!
pop
	^ stack removeLast! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 8/31/2003 21:43'!
pop: n 
	stack removeLast: n! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/29/2003 15:37'!
popIntoLiteralVariable: value 
	"Simulate the action of bytecode that removes the top of the stack and 
	stores it into a literal variable of my method."

	self pop! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/28/2003 14:51'!
popIntoReceiverVariable: offset 
	"Simulate the action of bytecode that removes the top of the stack and 
	stores it into an instance variable of my receiver."

	self pop! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/28/2003 14:51'!
popIntoTemporaryVariable: offset 
	"Simulate the action of bytecode that removes the top of the stack and 
	stores it into an instance variable of my receiver."

	self pop! !

!SendInfo methodsFor: 'initialization' stamp: 'NS 5/23/2005 11:12'!
prepareState
	| nrsArray |
	self newEmptyStack.
	savedStacks := QuickIntegerDictionary new: (sender endPC).
	isStartOfBlock := false.
	nrsArray := self class neverRequiredSelectors.
	self assert:[nrsArray size = 5] because: 'Size of neverRequiredSelectors has been changed; re-optimize (by hand) #tallySelfSendsFor:'.
	nr1 := nrsArray at: 1.
	nr2 := nrsArray at: 2.
	nr3 := nrsArray at: 3.
	nr4 := nrsArray at: 4.
	nr5 := nrsArray at: 5.! !

!SendInfo methodsFor: 'printing' stamp: 'NS 5/23/2005 10:57'!
printOn: aStream
	aStream nextPut: $[.
	aStream print: self selfSentSelectors asArray.
	aStream space.
	aStream print: self superSentSelectors asArray.
	aStream space.
	aStream print: self  classSentSelectors asArray.
	aStream nextPut: $].! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 8/30/2003 11:38'!
push: aValue
	stack addLast: aValue.! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/29/2003 15:53'!
pushActiveContext
	"Simulate the action of bytecode that pushes the active context on the 
	top of its own stack."

	self push: #block.! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/29/2003 01:57'!
pushConstant: value 

	self push: value! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/28/2003 14:49'!
pushLiteralVariable: value 

	self push: #stuff! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/28/2003 17:55'!
pushReceiver
	self push: #self.! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/29/2003 07:50'!
pushReceiverVariable: anOffset
	"Push the value of one of the receiver's instance variables."

	self push: #stuff.! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/28/2003 14:47'!
pushTemporaryVariable: offset 
	"Simulate the action of bytecode that pushes the contents of the 
	temporary variable whose index is the argument, index, on the top of 
	the stack."

	self push: #stuff! !

!SendInfo methodsFor: 'accessing' stamp: 'NS 5/23/2005 11:03'!
selfSentSelectors
	^ selfSentSelectors ifNil: [#()] ifNotNil: [selfSentSelectors].! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'NS 5/23/2005 11:01'!
send: selector super: superFlag numArgs: numArgs 
	"Simulate the action of bytecodes that send a message with  
	selector. superFlag, tells whether the receiver of the  
	message was 'super' in the source. The arguments  
	of the message are found in the top numArgs locations on the  
	stack and the receiver just below them."
	| stackTop |
	selector == #blockCopy:
		ifTrue: ["self assert: [numArgs = 1]."
			isStartOfBlock := true.
			numBlockArgs := self pop.
			^ self].
	self pop: numArgs.
	stackTop := self pop.
	superFlag
		ifTrue: [self addSuperSentSelector: selector]
		ifFalse: [stackTop == #self
				ifTrue: [self tallySelfSendsFor: selector].
			stackTop == #class
				ifTrue: [self addClassSentSelector: selector]].
	self
		push: ((selector == #class and: [stackTop == #self])
				ifTrue: [#class]
				ifFalse: [#stuff])! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/29/2003 15:39'!
storeIntoLiteralVariable: value 
	"Simulate the action of bytecode that stores the top of the stack into a 
	literal variable of my method."
! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/29/2003 15:34'!
storeIntoReceiverVariable: offset 
	"Simulate the action of bytecode that stores the top of the stack into an 
	instance variable of my receiver."
! !

!SendInfo methodsFor: 'instruction decoding' stamp: 'apb 8/28/2003 19:48'!
storeIntoTemporaryVariable: offset 
	"Simulate the action of bytecode that stores the top of the stack into one 
	of my temporary variables."
! !

!SendInfo methodsFor: 'accessing' stamp: 'NS 5/23/2005 11:04'!
superSentSelectors
	^  superSentSelectors ifNil: [#()] ifNotNil: [superSentSelectors].! !

!SendInfo methodsFor: 'private' stamp: 'NS 5/23/2005 10:59'!
tallySelfSendsFor: selector 
	"Logically, we do the following test: 
		(self neverRequiredSelectors includes: selector) ifTrue: [^ self].
	However, since this test alone was reponsible for 2.8% of the execution time,
	we replace it with the following:"
	selector == nr1 ifTrue:[^ self].
	selector == nr2 ifTrue:[^ self].
	selector == nr3 ifTrue:[^ self].
	selector == nr4 ifTrue:[^ self].
	selector == nr5 ifTrue:[^ self].
	selector == #class ifTrue:[^ self].
	self addSelfSentSelector: selector.! !

!SendInfo methodsFor: 'stack manipulation' stamp: 'apb 8/28/2003 19:28'!
top
	^ stack last! !

!Behavior methodsFor: '*traits-requires' stamp: 'NS 5/26/2005 12:11'!
classAndMethodFor: aSymbol do: binaryBlock ifAbsent: absentBlock
	"Looks up the selector aSymbol in the class chain. If it is found, binaryBlock is evaluated
	with the class that defines the selector and the associated method. Otherwise
	absentBlock is evaluated."

	| method |
	self withAllSuperclassesDo: [:class |
		method := class compiledMethodAt: aSymbol ifAbsent: [nil].
		method ifNotNil: [^ binaryBlock value: class value: method].
	].
	^ absentBlock value.! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 9/9/2005 19:45'!
classesComposedWithMe
	^{self}! !

!Behavior methodsFor: '*traits-requires' stamp: 'NS 5/26/2005 14:27'!
computeSelfSendersFromInheritedSelfSenders: inheritedCollection localSelfSenders: localCollection
	"Compute the set of all self-senders from the set of inherited self-senders and the set of local self-senders."

	| result mDict |
	mDict := self methodDict.
	result := IdentitySet new: inheritedCollection size + localCollection size.
	"This if-statement is just a performance optimization. 
	Both branches are semantically equivalent."
	inheritedCollection size > mDict size ifTrue: [
		result addAll: inheritedCollection.
		mDict keysDo: [:each | result remove: each ifAbsent: []].
	] ifFalse: [
		inheritedCollection do: [:each | (mDict includesKey: each) ifFalse: [result add: each]].
	].
	result addAll: localCollection.
	^ result.! !

!Behavior methodsFor: '*traits-requires' stamp: 'NS 5/26/2005 14:11'!
computeTranslationsAndUpdateUnreachableSet: unreachableCollection
	"This method computes the set of unreachable selectors in the superclass by altering the set of unreachable selectors in this class. In addition, it builds a dictionary mapping super-sent selectors to the selectors of methods sending these selectors."

	| translations reachableSenders oldUnreachable |
	oldUnreachable := unreachableCollection copy.
	translations := IdentityDictionary new.
	"Add selectors implemented in this class to unreachable set."
	self methodDict keysDo: [:s | unreachableCollection add: s].
	
	"Fill translation dictionary and remove super-reachable selectors from unreachable."
	self sendCaches superSentSelectorsAndSendersDo: [:sent :senders |
		reachableSenders := FixedIdentitySet readonlyWithAll: senders notIn: oldUnreachable.
		reachableSenders isEmpty ifFalse: [
			translations at: sent put: reachableSenders.
			unreachableCollection remove: sent ifAbsent: [].
		].
	].
	^ translations! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 8/4/2005 16:48'!
findSelfSendersOf: selector unreachable: unreachableCollection noInheritedSelfSenders: noInheritedBoolean
	"This method answers a subset of all the reachable methods (local or inherited) that self-send selector (empty set => no self-senders).
	See Nathanael Schrli's PhD for more details."
	
	| selfSenders reachableSelfSenders translations |
	"Check whether there are local methods that self-send selector and are reachable."
	selfSenders := self sendCaches selfSendersOf: selector.
	reachableSelfSenders := FixedIdentitySet readonlyWithAll: selfSenders notIn: unreachableCollection.
	(self superclass isNil or: [noInheritedBoolean or: [reachableSelfSenders notEmpty]]) 
		ifTrue: [^ reachableSelfSenders].

	"Compute the set of unreachable superclass methods and super-send translations and recurse."
	translations := self computeTranslationsAndUpdateUnreachableSet: unreachableCollection.
	reachableSelfSenders := superclass findSelfSendersOf: selector unreachable: unreachableCollection noInheritedSelfSenders: false.
	
	"Use the translations to replace selectors that are super-sent with the methods that issue the super-sends."
	reachableSelfSenders := self translateReachableSelfSenders: reachableSelfSenders translations: translations.
	^ reachableSelfSenders.! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 9/12/2005 11:44'!
requiredSelectors
	^RequiredSelectors current for: self! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 9/6/2005 13:14'!
requiredSelectorsCache
	^RequiredSelectors current cacheFor: self! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 9/1/2005 16:36'!
sendCaches
	^LocalSends current for: self! !

!Behavior methodsFor: '*traits-requires' stamp: 'NS 5/24/2005 16:38'!
translateReachableSelfSenders: senderCollection translations: translationDictionary
	| result superSenders |
	(translationDictionary isEmptyOrNil or: [senderCollection isEmpty]) ifTrue: [^ senderCollection].
	result := FixedIdentitySet new.
	senderCollection do: [:s |
		superSenders := translationDictionary at: s ifAbsent: [nil].
		superSenders isNil
			ifTrue: [result add: s]
			ifFalse: [result addAll: superSenders].
		result isFull ifTrue: [^ result].
	].
	^ result.! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 9/12/2005 18:28'!
updateRequiredStatusFor: selector inSubclasses: someClasses
	"Updates the requirements cache to reflect whether selector is required in this class and some of its subclasses."
	| inheritedMethod |
	inheritedMethod := self superclass ifNotNil: [self superclass lookupSelector: selector].
	^self updateRequiredStatusFor: selector  inSubclasses: someClasses parentSelfSenders: FixedIdentitySet new providedInParent: inheritedMethod noInheritedSelfSenders: false accumulatingInto: IdentitySet new.! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 8/9/2005 17:02'!
updateRequiredStatusFor: selector  inSubclasses: someClasses parentSelfSenders: inheritedSelfSenders providedInParent: providedBoolean noInheritedSelfSenders: noInheritedBoolean
	"Updates the requirements cache to reflect whether selector is required in this class and all of its subclasses. The parameter inheritedSelfSenders is a subset of the methods in the parent of this class that are known to self-send selector. providedBoolean indicates whether selector is provided in the parent. noInheritedBoolean is true if no self-senders could be found in the superclass.
	See Nathanael Schrli's PhD for more details."

	| selfSenders provided m |
	"Remove from the inherited selfSenders methods that are potentially unreachable."
	selfSenders := inheritedSelfSenders reject: [:each | self includesSelector: each].
	
	"Check whether the method is provided."
	m := self compiledMethodAt: selector ifAbsent: [nil].
	providedBoolean ifTrue: [
		provided := m isNil or: [m isDisabled not and: [m isExplicitlyRequired not and: [m isSubclassResponsibility not]]].
	] ifFalse: [
		provided := m notNil and: [m isProvided].
	].

	provided ifTrue: [
		"If it is provided, it cannot be required."
		self setRequiredStatusOf: selector to: false.
	] ifFalse: [
		"If there are non-overridden inherited selfSenders we know that it must be
		required. Otherwise, we search for self-senders."
		selfSenders isEmpty ifTrue: [selfSenders := self findSelfSendersOf: selector unreachable: IdentitySet new noInheritedSelfSenders: noInheritedBoolean].
		self setRequiredStatusOf: selector to: selfSenders notEmpty.
	].

	"Do the same for all subclasses."
	self subclassesDo: [:each |
		 (someClasses includes: each) ifTrue: 
			[each updateRequiredStatusFor: selector  
				inSubclasses: someClasses 
				parentSelfSenders: selfSenders 
				providedInParent: provided 
				noInheritedSelfSenders: (provided not and: [selfSenders isEmpty])]].! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 9/12/2005 19:43'!
updateRequiredStatusFor: selector inSubclasses: someClasses parentSelfSenders: inheritedSelfSenders providedInParent: inheritedMethod noInheritedSelfSenders: noInheritedBoolean accumulatingInto: requiringClasses 
	"Updates the requirements cache to reflect whether selector is required in this class and all of its subclasses. The parameter inheritedSelfSenders is a subset of the methods in the parent of this class that are known to self-send selector. providedBoolean indicates whether selector is provided in the parent. noInheritedBoolean is true if no self-senders could be found in the superclass.
	See Nathanael Schrli's PhD for more details."

	"Remove from the inherited selfSenders methods that are potentially unreachable."

	| selfSenders m relevantMethod required lookedForInheritedSelfSenders |
	lookedForInheritedSelfSenders := false.
	selfSenders := inheritedSelfSenders 
				reject: [:each | self includesSelector: each].

	"Check whether the method is provided."
	m := self compiledMethodAt: selector ifAbsent: [nil].
	relevantMethod := m ifNotNil: [m] ifNil: [inheritedMethod].
	relevantMethod 
		ifNotNil: [required := relevantMethod isSubclassResponsibility or: [
					relevantMethod isDisabled or: [
						relevantMethod isExplicitlyRequired]]]
		ifNil: ["If there are non-overridden inherited selfSenders we know that it must be
		required. Otherwise, we search for self-senders."

			selfSenders isEmpty 
				ifTrue: 
					[selfSenders := self 
								findSelfSendersOf: selector
								unreachable: IdentitySet new
								noInheritedSelfSenders: noInheritedBoolean.
					lookedForInheritedSelfSenders := true].
			required := selfSenders notEmpty].

	required ifTrue: [requiringClasses add: self].

	"Do the same for all subclasses."
	self subclassesDo: 
			[:each | 
			(someClasses includes: each) 
				ifTrue: 
					[each 
						updateRequiredStatusFor: selector
						inSubclasses: someClasses
						parentSelfSenders: selfSenders
						providedInParent: relevantMethod
						noInheritedSelfSenders: (lookedForInheritedSelfSenders and: [selfSenders isEmpty])
						accumulatingInto: requiringClasses]].
	^requiringClasses! !

!Behavior methodsFor: '*traits-requires' stamp: 'dvf 8/9/2005 15:39'!
withInheritanceTraitCompositionIncludes: aTrait
	^self withAllSuperclasses anySatisfy: [:c | c traitCompositionIncludes: aTrait]! !

Error subclass: #TraitException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!TraitException commentStamp: '<historical>' prior: 0!
General exception used for example to signal invalid trait compositions!

TraitException subclass: #TraitCompositionException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitCompositionException commentStamp: '<historical>' prior: 0!
Signal invalid trait compositions.!

Array variableSubclass: #FixedIdentitySet
	instanceVariableNames: 'tally capacity'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Requires'!

!FixedIdentitySet commentStamp: 'NS 5/26/2005 13:00' prior: 0!
This is a fast but lazy implementation of fixed size identity sets. The two main difference to regular identity sets are:

1) These identity sets have a fixed size. If they are full, adding another element doesn't have any effect.
2) No rehashing. If two elements were to be stored on the same position in the underlying array, one of them is simply discarded.

As a consequence of (1) and (2), these identity sets are very fast!! Note that this class inherits form Array. This is not clean but reduces memory overhead when instances are created.!

!FixedIdentitySet classSide methodsFor: 'constants' stamp: 'NS 5/26/2005 13:02'!
arraySizeForCapacity: anInteger
	"Because of the hash performance, the array size is always a power of 2 
	and at least twice as big as the capacity anInteger"

	^ anInteger <= 0 
		ifTrue: [0]
		ifFalse: [1 << (anInteger << 1 - 1) highBit].! !

!FixedIdentitySet classSide methodsFor: 'constants' stamp: 'NS 5/26/2005 13:03'!
defaultSize
	^ 4! !

!FixedIdentitySet classSide methodsFor: 'constants' stamp: 'NS 5/23/2005 13:09'!
new
	^ self new: self defaultSize! !

!FixedIdentitySet classSide methodsFor: 'constants' stamp: 'NS 5/23/2005 17:40'!
new: anInteger
	^ (super new: (self arraySizeForCapacity: anInteger)) initializeCapacity: anInteger! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/26/2005 13:05'!
readonlyWithAll: aCollection notIn: notCollection
	"For performance reasons, this method may return an array rather than a FixedIdentitySet. 
	Therefore it should only be used if the return value does not need to be modified.
	Use #withAll:notIn: if the return value might need to be modified."

	| size |
	aCollection isEmpty ifTrue: [^ #()].
	size := aCollection size = 1 
		ifTrue: [1]
		ifFalse: [self sizeFor: aCollection].
	^ (self new: size) addAll: aCollection notIn: notCollection; yourself! !

!FixedIdentitySet classSide methodsFor: 'private' stamp: 'NS 5/24/2005 13:57'!
sizeFor: aCollection
	^ aCollection species == self 
		ifTrue: [aCollection capacity]
		ifFalse: [self defaultSize].! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/24/2005 08:51'!
with: anObject 
	"Answer an instance of me containing anObject."

	^ self new
		add: anObject;
		yourself! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/24/2005 08:51'!
with: firstObject with: secondObject 
	"Answer an instance of me containing the two arguments as elements."

	^ self new
		add: firstObject;
		add: secondObject;
		yourself! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/24/2005 08:51'!
with: firstObject with: secondObject with: thirdObject 
	"Answer an instance of me containing the three arguments as elements."

	^ self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		yourself! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/24/2005 08:51'!
with: firstObject with: secondObject with: thirdObject with: fourthObject 
	"Answer an instance of me, containing the four arguments as the elements."

	^ self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		add: fourthObject;
		yourself! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/24/2005 08:52'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject
	"Answer an instance of me, containing the five arguments as the elements."

	^ self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		add: fourthObject;
		add: fifthObject;
		yourself! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/24/2005 08:52'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject with: sixthObject
	"Answer an instance of me, containing the six arguments as the elements."

	^ self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		add: fourthObject;
		add: fifthObject;
		add: sixthObject;
		yourself! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/24/2005 13:57'!
withAll: aCollection
	"Create a new collection containing all the elements from aCollection."

	^ (self new: (self sizeFor: aCollection))
		addAll: aCollection;
		yourself! !

!FixedIdentitySet classSide methodsFor: 'instance creation' stamp: 'NS 5/24/2005 13:58'!
withAll: aCollection notIn: notCollection
	^ (self new: (self sizeFor: aCollection)) addAll: aCollection notIn: notCollection; yourself! !

!FixedIdentitySet methodsFor: 'comparing' stamp: 'NS 5/24/2005 08:59'!
= aCollection
	self == aCollection ifTrue: [^ true].
	aCollection size = self size ifFalse: [^ false].
	aCollection do: [:each | (self includes: each) ifFalse: [^ false]].
	^ true! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/24/2005 10:47'!
add: anObject
	| index old |
	self isFull ifTrue: [^ false].
	index := self indexOf: anObject.
	old := self basicAt: index.
	old == anObject ifTrue: [^ true].
	old ifNotNil: [^ false].
	self basicAt: index put: anObject.
	tally := tally + 1.
	^ true! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/24/2005 10:49'!
addAll: aCollection
	aCollection do: [:each | 
		self isFull ifTrue: [^ self].
		self add: each.
	].! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/24/2005 10:50'!
addAll: aCollection notIn: notCollection
	aCollection do: [:each | 
		self isFull ifTrue: [^ self].
		(notCollection includes: each) ifFalse: [self add: each].
	].! !

!FixedIdentitySet methodsFor: 'private' stamp: 'NS 5/23/2005 18:08'!
arraySize
	^ super size! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/24/2005 08:44'!
at: index
	self shouldNotImplement! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/24/2005 08:43'!
at: index put: anObject
	self shouldNotImplement! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/23/2005 17:40'!
capacity
	^ capacity! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/24/2005 10:46'!
destructiveAdd: anObject
	| index old |
	self isFull ifTrue: [^ false].
	index := self indexOf: anObject.
	old := self basicAt: index.
	self basicAt: index put: anObject.
	old ifNil: [tally := tally + 1].
	^ true! !

!FixedIdentitySet methodsFor: 'enumerating' stamp: 'NS 5/24/2005 09:04'!
do: aBlock
	| obj count |
	count := 0.
	1 to: self basicSize do: [:index |
		count >= tally ifTrue: [^ self].
		obj := self basicAt: index.
		obj ifNotNil: [count := count + 1. aBlock value: obj].
	].
! !

!FixedIdentitySet methodsFor: 'comparing' stamp: 'NS 5/24/2005 08:56'!
hash
	"Answer an integer hash value for the receiver such that,
	  -- the hash value of an unchanged object is constant over time, and
	  -- two equal objects have equal hash values"

	| hash |
	hash := self species hash.
	self size <= 10 ifTrue:
		[self do: [:elem | hash := hash bitXor: elem hash]].
	^hash bitXor: self size hash! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/24/2005 08:45'!
includes: anObject
	^ (self basicAt: (self indexOf: anObject)) == anObject! !

!FixedIdentitySet methodsFor: 'private' stamp: 'NS 5/24/2005 10:48'!
indexOf: anObject
	anObject isNil ifTrue: [self error: 'This class collection cannot handle nil as an element'].
	^ (anObject identityHash bitAnd: self basicSize - 1) + 1! !

!FixedIdentitySet methodsFor: 'initialization' stamp: 'NS 5/23/2005 17:39'!
initializeCapacity: anInteger
	tally := 0.
	capacity := anInteger.! !

!FixedIdentitySet methodsFor: 'testing' stamp: 'NS 5/24/2005 10:45'!
isFull
	^ tally >= capacity! !

!FixedIdentitySet methodsFor: 'testing' stamp: 'NS 5/24/2005 10:46'!
notFull
	^ tally < capacity! !

!FixedIdentitySet methodsFor: 'printing' stamp: 'NS 5/23/2005 18:32'!
printOn: aStream
	| count |
	aStream nextPutAll: '#('.
	count := 0.
	self do: [:each | 
		count := count + 1.
		each printOn: aStream.
		count < self size ifTrue: [aStream nextPut: $ ]
	].
	aStream nextPut: $).! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/24/2005 13:12'!
remove: anObject ifAbsent: aBlock
	| index |
	index := self indexOf: anObject.
	^ (self basicAt: index) == anObject 
		ifTrue: [self basicAt: index put: nil. tally := tally - 1. anObject]
		ifFalse: [aBlock value].! !

!FixedIdentitySet methodsFor: 'enumerating' stamp: 'NS 5/24/2005 13:52'!
select: aBlock
	| result |
	result := self species new: self capacity.
	self do: [:each | (aBlock value: each) ifTrue: [result add: each]].
	^ result.! !

!FixedIdentitySet methodsFor: 'accessing' stamp: 'NS 5/23/2005 13:13'!
size
	^ tally! !

Array variableSubclass: #QuickIntegerDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

!QuickIntegerDictionary commentStamp: 'dvf 8/4/2005 11:06' prior: 0!
This implementation serves as a very quick dictionary under the assumption that the keys are small natural numbers.!

!QuickIntegerDictionary methodsFor: 'accessing' stamp: 'apb 8/30/2003 18:33'!
includesKey: anIntegerKey
	^ (self at: anIntegerKey) notNil ! !

!QuickIntegerDictionary methodsFor: 'accessing' stamp: 'apb 8/30/2003 18:39'!
removeKey: anIntegerKey
	^ self at: anIntegerKey put: nil.! !

Array variableSubclass: #QuickStack
	instanceVariableNames: 'top'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

!QuickStack commentStamp: 'apb 8/30/2003 10:55' prior: 0!
This class is a quick and dirty implementation of a stack, designed to be used in the
SendsInfo abstract interpreter.  As opposed to using an OrderedCollection, this stack is quick because it can be emptied in a single assignment, and dirty because elements above the logical top of the stack (i.e., those that have been popped off) are not nil'ed out.  For our application, these are important optimizations with no ill effects.

QuickStacks will expand beyond their initial size if required, but we intend that the initial size will always be sufficient, so the efficiency of this feature is not important.
!

!QuickStack classSide methodsFor: 'instance creation' stamp: 'apb 8/30/2003 11:38'!
new
	^ (super new: 16) initialize
	"Why 16?  Because in performing an abstract interpretation of every
	method in every Class <= Object, the largest stack that was found 
	to be necessary was 15"! !

!QuickStack methodsFor: 'accessing' stamp: 'apb 8/30/2003 18:09'!
addLast: aValue
	top = self basicSize ifTrue: [self grow].
	top := top + 1.
	^ self at: top put: aValue! !

!QuickStack methodsFor: 'accessing' stamp: 'apb 9/7/2004 10:25'!
becomeEmpty
	top := 0! !

!QuickStack methodsFor: 'copying' stamp: 'apb 9/7/2004 10:37'!
copy
	"Answer a copy of a myself"
	| newSize |
	newSize := self basicSize.
	^ (self class new: newSize)
		replaceFrom: 1
		to: top
		with: self
		startingAt: 1;
		 setTop: top! !

!QuickStack methodsFor: 'private' stamp: 'apb 8/30/2003 17:56'!
grow
	| newStack |
	newStack := self class new: (self basicSize * 2).
	newStack replaceFrom: 1 to: top with: self startingAt: 1.
	newStack setTop: top.
	self becomeForward: newStack.
! !

!QuickStack methodsFor: 'initialization' stamp: 'apb 8/30/2003 11:34'!
initialize
	top := 0! !

!QuickStack methodsFor: 'accessing' stamp: 'apb 8/30/2003 11:36'!
isEmpty
	^ top = 0! !

!QuickStack methodsFor: 'accessing' stamp: 'apb 8/30/2003 11:43'!
removeLast
	| answer |
	answer := self at: top.
	top := top - 1.
	^ answer! !

!QuickStack methodsFor: 'accessing' stamp: 'apb 8/31/2003 21:42'!
removeLast: n

	top := top - n! !

!QuickStack methodsFor: 'private' stamp: 'apb 8/30/2003 11:55'!
setTop: t
	top := t! !

!QuickStack methodsFor: 'accessing' stamp: 'apb 8/30/2003 11:36'!
size
	^ top! !

TestCase subclass: #ATestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!ATestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/18/2006 21:28'!
testRequirement
  "
  self debug: #testRequirement
  "
  | class |
  class := Object
            subclass: #AClassForTest
            instanceVariableNames: ''
            classVariableNames: ''
            poolDictionaries: ''
            category: self class category.
  [
   class compile: 'call
                    ^ self isCalled'.
   self assert: (class requiredSelectors includes: #isCalled).


   class compile: 'isCalled
                    ^ 1'.
   "Fail here:"
   self deny: (class requiredSelectors includes: #isCalled)]

  ensure: [class removeFromSystem] ! !

TestCase subclass: #SendsInfoTest
	instanceVariableNames: 'state'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/2/2006 22:29'!
assert: aSelector sends: sendsCollection supersends: superCollection classSends: classCollection
	| theMethod info |
	theMethod := self class >> aSelector.
	info := (SendInfo on: theMethod) collectSends.
	self assert: #self sendsIn: info for: aSelector are: sendsCollection.
	self assert: #super sendsIn: info for: aSelector are: superCollection.
	self assert: #class sendsIn: info for: aSelector are: classCollection.
! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/2/2006 22:28'!
assert: levelSymbol sendsIn: aSelector are: anArrayOfSelectors 

! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/2/2006 22:39'!
assert: alevel sendsIn: aSendInfo for: aSelector are: aCollectionOfSelectors
	| detectedSends message |
	detectedSends := aSendInfo perform: (alevel, 'SentSelectors') asSymbol.
	message := alevel, ' sends wrong for ', aSelector.
	self assert: ((detectedSends isEmpty and: [aCollectionOfSelectors isEmpty]) or:
				[detectedSends = (aCollectionOfSelectors asIdentitySet)]) description: message! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/3/2006 01:03'!
branch
	"This method is never run. It is here just so that the sends in it can be
	tallied by the SendInfo interpreter."
	(state
		ifNil: [self]
		ifNotNil: [state]) clip.
	(state isNil
		ifTrue: [self]
		ifFalse: [state]) truncate.
! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/3/2006 00:36'!
classBranch
	self
		shouldnt: [state isNil
				ifTrue: [self tell]
				ifFalse: [self class tell]]
		raise: MessageNotUnderstood! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/2/2006 23:55'!
clip	
	"This method is never run. It is here just so that the sends in it can be
	tallied by the SendInfo interpreter."
	| temp |
	self printString.
	temp := self.
	temp error: 4 + 5! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/2/2006 23:55'!
clipRect: aRectangle 
	"This method is never run. It is here just so that the sends in it can be
	tallied by the SendInfo interpreter."
	super clipRect: aRectangle.
	(state notNil
			and: [self bitBlt notNil])
		ifTrue: [state bitBlt clipRect: aRectangle]! !

!SendsInfoTest methodsFor: 'tests' stamp: 'al 7/21/2006 21:45'!
expectedFailures
	"See #testCompileSuperBranch"

	^#(testCompileSuperBranch testTestCompileSuperBranch)! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/2/2006 23:55'!
pseudoCopy
	"This method is never run. It is here just so that the sends in it can be
	tallied by the SendInfo interpreter."
	| array |
	array := self class new: self basicSize.
	self
		instVarsWithIndexDo: [:each :i | array at: i put: each].
	^ array! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/3/2006 00:30'!
superBranch
	self
		should: [state isNil
				ifTrue: [super tell]
				ifFalse: [self tell]]
		raise: MessageNotUnderstood! !

!SendsInfoTest methodsFor: 'test subjects' stamp: 'apb 3/3/2006 00:14'!
tell
	"this method should not be defined in super"! !

!SendsInfoTest methodsFor: 'tests' stamp: 'apb 3/3/2006 00:10'!
testBranch
	self assert: #branch sends: #(clip truncate) supersends: #() classSends: #()! !

!SendsInfoTest methodsFor: 'tests' stamp: 'apb 3/3/2006 00:39'!
testClassBranch
	self assert: #classBranch sends: #(tell shouldnt:raise:) supersends: #() classSends: #(tell).
	self classBranch.! !

!SendsInfoTest methodsFor: 'tests' stamp: 'apb 3/2/2006 22:18'!
testClip
	self assert: #clip sends: #(printString) supersends: #() classSends: #()! !

!SendsInfoTest methodsFor: 'tests' stamp: 'apb 3/3/2006 00:03'!
testClipRect
	self assert: #clipRect:  sends: #(bitBlt)  supersends: #(clipRect:)  classSends: #() 
! !

!SendsInfoTest methodsFor: 'tests' stamp: 'apb 3/3/2006 00:38'!
testCompileSuperBranch
	"The compiler accepts this code 
		condition ifTrue: [super] ifFalse: [self]
	but compiles it as 
		condition ifTrue: [super] ifFalse: [self].
	(Use the show decompiled button).  So the SendsInfo interpreter, interpretting 
	the bad code, comes up with the wrong answer."

	self should: [(state isNil
		ifTrue: [super]
		ifFalse: [self]) tell] raise: MessageNotUnderstood description: 'Compiler treats super as self'
	! !

!SendsInfoTest methodsFor: 'tests' stamp: 'apb 3/2/2006 23:38'!
testPseudoCopy
	self assert: #pseudoCopy sends: #(instVarsWithIndexDo: basicSize) supersends: #() classSends: #(#new:)! !

!SendsInfoTest methodsFor: 'tests' stamp: 'apb 3/3/2006 00:33'!
testSuperBranch
	self assert: #superBranch sends: #(tell should:raise:) supersends: #(tell) classSends: #().
	self superBranch.! !

!SendsInfoTest methodsFor: 'tests' stamp: 'apb 3/3/2006 00:24'!
testTestCompileSuperBranch
	
	self assert: #testCompileSuperBranch sends: #(tell) supersends: #(tell) classSends: #()! !

TestCase subclass: #TimeMeasuringTest
	instanceVariableNames: 'realTime shouldProfile'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Requires'!

TimeMeasuringTest subclass: #RequiresSpeedTestCase
	instanceVariableNames: 'displayedClasses focusedClasses interestingCategories'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Requires'!

!RequiresSpeedTestCase commentStamp: 'al 2/17/2006 08:50' prior: 0!
This class sets some performance requirements for the requirements algorithm. Subclasses set up and test different caching strategies.

Test methods are prefixed with "performance" to exclude them from normal test runs.!

RequiresSpeedTestCase subclass: #FullMERequiresSpeedTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Requires'!

!FullMERequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/8/2005 11:37'!
getInformationFor: classes 
	classes do: 
			[:interestingCl | 
			interestingCl withAllSuperclassesDo: 
					[:cl | 
					LocalSends current for: cl.
					ProvidedSelectors current for: cl.
					RequiredSelectors current for: cl]]! !

!FullMERequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/9/2005 23:15'!
loseInterestInClasses: classes 
	classes do: [:interestingCl | 
		RequiredSelectors current lostInterest: self in: interestingCl.
		interestingCl withAllSuperclassesDo: [:cl | 
			ProvidedSelectors current lostInterest: self in: cl.
			LocalSends current lostInterest: self in: cl]]! !

!FullMERequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/9/2005 23:11'!
noteInterestInClasses: classes 
	classes do: 
			[:interestingCl | 
			interestingCl withAllSuperclassesDo: 
					[:cl | 
					LocalSends current noteInterestOf: self in: cl.
					ProvidedSelectors current noteInterestOf: self in: cl].
				RequiredSelectors current noteInterestOf: self in: interestingCl]! !

!FullMERequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/8/2005 11:37'!
prepareAllCaches
	| classes |
	classes := displayedClasses , focusedClasses.
	self noteInterestInClasses: classes.
	self getInformationFor: classes! !

!RequiresSpeedTestCase classSide methodsFor: 'as yet unclassified' stamp: 'dvf 9/1/2005 13:02'!
isAbstract
	^self == RequiresSpeedTestCase ! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/8/2005 11:32'!
classesInCategories: currentCats 
	^currentCats gather: 
					[:c | 
					(SystemOrganization listAtCategoryNamed: c) 
						collect: [:name | Smalltalk at: name]]! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/8/2005 11:32'!
decideInterestingClasses
	interestingCategories := {
				'Morphic-Basic'.
				'Morphic-Books'.
				'Morphic-Demo'.
				'System-Compression'.
				'System-Compiler'
			}.
	displayedClasses := self classesInCategories: interestingCategories.
	focusedClasses := {
				AtomMorph.
				AlignmentMorph.
				BookMorph.
				GZipReadStream.
				CommentNode
			}! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'al 2/17/2006 08:46'!
performanceTestFileInScenario
	self prepareAllCaches.
	"decide the interesting sets"
	"set them up as such"
	"decide the classes and methods to be touched"
	self measure: 
		["touch the code as decided"
		"ask isAbsract of many classes"
		"ask requiredSelectors of a few"].
	self assert: realTime < 1000
! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'al 2/17/2006 08:47'!
performanceTestFullRequires
	self prepareAllCaches.
	"note that we do not invalidate any caches"
	self measure: [AlignmentMorph requiredSelectors].
	"assuming we want 5 browsers to update their requiredSelectors list in 0.1 second"
	self assert: realTime < 20! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'al 2/17/2006 08:45'!
performanceTestMethodChangeScenario
	RequiredSelectors doWithTemporaryInstance: [
		LocalSends doWithTemporaryInstance: [
			ProvidedSelectors doWithTemporaryInstance: [
				self prepareAllCaches.
				self measure: 
						[self touchObjectHalt.
						displayedClasses do: [:cl | cl hasRequiredSelectors].
						focusedClasses do: [:cl | cl requiredSelectors]].
				self assert: realTime < 200]]]! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'al 2/17/2006 08:47'!
performanceTestMorphMethodChangeScenario
	RequiredSelectors doWithTemporaryInstance: 
			[LocalSends doWithTemporaryInstance: 
					[ProvidedSelectors doWithTemporaryInstance: 
							[self prepareAllCaches.
							self measure: 
									[self touchMorphStep.
									displayedClasses do: [:cl | cl hasRequiredSelectors].
									focusedClasses do: [:cl | cl requiredSelectors]].
							self assert: realTime < 200]]]! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'al 2/17/2006 08:47'!
performanceTestParseNodeMethodChangeScenario
	RequiredSelectors doWithTemporaryInstance: 
			[LocalSends doWithTemporaryInstance: 
					[ProvidedSelectors doWithTemporaryInstance: 
							[self prepareAllCaches.
							self measure: 
									[self touchParseNodeComment.
									displayedClasses do: [:cl | cl hasRequiredSelectors].
									focusedClasses do: [:cl | cl requiredSelectors]].
							self assert: realTime < 100]]]! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'al 2/17/2006 08:47'!
performanceTestSwitchToMorphClassCategoryScenario
	"When changing in one browser the selected category, we add some interesting classes, remove some others, and calculate some values. So this is a pretty full life cycle test."
	| noLongerInteresting newInteresting |
	RequiredSelectors doWithTemporaryInstance: 
			[LocalSends doWithTemporaryInstance: 
					[ProvidedSelectors doWithTemporaryInstance: 
							[self prepareAllCaches.
							noLongerInteresting := self classesInCategories: {'Morphic-Basic'}.
							newInteresting := self classesInCategories: {'Morphic-Kernel'}.
							self measure: 
									[self noteInterestInClasses: newInteresting.
									self loseInterestInClasses: noLongerInteresting.
									newInteresting do: [:cl | cl hasRequiredSelectors].
									self loseInterestInClasses: newInteresting.
									self noteInterestInClasses: noLongerInteresting.].
							self assert: realTime < 500]]]! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/1/2005 13:01'!
prepareAllCaches
	self subclassResponsibility.! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/1/2005 17:55'!
setUp
	self decideInterestingClasses.
! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 18:10'!
touchMorphStep
	Morph compile: (Morph sourceCodeAt: #step ifAbsent: []) asString! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 15:27'!
touchObjectHalt
	^Object compile: (Object sourceCodeAt: #halt ifAbsent: []) asString! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 18:12'!
touchParseNodeComment
	ParseNode compile: (ParseNode sourceCodeAt: #comment ifAbsent: []) asString! !

!RequiresSpeedTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 14:58'!
workingCopyPredicate
	^[:e | {'TraitsOmniBrowser'. 'Traits'} includes: e package name]! !

TimeMeasuringTest subclass: #SendCachePerformanceTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

!SendCachePerformanceTest methodsFor: 'as yet unclassified' stamp: 'al 2/17/2006 08:42'!
performanceTestBaseline
	LocalSends current for: Morph.
	self assert: [LocalSends current for: Morph] timeToRun < 1.
	Morph clearSendCaches.
	self measure: [LocalSends current for: Morph].
	self assert: realTime < 100.
	self assert: [LocalSends current for: Morph] timeToRun < 1! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'md 2/22/2006 14:27'!
debug
	self resources do: [:res | 
		res isAvailable ifFalse: [^res signalInitializationError]].
	[(self class selector: testSelector) setToDebug; runCase] 
		ensure: [self resources do: [:each | each reset]]
			! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 15:09'!
initialize
	shouldProfile := false.! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 16:07'!
measure: measuredBlock 
	shouldProfile 
		ifTrue: [TimeProfileBrowser onBlock: [10 timesRepeat: measuredBlock]].
	realTime := measuredBlock timeToRun! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 15:09'!
openDebuggerOnFailingTestMethod
	shouldProfile := true! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 14:53'!
reportPerformance
	| str |
	str := CrLfFileStream fileNamed: 'performanceReports.txt'.
	str setToEnd;
		nextPutAll: ' test: ', testSelector;
		nextPutAll: ' time: ', realTime asString; 
		nextPutAll: ' version: ', self versionInformation;
		cr; 
		close! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 14:58'!
runCase
	[super runCase] ensure: [self reportPerformance]! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 15:13'!
setToDebug
	shouldProfile := true
! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 14:59'!
versionInfoForWorkingCopiesThat: wcPredicate 
	^(MCWorkingCopy allManagers select: wcPredicate) inject: ''
		into: [:s :e | s , e description]! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 14:48'!
versionInformation
	| wcPredicate |
	wcPredicate := self workingCopyPredicate.
	^self versionInfoForWorkingCopiesThat: wcPredicate! !

!TimeMeasuringTest methodsFor: 'as yet unclassified' stamp: 'dvf 8/31/2005 14:48'!
workingCopyPredicate
	^[:e | '*Traits*' match: e package name]! !

TestCase subclass: #TraitsTestCase
	instanceVariableNames: 'createdClassesAndTraits'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

TraitsTestCase subclass: #ClassTraitTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!ClassTraitTest methodsFor: 'testing' stamp: 'al 3/26/2006 12:15'!
testChanges
	"Test the most important features to ensure that
	general functionality of class traits are working."

	"self run: #testChanges"

	| classTrait |
	classTrait := self t1 classTrait.
	classTrait compile: 'm1ClassSide ^17' classified: 'mycategory'.

	"local selectors"
	self assert: (classTrait includesLocalSelector: #m1ClassSide).
	self deny: (classTrait includesLocalSelector: #otherSelector).

	"propagation"
	self assert: (self t5 classSide methodDict includesKey: #m1ClassSide).
	self assert: (self c2 class methodDict includesKey: #m1ClassSide).
	self shouldnt: [self c2 m1ClassSide] raise: Error.
	self assert: self c2 m1ClassSide = 17.

	"category"
	self assert: (self c2 class organization categoryOfElement: #m1ClassSide) 
				= 'mycategory'.

	"conflicts"
	self t2 classSide compile: 'm1ClassSide' classified: 'mycategory'.
	self assert: (self c2 class methodDict includesKey: #m1ClassSide).
	self deny: (self c2 class includesLocalSelector: #m1ClassSide).
	self should: [self c2 m1ClassSide] raise: Error.

	"conflict category"
	self assert: (self c2 class organization categoryOfElement: #m1ClassSide) 
				= #mycategory! !

!ClassTraitTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testConflictsAliasesAndExclusions
	"conflict"

	self t1 classTrait compile: 'm2ClassSide: x ^99' classified: 'mycategory'.
	self assert: (self t1 classTrait includesLocalSelector: #m2ClassSide:).
	self assert: (self t5 classTrait >> #m2ClassSide:) isConflict.
	self assert: (self c2 class >> #m2ClassSide:) isConflict.

	"exclusion and alias"
	self assert: self t5 classSide traitComposition asString 
				= 'T1 classTrait + T2 classTrait'.
	self t5 classSide 
		setTraitCompositionFrom: (self t1 classTrait @ { (#m2ClassSideAlias1: -> #m2ClassSide:) } 
				+ self t2 classTrait) @ { (#m2ClassSideAlias2: -> #m2ClassSide:) } 
				- { #m2ClassSide: }.
	self deny: (self t5 classTrait >> #m2ClassSide:) isConflict.
	self deny: (self c2 class >> #m2ClassSide:) isConflict.
	self assert: (self c2 m2ClassSideAlias1: 13) = 99.
	self assert: (self c2 m2ClassSideAlias2: 13) = 13! !

!ClassTraitTest methodsFor: 'testing' stamp: 'dvf 8/30/2005 16:17'!
testInitialization
	"self run: #testInitialization"

	| classTrait |
	classTrait := self t1 classTrait.
	self assert: self t1 hasClassTrait.
	self assert: self t1 classTrait == classTrait.
	self assert: classTrait isClassTrait.
	self assert: classTrait classSide == classTrait.
	self deny: classTrait isBaseTrait.
	self assert: classTrait baseTrait == self t1.

	"assert classtrait methods are propagated to users when setting traitComposition"
	self assert: self t4 hasClassTrait.
	self assert: self t5 hasClassTrait.
	self assert: (self t2 classSide includesLocalSelector: #m2ClassSide:).
	self assert: (self t4 classSide methodDict includesKey: #m2ClassSide:).
	self assert: (self t5 classSide methodDict includesKey: #m2ClassSide:).
	self assert: (self c2 m2ClassSide: 17) = 17! !

!ClassTraitTest methodsFor: 'testing' stamp: 'al 3/26/2006 12:06'!
testUsers
	self assert: self t2 classSide users size = 3.
	self assert: (self t2 classSide users includesAllOf: {				
		(self t4 classTrait).
		(self t5 classTrait).
		(self t6 classTrait) }).
	self assert: self t5 classSide users size = 1.
	self assert: self t5 classSide users anyOne = self c2 class.
	self c2 setTraitCompositionFrom: self t1 + self t5.
	self assert: self t5 classSide users size = 1.
	self assert: self t5 classSide users anyOne = self c2 class.
	self c2 setTraitComposition: self t2 asTraitComposition.
	self assert: self t5 classSide users isEmpty! !

TraitsTestCase subclass: #LocatedMethodTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!LocatedMethodTest methodsFor: 'setup' stamp: 'dvf 8/26/2005 14:21'!
setUp
	super setUp.
	self t1 compile: '+ aNumber ^aNumber + 17'.
	self t1 compile: '!!aNumber
		| temp |
		^aNumber + 17'.
	self t1 compile: '&& anObject'.
	self t1 compile: '@%+anObject'.
	self t1 compile: 'mySelector "a comment"'.
	self t1 compile: 'mySelector: something
		^17'.
	self t1 compile: 'mySelector: something and:somethingElse ^true'! !

!LocatedMethodTest methodsFor: 'running' stamp: 'dvf 8/26/2005 14:25'!
testArgumentNames

	self assert: (LocatedMethod location: self t1 selector: #+) argumentNames = #(aNumber).
	self assert: (LocatedMethod location: self t1 selector: #!!) argumentNames = #(aNumber).
	self assert: (LocatedMethod location: self t1 selector: #&&) argumentNames = #(anObject).
	self assert: (LocatedMethod location: self t1 selector: #@%+) argumentNames = #(anObject).
	
	self assert: (LocatedMethod location: self t1 selector: #mySelector) argumentNames = #().
	self assert: (LocatedMethod location: self t1 selector: #mySelector:) argumentNames = #(something).
	self assert: (LocatedMethod location: self t1 selector: #mySelector:and:) argumentNames = #(something somethingElse).
	
! !

!LocatedMethodTest methodsFor: 'running' stamp: 'dvf 8/26/2005 14:25'!
testBinarySelectors
	self assert: (LocatedMethod location: self t1 selector: #+) isBinarySelector.
	self assert: (LocatedMethod location: self t1 selector: #!!) isBinarySelector.
	self assert: (LocatedMethod location: self t1 selector: #&&) isBinarySelector.
	self assert: (LocatedMethod location: self t1 selector: #@%+) isBinarySelector.
	
	self deny: (LocatedMethod location: self t1 selector: #mySelector) isBinarySelector.
	self deny: (LocatedMethod location: self t1 selector: #mySelector:) isBinarySelector.
	self deny: (LocatedMethod location: self t1 selector: #mySelector:and:) isBinarySelector.
	
	! !

!LocatedMethodTest methodsFor: 'running' stamp: 'al 1/23/2004 17:16'!
testEquality
	| locatedMethod1 locatedMethod2 |
	locatedMethod1 := LocatedMethod location: self class selector: #testEquality.
	locatedMethod2 := LocatedMethod location: self class selector: #testEquality.
	self assert: locatedMethod1 = locatedMethod2.
	self assert: locatedMethod1 hash = locatedMethod2 hash! !

TraitsTestCase subclass: #PureBehaviorTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'al 3/26/2006 12:24'!
testClassesWithTraits
	"self debug: #testClassesWithTraits"

	self assert: (self c1 methodDict includesKey: #foo).
	self assert: (self c2 methodDict includesKey: #bar).
	self assert: (self c2 methodDict includesKey: #m51).
	self assert: (self c2 methodDict includesKey: #m12).
	self assert: (self c2 methodDict includesKey: #m13).
	self assert: (self c2 methodDict includesKey: #m21).
	self assert: (self c2 methodDict includesKey: #m22).
	
	self deny: self c1 class hasTraitComposition.
	self assert: self c2 class hasTraitComposition.

	self assert: (self c2 class traitComposition size = 1).
	self assert: (self c2 class traitComposition includesTrait: self t5 classTrait)! !

!PureBehaviorTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:33'!
testIsAliasSelector
	self deny: (self t1 isAliasSelector: #m11).
	self deny: (self t1 isAliasSelector: #foo).

	"directly"
	self assert: (self t6 isAliasSelector: #m22Alias).
	self deny: (self t6 isAliasSelector: #m22).

	"indirectly"
	self c1 setTraitCompositionFrom: self t6.
	self assert: (self c1 isAliasSelector: #m22Alias).
	self deny: (self c1 isAliasSelector: #m22)! !

!PureBehaviorTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:33'!
testIsLocalAliasSelector
	self deny: (self t1 isLocalAliasSelector: #m11).
	self deny: (self t1 isLocalAliasSelector: #foo).

	"directly"
	self assert: (self t6 isLocalAliasSelector: #m22Alias).
	self deny: (self t6 isLocalAliasSelector: #m22).

	"indirectly"
	self c1 setTraitComposition: self t6 asTraitComposition.
	self deny: (self c1 isLocalAliasSelector: #m22Alias).
	self deny: (self c1 isLocalAliasSelector: #m22)! !

!PureBehaviorTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testLocalSelectors
	"self run: #testLocalSelectors"

	self assert: self t3 localSelectors size = 3.
	self assert: (self t3 localSelectors includesAllOf: #(#m31 #m32 #m33 )).
	self assert: (self t3 includesLocalSelector: #m32).
	self deny: (self t3 includesLocalSelector: #inexistantSelector).
	self assert: self t5 localSelectors size = 3.
	self assert: (self t5 localSelectors includes: #m51).
	self assert: (self t5 includesLocalSelector: #m51).
	self deny: (self t5 includesLocalSelector: #m11).
	self t5 removeSelector: #m51.
	self deny: (self t3 includesLocalSelector: #m51).
	self deny: (self t5 includesLocalSelector: #m11).
	self assert: self t5 localSelectors size = 2.
	self t5 compile: 'm52 ^self'.
	self assert: self t5 localSelectors size = 2.
	self assert: (self t5 localSelectors includes: #m52).

	"test that propagated methods do not get in as local methods"
	self t2 compile: 'local2 ^self'.
	self deny: (self t5 includesLocalSelector: #local2).
	self assert: self t5 localSelectors size = 2.
	self assert: (self t5 localSelectors includes: #m52).
	self assert: self c2 localSelectors size = 2.
	self assert: (self c2 localSelectors includesAllOf: #(#foo #bar ))! !

!PureBehaviorTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testMethodCategoryReorganization
	"self run: #testMethodCategory"

	self t1 compile: 'm1' classified: 'category1'.
	self assert: (self t5 organization categoryOfElement: #m1) = #category1.
	self assert: (self c2 organization categoryOfElement: #m1) = #category1.
	self t1 organization 
		classify: #m1
		under: #category2
		suppressIfDefault: true.
	self assert: (self t5 organization categoryOfElement: #m1) = #category2.
	self assert: (self c2 organization categoryOfElement: #m1) = #category2! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:21'!
testOwnMethodsTakePrecedenceOverTraitsMethods
	"First create a trait with no subtraits and then
	add subtrait t1 which implements m11 as well."

	| trait |
	trait := self createTraitNamed: #T
				uses: { }.
	trait compile: 'm11 ^999'.
	self assert: trait methodDict size = 1.
	self assert: (trait methodDict at: #m11) decompileString = 'm11
	^ 999'.
	Trait 
		named: #T
		uses: self t1
		category: self class category.
	self assert: trait methodDict size = 3.
	self assert: (trait methodDict keys includesAllOf: #(#m11 #m12 #m13 )).
	self assert: (trait methodDict at: #m11) decompileString = 'm11
	^ 999'.
	self assert: (trait methodDict at: #m12) decompileString = 'm12
	^ 12'! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:32'!
testPropagationOfChangesInTraits
	| aC2 |
	aC2 := self c2 new.
	self assert: self c2 methodDict size = 9.
	self t1 compile: 'zork ^false'.
	self assert: self c2 methodDict size = 10.
	self deny: aC2 zork.
	self t1 removeSelector: #m12.
	self assert: self c2 methodDict size = 9.
	self should: [aC2 m12] raise: MessageNotUnderstood.
	self assert: aC2 m21 = 21.
	self t2 compile: 'm21 ^99'.
	self assert: aC2 m21 = 99! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:33'!
testPropagationOfChangesInTraitsToAliasMethods
	| anObject |
	anObject := (self 
				createClassNamed: #AliasTestClass
				superclass: Object
				uses: self t6) new.
	self assert: anObject m22Alias = 22.

	"test update alias method"
	self t2 compile: 'm22 ^17'.
	self assert: anObject m22Alias = 17.

	"removing original method should also remove alias method"
	self t2 removeSelector: #m22.
	self should: [anObject m22Alias] raise: MessageNotUnderstood! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:21'!
testPropagationOfChangesInTraitsToAliasMethodsWhenOriginalMethodIsExcluded
	"Assert that alias method is updated although
	the original method is excluded from this user."

	| anObject |
	anObject := (self 
				createClassNamed: #AliasTestClass
				superclass: Object
				uses: self t1 @ { (#aliasM11 -> #m11) } - { #m11 }) new.
	self assert: anObject aliasM11 = 11.
	self deny: (anObject class methodDict includesKey: #m11).
	self t1 compile: 'm11 ^17'.
	self assert: anObject aliasM11 = 17! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:32'!
testPropagationWhenTraitCompositionModifications
	"Test that the propagation mechanism works when
	setting new traitCompositions."

	self assert: self c2 methodDict size = 9.	"2 + (3+(3+2))-1"

	"removing methods"
	Trait 
		named: #T5
		uses: self t1 + self t2 - { #m21. #m22 }
		category: self class category.
	self assert: self c2 methodDict size = 7.

	"adding methods"
	Trait 
		named: #T2
		uses: self t3
		category: self class category.
	self assert: self c2 methodDict size = 10.
	self assert: (self c2 methodDict keys includesAllOf: #(#m31 #m32 #m33 ))! !

!PureBehaviorTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testRemovingMethods
	"When removing a local method, assure that the method
	from the trait is installed instead and that the users are 
	updated."

	"self run: #testRemovingMethods"

	"Classes"

	self c2 compile: 'm12 ^0' classified: #xxx.
	self assert: (self c2 includesLocalSelector: #m12).
	self c2 removeSelector: #m12.
	self deny: (self c2 includesLocalSelector: #m12).
	self assert: (self c2 selectors includes: #m12).

	"Traits"
	self t5 compile: 'm12 ^0' classified: #xxx.
	self assert: self c2 new m12 = 0.
	self t5 removeSelector: #m12.
	self deny: (self t5 includesLocalSelector: #m12).
	self assert: (self t5 selectors includes: #m12).
	self assert: self c2 new m12 = 12! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:32'!
testSuperSends
	| aC2 |
	aC2 := self c2 new.
	self assert: aC2 m51.
	self deny: aC2 foo.
	self deny: aC2 bar! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:33'!
testTraitCompositionModifications
	self assert: self t6 methodDict size = 6.
	self assert: (self t6 sourceCodeAt: #m22Alias) asString = 'm22Alias ^22'.
	self t6 setTraitComposition: self t2 asTraitComposition.
	self assert: self t6 methodDict size = 2.
	self deny: (self t6 methodDict includesKey: #m22Alias).
	self t6 
		setTraitCompositionFrom: self t1 @ { (#m13Alias -> #m13) } - { #m11. #m12 } 
				+ self t2.
	self assert: self t6 methodDict size = 4.
	self 
		assert: (self t6 methodDict keys includesAllOf: #(#m13 #m13Alias #m21 #m22 )).
	self 
		assert: (self t6 sourceCodeAt: #m13Alias) asString = 'm13Alias ^self m12'! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:31'!
testTraitCompositionWithCycles
	self should: [self t1 setTraitComposition: self t1 asTraitComposition]
		raise: Error.
	self t2 setTraitComposition: self t3 asTraitComposition.
	self should: [self t3 setTraitComposition: self t5 asTraitComposition]
		raise: Error! !

!PureBehaviorTest methodsFor: 'testing-applying trait composition' stamp: 'dvf 8/26/2005 14:32'!
testUpdateWhenLocalMethodRemoved
	| aC2 |
	aC2 := self c2 new.
	self t5 compile: 'foo ^123'.
	self deny: aC2 foo.
	self c2 removeSelector: #foo.
	self assert: aC2 foo = 123! !

!PureBehaviorTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:33'!
traitOrClassOfSelector
	"self run: #traitOrClassOfSelector"

	"locally defined in trait or class"

	self assert: (self t1 traitOrClassOfSelector: #m12) = self t1.
	self assert: (self c1 traitOrClassOfSelector: #foo) = self c1.

	"not locally defined - simple"
	self assert: (self t4 traitOrClassOfSelector: #m21) = self t2.
	self assert: (self c2 traitOrClassOfSelector: #m51) = self t5.

	"not locally defined - into nested traits"
	self assert: (self c2 traitOrClassOfSelector: #m22) = self t2.

	"not locally defined - aliases"
	self assert: (self t6 traitOrClassOfSelector: #m22Alias) = self t2.

	"class side"
	self assert: (self t2 classSide traitOrClassOfSelector: #m2ClassSide:) 
				= self t2 classSide.
	self assert: (self t6 classSide traitOrClassOfSelector: #m2ClassSide:) 
				= self t2 classSide! !

TraitsTestCase subclass: #RequiresTestCase
	instanceVariableNames: 't7 t8 t9 t10 t11 c9 c10 c11 c12 t13 c13 ta tb tc ca cb cc cd ce cf ch ci cg'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

RequiresTestCase subclass: #RequiresOriginalTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!RequiresOriginalTestCase classSide methodsFor: 'as yet unclassified' stamp: 'al 1/13/2006 00:02'!
updateRequiredStatusFor: selector in: aClass 
	aClass updateRequiredStatusFor: selector inSubclasses: self systemNavigation allClassesAndTraits ! !

!RequiresTestCase classSide methodsFor: 'as yet unclassified' stamp: 'dvf 9/1/2005 16:47'!
isAbstract
	^self == RequiresTestCase! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/9/2005 15:09'!
for: ba classesComposedWith: aBehavior 
	^(ba includes: aBehavior) 
		or: [(ba gather: [:c | c traitComposition allTraits]) includes: aBehavior]! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 11:58'!
loseInterestsFor: behavior 
	RequiredSelectors current lostInterest: self in: behavior.
	LocalSends current lostInterest: self in: behavior.
	^ProvidedSelectors current lostInterest: self
		inAll: behavior withAllSuperclasses! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 14:16'!
loseInterestsInAll
	^self createdClassesAndTraits 
		, TraitsResource current createdClassesAndTraits 
			do: [:e | self loseInterestsFor: e]! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 11:58'!
noteInterestsFor: behavior 
	RequiredSelectors current noteInterestOf: self in: behavior.
	LocalSends current noteInterestOf: self in: behavior.
	ProvidedSelectors current noteInterestOf: self
		inAll: behavior withAllSuperclasses! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 12:32'!
noteInterestsForAll
	self createdClassesAndTraits 
		, TraitsResource current createdClassesAndTraits 
			do: [:e | self noteInterestsFor: e]! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/9/2005 17:27'!
requiredMethodsForTrait: aTrait 
	^aTrait requiredSelectors! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/9/2005 17:28'!
requiredMethodsOfTrait: basicTrait inContextOf: composedTrait 
	| interestingSelectors sss |
	interestingSelectors := (composedTrait traitComposition 
				transformationOfTrait: basicTrait) allSelectors.
	sss := composedTrait selfSentSelectorsFromSelectors: interestingSelectors.
	^sss copyWithoutAll: composedTrait allSelectors! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/8/2005 17:45'!
selfSentSelectorsInTrait: aTrait 
	^self selfSentSelectorsInTrait: aTrait fromSelectors: aTrait allSelectors 
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/9/2005 17:27'!
selfSentSelectorsInTrait: composedTrait fromSelectors: interestingSelectors 
	^composedTrait selfSentSelectorsFromSelectors: interestingSelectors 
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 12:32'!
setUp
	super setUp.
	t7 := self createTraitNamed: #T7
				uses: { }.
	t7 compile: 'm13 ^self m12' classified: #cat3.
	t8 := self createTraitNamed: #T8
				uses: { (t7 - { #m13 }) }.
	t9 := self createTraitNamed: #T9
				uses: { }.
	t9 compile: 'm13 ^self m12' classified: #cat3.
	t9 compile: 'm12 ^3' classified: #cat3.
	t10 := self createTraitNamed: #T10
				uses: { (t9 - { #m12 }) }.

	t11 := self createTraitNamed: #T11
				uses: { (t9 @ { (#m11 -> #m12) } - { #m12 }) }.

	c9 := self 
			createClassNamed: #C9
			superclass: ProtoObject
			uses: t7.

	
	c10 := self 
			createClassNamed: #C10
			superclass: ProtoObject
			uses: t7.
	c10 compile: 'm12 ^3'.

	c11 := self createClassNamed: #C11
			superclass: ProtoObject
			uses: {}.
	c11 compile: 'm12 ^3'.
	c12 := self createClassNamed: #C12
			superclass: c11
			uses: {t7}.
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/9/2005 16:34'!
setUpHierarchy
	ta := self createTraitNamed: #TA
				uses: { }.
	tb := self createTraitNamed: #TB
				uses: { }.
	tc := self createTraitNamed: #TC uses: tb.
	ca := self 
				createClassNamed: #CA
				superclass: ProtoObject
				uses: { }.
	cb := self 
				createClassNamed: #CB
				superclass: ca
				uses: ta + tb.
	cc := self 
				createClassNamed: #CC
				superclass: cb
				uses: tb.
	cd := self 
				createClassNamed: #CD
				superclass: cc
				uses: { }.
	ce := self 
				createClassNamed: #CE
				superclass: cc
				uses: { }.
	cf := self 
				createClassNamed: #CF
				superclass: cb
				uses: { }.
	cg := self 
				createClassNamed: #CG
				superclass: cf
				uses: { }.
	ch := self 
				createClassNamed: #CH
				superclass: ca
				uses: { ta }.
	ci := self 
				createClassNamed: #CI
				superclass: ch
				uses: { }.

	ca compile: 'mca ^self ssca'.
	cb compile: 'mca ^3'.
	cb compile: 'mcb super mca'.
	cc compile: 'mcb ^3'.
	cc compile: 'mcb ^self sscc'.
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/12/2005 16:41'!
testAffectedClassesAndTraits
	| rscc |
	self setUpHierarchy.
	rscc := RequiredSelectorsChangesCalculator onModificationOf: {tb} withTargets: {ta. cg. ci. cd. tc}.
	self assert: rscc rootClasses asSet = (Set withAll: {cc. cb}).
	self assert: rscc classesToUpdate asSet = (Set withAll: {cg. cd. cf. cc. cb}).
	self assert: rscc traitsToUpdate asSet = (Set withAll: {tc}).
	self assert: (#(sscc) copyWithoutAll: (rscc selectorsToUpdateIn: cc)) isEmpty.
	self assert: (#(ssca) copyWithoutAll: (rscc selectorsToUpdateIn: cb)) isEmpty.! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/8/2005 16:45'!
testExclusionWithAliasing
	self assert: ((self requiredMethodsForTrait: t11) = (Set with: #m12)).
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/8/2005 16:34'!
testExlcusionInTraits
	self assert: ((self requiredMethodsForTrait: t8) = (Set new)).
	self assert: ((self requiredMethodsForTrait: t9) = (Set new)).
	self assert: ((self requiredMethodsForTrait: t10) = (Set with:#m12)).
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 12:33'!
testOneLevelRequires
	
	[self noteInterestsForAll.
	self assert: self c3 localSelectors size = 1.
	self assert: (self c3 sendCaches selfSendersOf: #bla) = #(#foo ).
	self c3 requiredSelectors.
	self assert: self c3 requirements = (Set withAll: #(#bla ))] 
			ensure: [self loseInterestsInAll]! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/9/2005 11:02'!
testRequiresOfTraitInContextOfClass
	"a class providing nothing, leaves the requirements of the trait intact"
	self assert: (self requiredMethodsOfTrait: t7 inContextOf: c9) = (Set with: #m12).
	"a class can provide the Trait requirement"
	self assert: (self requiredMethodsOfTrait: t7 inContextOf: c10) = (Set new).
	"a class' superclass can provide the Trait requirement"
	self assert: (self requiredMethodsOfTrait: t7 inContextOf: c12) = (Set new).
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/8/2005 17:55'!
testSimpleCompositionContexts
	self assert: (self requiredMethodsOfTrait: t7 inContextOf: t8) = (Set new).
	self assert: (self requiredMethodsOfTrait: t9 inContextOf: t10) = (Set with: #m12).
	self assert: (self requiredMethodsOfTrait: t9 inContextOf: t11) = (Set with: #m12).! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'noha 6/11/2008 18:45'!
testSins
	| caa cab cac cad |
	caa := self 
				createClassNamed: #CAA
				superclass: ProtoObject
				uses: { }.
	ProtoObject removeSubclass: caa.
	caa superclass: nil.
	cab := self 
				createClassNamed: #CAB
				superclass: caa
				uses: {}.
	cac := self 
				createClassNamed: #CAC
				superclass: cab
				uses: {}.
	cad := self 
				createClassNamed: #CAD
				superclass: cac
				uses: { }.

	caa compile: 'ma self foo'.
	caa compile: 'md self explicitRequirement'.
	cac compile: 'mb self bar'.
	self noteInterestsFor: cad.
	self assert: (cad requiredSelectors = (Set withAll: #(foo bar md))).
	cab compile: 'mc ^3'.
	self assert: (cad requiredSelectors = (Set withAll: #(foo bar md))).
	self loseInterestsFor: cad.! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/8/2005 16:33'!
testStandAloneTrait
	self assert: ((self requiredMethodsForTrait: t7) = (Set with: #m12)).! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 14:31'!
testTwoLevelRequires
	[self noteInterestsForAll.
	self assert: self c4 localSelectors size = 1.
	self assert: self c5 localSelectors size = 1.
	self assert: (self c4 sendCaches selfSendersOf: #blew) = #(#foo ).
	self assert: (self c5 sendCaches selfSendersOf: #blah) = #(#foo ).
	self c4 requiredSelectors.
	self assert: self c4 requirements = (Set withAll: #(#blew )).
	self assert: self c5 requirements = (Set withAll: #(#blah ))]
		ensure: [self loseInterestsInAll ]! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 12:35'!
testTwoLevelRequiresWithUnalignedSuperSends
	[self noteInterestsForAll.
	self updateRequiredStatusFor: #x in: self c6.
	self updateRequiredStatusFor: #blah in: self c8.
	self assert: self c6 requirements = (Set with: #x).
	self assert: self c7 requirements = (Set with: #x).
	self assert: self c8 requirements = (Set with: #blah).]
		ensure: [self loseInterestsInAll]
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 12:35'!
testTwoLevelRequiresWithUnalignedSuperSendsStartLate
	[self noteInterestsForAll.
	self updateRequiredStatusFor: #x in: self c8.
	self updateRequiredStatusFor: #blah in: self c8.
	self assert: self c8 requirements = (Set with: #blah).
	self updateRequiredStatusFor: #x in: self c7.
	self updateRequiredStatusFor: #blah in: self c7.
	self assert: self c7 requirements = (Set with: #x)]
		ensure: [self loseInterestsInAll]
! !

!RequiresTestCase methodsFor: 'as yet unclassified' stamp: 'dvf 8/8/2005 14:46'!
updateRequiredStatusFor: selector in: aClass
	self class updateRequiredStatusFor: selector in: aClass
! !

TraitsTestCase subclass: #SystemTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!SystemTest methodsFor: 'testing' stamp: 'al 4/9/2006 15:22'!
testAllClassesAndTraits
	"self debug: #testAllClassesAndTraits"
	
	| trait |
	trait := self t1.
	self assert: (Smalltalk allClassesAndTraits includes: trait).
	self deny: (Smalltalk allClasses includes: trait).
	! !

!SystemTest methodsFor: 'testing' stamp: 'al 1/13/2006 12:27'!
testAllImplementedMessagesWithout
	"self debug: #testAllImplementedMessagesWithout"

	self t6 compile: 'das2qwdqwd'.
	self assert: (SystemNavigation default allImplementedMessages includes: #das2qwdqwd).
	self deny: (SystemNavigation default allImplementedMessages includes: #qwdqwdqwdc).! !

!SystemTest methodsFor: 'testing' stamp: 'al 1/13/2006 11:49'!
testAllSentMessages
	"self debug: #testAllSentMessages"

	self t1 compile: 'foo 1 dasoia'.
	self assert: (SystemNavigation default allSentMessages includes: 'dasoia' asSymbol).
	self deny: (SystemNavigation default allSentMessages includes: 'nioaosi' asSymbol).! !

!SystemTest methodsFor: 'testing' stamp: 'al 6/7/2008 14:54'!
testClassFromPattern
	"self debug: #testClassFromPattern"

	self assert: (Utilities classFromPattern: 'TComposingD' withCaption: '') = TComposingDescription! !

TraitsTestCase subclass: #TraitCompositionTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!TraitCompositionTest methodsFor: 'testing-basic' stamp: 'dvf 8/26/2005 14:31'!
testAliasCompositions
	"unary"

	self 
		shouldnt: [self t2 setTraitCompositionFrom: self t1 @ { (#aliasM11 -> #m11) }]
		raise: TraitCompositionException.
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#alias: -> #m11) }]
		raise: TraitCompositionException.
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#alias:x:y: -> #m11) }]
		raise: TraitCompositionException.

	"binary"
	self t1 compile: '= anObject'.
	self 
		shouldnt: [self t2 setTraitCompositionFrom: self t1 @ { (#equals: -> #=) }]
		raise: TraitCompositionException.
	self shouldnt: [self t2 setTraitCompositionFrom: self t1 @ { (#% -> #=) }]
		raise: TraitCompositionException.
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#equals -> #=) }]
		raise: TraitCompositionException.
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#equals:x: -> #=) }]
		raise: TraitCompositionException.

	"keyword"
	self t1 compile: 'x: a y: b z: c'.
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#'==' -> #x:y:z:) }]
		raise: TraitCompositionException.
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#x -> #x:y:z:) }]
		raise: TraitCompositionException.
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#x: -> #x:y:z:) }]
		raise: TraitCompositionException.
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#x:y: -> #x:y:z:) }]
		raise: TraitCompositionException.
	self shouldnt: 
			[self t2 setTraitCompositionFrom: self t1 @ { (#myX:y:z: -> #x:y:z:) }]
		raise: TraitCompositionException.

	"alias same as selector"
	self 
		should: [self t2 setTraitCompositionFrom: self t1 @ { (#m11 -> #m11) }]
		raise: TraitCompositionException.

	"same alias name used twice"
	self should: 
			[self t2 
				setTraitCompositionFrom: self t1 @ { (#alias -> #m11). (#alias -> #m12) }]
		raise: TraitCompositionException.

	"aliasing an alias"
	self should: 
			[self t2 
				setTraitCompositionFrom: self t1 @ { (#alias -> #m11). (#alias2 -> #alias) }]
		raise: TraitCompositionException! !

!TraitCompositionTest methodsFor: 'testing-enquiries' stamp: 'dvf 8/26/2005 14:31'!
testClassMethodsTakePrecedenceOverTraitsMethods
	| keys |
	keys := Set new.
	self t4 methodDict bindingsDo: [:each | keys add: each key].
	self assert: keys size = 6.
	self 
		assert: (keys includesAllOf: #(
						#m12
						#m13
						#m13
						#m21
						#m22
						#m11
						#m42
					)).
	self assert: (self t4 methodDict at: #m11) decompileString = 'm11
	^ 41'! !

!TraitCompositionTest methodsFor: 'testing-basic' stamp: 'dvf 8/30/2005 14:07'!
testCompositionFromArray
	| composition |
	composition := { (self t1) } asTraitComposition.
	self assert: (composition isKindOf: TraitComposition).
	self assert: (composition traits includes: self t1).
	self assert: composition traits size = 1.
	composition := { (self t1). self t2 } asTraitComposition.
	self assert: (composition isKindOf: TraitComposition).
	self assert: (composition traits includes: self t1).
	self assert: (composition traits includes: self t2).
	self assert: composition traits size = 2! !

!TraitCompositionTest methodsFor: 'testing-basic' stamp: 'al 12/15/2003 20:51'!
testEmptyTrait
	| composition |
	composition := {} asTraitComposition.
	
	self assert: (composition isKindOf: TraitComposition).
	self assert: composition transformations isEmpty.
	self assert: composition traits isEmpty! !

!TraitCompositionTest methodsFor: 'testing-basic' stamp: 'dvf 8/26/2005 14:31'!
testInvalidComposition
	self should: [self t1 @ { (#a -> #b) } @ { (#x -> #y) }]
		raise: TraitCompositionException.
	self should: [(self t1 + self t2) @ { (#a -> #b) } @ { (#x -> #y) }]
		raise: TraitCompositionException.
	self should: [self t1 - { #a } - { #b }] raise: TraitCompositionException.
	self should: [self t1 + self t2 - { #a } - { #b }]
		raise: TraitCompositionException.
	self should: [(self t1 - { #x }) @ { (#a -> #b) }]
		raise: TraitCompositionException.
	self should: [(self t1 + self t2 - { #x }) @ { (#a -> #b) }]
		raise: TraitCompositionException.
	self should: [self t1 + self t1] raise: TraitCompositionException.
	self should: [(self t1 + self t2) @ { (#a -> #b) } + self t1]
		raise: TraitCompositionException.
	self should: [self t1 @ { (#a -> #m11). (#a -> #m12) }]
		raise: TraitCompositionException.
	self should: [self t1 @ { (#a -> #m11). (#b -> #a) }]
		raise: TraitCompositionException! !

!TraitCompositionTest methodsFor: 'testing-basic' stamp: 'dvf 8/26/2005 14:31'!
testPrinting
	| composition1 composition2 |
	composition1 := ((self t1 - { #a } + self t2) @ { (#z -> #c) } - { #b. #c } 
				+ self t3 - { #d. #e } 
				+ self t4) @ { (#x -> #a). (#y -> #b) }.
	composition2 := self t4 @ { (#x -> #a). (#y -> #b) } + self t1 - { #a } 
				+ self t3 - { #d. #e } 
				+ self t2 - { #b. #c }.
	self assertPrints: composition1 printString
		like: 'T1 - {#a} + T2 @ {#z->#c} - {#b. #c} + T3 - {#d. #e} + T4 @ {#x->#a. #y->#b}'.
	self assertPrints: composition2 printString
		like: 'T4 @ {#x->#a. #y->#b} + T1 - {#a} + T3 - {#d. #e} + T2 - {#b. #c}'! !

!TraitCompositionTest methodsFor: 'testing-enquiries' stamp: 'dvf 8/26/2005 14:31'!
testProvidedMethodBindingsWithConflicts
	| traitWithConflict methodDict |
	traitWithConflict := self createTraitNamed: #TraitWithConflict
				uses: self t1 + self t4.
	methodDict := traitWithConflict methodDict.
	self assert: methodDict size = 6.
	self 
		assert: (methodDict keys includesAllOf: #(
						#m11
						#m12
						#m13
						#m21
						#m22
						#m42
					)).
	self 
		assert: (methodDict at: #m11) decompileString = 'm11
	self traitConflict'! !

!TraitCompositionTest methodsFor: 'testing-basic' stamp: 'dvf 8/26/2005 14:31'!
testSum
	| composition |
	composition := self t1 + self t2 + self t3.
	self assert: (composition isKindOf: TraitComposition).
	self assert: (composition traits includes: self t1).
	self assert: (composition traits includes: self t2).
	self assert: (composition traits includes: self t3).
	self assert: composition traits size = 3! !

!TraitCompositionTest methodsFor: 'testing-basic' stamp: 'dvf 8/26/2005 14:31'!
testSumWithParenthesis
	| composition |
	composition := self t1 + (self t2 + self t3).
	self assert: (composition isKindOf: TraitComposition).
	self assert: (composition traits includes: self t1).
	self assert: (composition traits includes: self t2).
	self assert: (composition traits includes: self t3).
	self assert: composition traits size = 3.
	self assert: composition size = 3! !

TraitsTestCase subclass: #TraitFileOutTest
	instanceVariableNames: 'ca cb ta tb tc td'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!TraitFileOutTest methodsFor: 'running' stamp: 'al 4/25/2004 16:29'!
categoryName
	^'Traits-Tests-FileOut'! !

!TraitFileOutTest methodsFor: 'running' stamp: 'al 4/25/2004 13:26'!
setUp
	super setUp.
	SystemOrganization addCategory: self categoryName.
	
	td := self createTraitNamed: #TD uses: {}.		
	td compile: 'd' classified: #cat1.
	tc := self createTraitNamed: #TC uses: td.		
	tc compile: 'c' classified: #cat1.
	tb := self createTraitNamed: #TB uses: td.		
	tb compile: 'b' classified: #cat1.
	ta := self createTraitNamed: #TA uses: tb + tc @ {#cc->#c} - {#c}.
	ta compile: 'a' classified: #cat1.
	
	ca := self createClassNamed: #CA superclass: Object uses: {}.
	ca compile: 'ca' classified: #cat1.
	cb := self createClassNamed: #CB superclass: ca uses: ta.
	cb compile: 'cb' classified: #cat1.
	
	"make the class of cb also use tc:"
	cb class uses: ta classTrait + tc instanceVariableNames: ''.! !

!TraitFileOutTest methodsFor: 'running' stamp: 'dvf 8/30/2005 15:07'!
tearDown
	| dir |
	dir := FileDirectory default.
	self createdClassesAndTraits, self resourceClassesAndTraits  do: [:each |
		dir deleteFileNamed: each asString , '.st' ifAbsent: []].
	dir deleteFileNamed: self categoryName , '.st' ifAbsent: [].
	SystemOrganization removeSystemCategory: self categoryName.
	super tearDown! !

!TraitFileOutTest methodsFor: 'testing' stamp: 'al 7/19/2004 20:50'!
testFileOutCategory
	"File out whole system category, delete all classes and traits and then
	file them in again."

	"self run: #testFileOutCategory"

	| file |
	SystemOrganization fileOutCategory: self categoryName.
	SystemOrganization removeSystemCategory: self categoryName.
	self deny: (Smalltalk keys includesAnyOf: #(CA CB TA TB TC TD)).
	[	file := FileStream readOnlyFileNamed: self categoryName , '.st'.
		file fileIn]
		ensure: [file close].

	self assert: (Smalltalk keys includesAllOf: #(CA CB TA TB TC TD)).

	ta := Smalltalk at: #TA.
	self assert: ta traitComposition asString = 'TB + TC @ {#cc->#c} - {#c}'.
	self assert: (ta methodDict keys includesAllOf: #(a b cc)).

	cb := Smalltalk at: #CB.
	self assert: cb traitComposition asString = 'TA'.
	self assert: (cb methodDict keys includesAllOf: #(cb a b cc)).

	"test classSide traitComposition of CB"

	self assert: cb classSide traitComposition asString =  'TA classTrait + TC'.
	self assert: (cb classSide methodDict keys includesAllOf: #(d c))
	! !

!TraitFileOutTest methodsFor: 'testing' stamp: 'al 12/1/2005 16:30'!
testFileOutTrait
	"fileOut trait T6, remove it from system and then file it in again"

	"self run: #testFileOutTrait"

	| fileName file |
	self t6 compile: 'localMethod: argument ^argument'.
	self t6 classSide compile: 'localClassSideMethod: argument ^argument'.
	self t6 fileOut.
	fileName := self t6 asString , '.st'.
	self resourceClassesAndTraits remove: self t6.
	self t6 removeFromSystem.
	
	[file := FileStream readOnlyFileNamed: fileName.
	file fileIn] 
			ensure: [file close].
	self assert: (Smalltalk includesKey: #T6).
	TraitsResource current t6: (Smalltalk at: #T6).
	self resourceClassesAndTraits add: self t6.
	self 
		assert: self t6 traitComposition asString = 'T1 + T2 @ {#m22Alias->#m22}'.
	self 
		assert: (self t6 methodDict keys includesAllOf: #(
						#localMethod:
						#m11
						#m12
						#m13
						#m21
						#m22
						#m22Alias
					)).
	self assert: self t6 classSide methodDict size = 2.
	self 
		assert: (self t6 classSide methodDict keys includesAllOf: #(#localClassSideMethod: #m2ClassSide: ))! !

!TraitFileOutTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testRemovingMethods
	"When removing a local method, assure that the method
	from the trait is installed instead and that the users are 
	updated."

	"self run: #testRemovingMethods"

	"Classes"

	self c2 compile: 'm12 ^0' classified: #xxx.
	self assert: (self c2 includesLocalSelector: #m12).
	self c2 removeSelector: #m12.
	self deny: (self c2 includesLocalSelector: #m12).
	self assert: (self c2 selectors includes: #m12).

	"Traits"
	self t5 compile: 'm12 ^0' classified: #xxx.
	self assert: self c2 new m12 = 0.
	self t5 removeSelector: #m12.
	self deny: (self t5 includesLocalSelector: #m12).
	self assert: (self t5 selectors includes: #m12).
	self assert: self c2 new m12 = 12! !

TraitsTestCase subclass: #TraitMethodDescriptionTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!TraitMethodDescriptionTest methodsFor: 'running' stamp: 'dvf 8/26/2005 14:31'!
testArgumentNames
	self t1 compile: 'zork1: myArgument zork2: mySecondArgument ^true'.
	self t2 compile: 'zork1: myArgument zork2: somethingElse ^false'.
	self assert: ((self t5 sourceCodeAt: #zork1:zork2:) asString 
				beginsWith: 'zork1: myArgument zork2: arg2').
	self t1 compile: 'zork1: myArgument zork2: mySecondArgument ^true'.
	self t2 compile: 'zork1: somethingElse zork2: myArgument ^false'.
	self assert: ((self t5 sourceCodeAt: #zork1:zork2:) asString 
				beginsWith: 'zork1: arg1 zork2: arg2')! !

!TraitMethodDescriptionTest methodsFor: 'running' stamp: 'dvf 8/26/2005 14:33'!
testCategories
	self assert: (self t4 organization categoryOfElement: #m21) = #cat1.
	self assert: (self t4 organization categoryOfElement: #m22) = #cat2.
	self assert: (self t4 organization categoryOfElement: #m11) = #catX.
	self assert: (self t4 organization categoryOfElement: #m12) = #cat2.
	self assert: (self t4 organization categoryOfElement: #m13) = #cat3.
	self assert: (self t6 organization categoryOfElement: #m22Alias) = #cat2.
	self t2 organization classify: #m22 under: #catX.
	self assert: (self t4 organization categoryOfElement: #m22) = #catX.
	self assert: (self t6 organization categoryOfElement: #m22Alias) = #catX.
	self t6 organization classify: #m22 under: #catY.
	self t6 organization classify: #m22Alias under: #catY.
	self t2 organization classify: #m22 under: #catZ.
	self assert: (self t6 organization categoryOfElement: #m22) = #catY.
	self assert: (self t6 organization categoryOfElement: #m22Alias) = #catY.
	self t1 compile: 'mA' classified: #catA.
	self assert: (self t4 organization categoryOfElement: #mA) = #catA.
	self t1 organization classify: #mA under: #cat1.
	self assert: (self t4 organization categories includes: #catA) not! !

!TraitMethodDescriptionTest methodsFor: 'running' stamp: 'dvf 8/26/2005 14:32'!
testConflictMethodCreation
	"Generate conflicting methods between t1 and t2
	and check the resulting method in Trait t5 (or c2).
	Also test selectors like foo:x (without space) or selectors with CRs."

	"unary"

	self t2 compile: 'm12 ^false'.
	self assert: ((self t5 sourceCodeAt: #m12) asString beginsWith: 'm12').
	self should: [self c2 new m12] raise: Error.

	"binary"
	self t1 compile: '@ myArgument ^true'.
	self t2 compile: '@myArgument ^false'.
	self 
		assert: ((self t5 sourceCodeAt: #@) asString beginsWith: '@ myArgument').
	self should: [self c2 new @ 17] raise: Error.

	"keyword"
	self t1 compile: 'zork: myArgument
		^true'.
	self t2 compile: 'zork: myArgument ^false'.
	self assert: ((self t5 sourceCodeAt: #zork:) asString 
				beginsWith: 'zork: myArgument').
	self should: [self c2 new zork: 17] raise: Error.
	self t1 compile: 'zork:myArgument ^true'.
	self t2 compile: 'zork:myArgument ^false'.
	self assert: ((self t5 sourceCodeAt: #zork:) asString 
				beginsWith: 'zork: myArgument').
	self should: [self c2 new zork: 17] raise: Error.
	self t1 compile: 'zork1: myArgument zork2: mySecondArgument ^true'.
	self t2 compile: 'zork1: anObject zork2: anotherObject ^false'.
	self assert: ((self t5 sourceCodeAt: #zork1:zork2:) asString 
				beginsWith: 'zork1: arg1 zork2: arg2').
	self should: [self c2 new zork1: 1 zork2: 2] raise: Error! !

!TraitMethodDescriptionTest methodsFor: 'running' stamp: 'dvf 8/26/2005 14:31'!
testConflictingCategories
	| t7 t8 |
	self t2 compile: 'm11' classified: #catY.
	self assert: (self t4 organization categoryOfElement: #m11) = #catX.
	self assert: (self t5 organization categoryOfElement: #m11) = #cat1.
	t7 := self createTraitNamed: #T7 uses: self t1 + self t2.
	self assert: (t7 organization categoryOfElement: #m11) 
				= ClassOrganizer ambiguous.
	self t1 removeSelector: #m11.
	self assert: (self t4 organization categoryOfElement: #m11) = #catX.
	self assert: (self t5 organization categoryOfElement: #m11) = #catY.
	self assert: (t7 organization categoryOfElement: #m11) = #catY.
	self 
		assert: (t7 organization categories includes: ClassOrganizer ambiguous) not.
	self t1 compile: 'm11' classified: #cat1.
	t8 := self createTraitNamed: #T8 uses: self t1 + self t2.
	t8 organization classify: #m11 under: #cat1.
	self t1 organization classify: #m11 under: #catZ.
	self assert: (self t4 organization categoryOfElement: #m11) = #catX.
	self assert: (self t5 organization categoryOfElement: #m11) = #catY.
	self assert: (t8 organization categoryOfElement: #m11) = #catZ! !

!TraitMethodDescriptionTest methodsFor: 'running' stamp: 'al 1/23/2004 12:57'!
testInitialize
	| empty |
	empty := TraitMethodDescription new.
	self assert: empty isEmpty.
	self deny: empty isConflict.
	self deny: empty isProvided.
	self deny: empty isRequired! !

TraitsTestCase subclass: #TraitTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Tests'!

!TraitTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testAddAndRemoveMethodsFromSubtraits
	| aC2 |
	aC2 := self c2 new.
	self assert: aC2 m51.
	self t5 removeSelector: #m51.
	self should: [aC2 m51] raise: MessageNotUnderstood.
	self t1 compile: 'foo ^true'.
	self deny: aC2 foo.
	self t1 compile: 'm51 ^self'.
	self shouldnt: [aC2 m51] raise: MessageNotUnderstood.
	self assert: aC2 m51 == aC2! !

!TraitTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testAddAndRemoveMethodsInClassOrTrait
	| aC2 |
	aC2 := self c2 new.
	self assert: aC2 m51.
	self c2 compile: 'm51 ^123'.
	self assert: aC2 m51 = 123.
	self c2 removeSelector: #m51.
	self shouldnt: [aC2 m51] raise: MessageNotUnderstood.
	self assert: aC2 m51.
	self t4 removeSelector: #m11.
	self assert: (self t4 methodDict includesKey: #m11)! !

!TraitTest methodsFor: 'testing' stamp: 'ms 5/8/2007 19:23'!
testAllClassVarNames
	
	
	self assert: self t1 allClassVarNames isEmpty! !

!TraitTest methodsFor: 'testing' stamp: 'dvf 8/30/2005 14:10'!
testCompositionCopy
	| t6compositionCopyFirst c2compositionCopy |
	self assert: (self t1 + self t2) allTraits 
				= (self t1 + self t2) copyTraitExpression allTraits.
	self assert: (self t1 classTrait + self t2 classTrait) allTraits 
				= (self t1 classTrait + self t2 classTrait) copyTraitExpression allTraits.
	self assert: self t6 traitComposition allTraits 
				= self t6 traitComposition copyTraitExpression allTraits.
	self 
		assert: self t6 asTraitComposition copyTraitExpression allTraits = { (self t1). (self t2). (self t6) }.
	"make no undue sharing happens of exclusions and aliases after an expression copy"
	t6compositionCopyFirst := self t6 traitComposition copyTraitExpression.
	t6compositionCopyFirst transformations at: 1 put: #m22Alias -> #m33.
	self 
		assert: self t6 traitComposition transformations second aliases first value 
				= #m22.
	c2compositionCopy := self c2 traitComposition copyTraitExpression.
	c2compositionCopy transformations first exclusions at: 1 put: #m4.
	self c2 traitComposition transformations first exclusions = #(#m11 )! !

!TraitTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testExplicitRequirement
	"self run: #testExplicitRequirement"

	self t1 compile: 'm self explicitRequirement'.
	self t2 compile: 'm ^true'.
	self assert: self t4 >> #m == (self t2 >> #m).
	self assert: self c2 new m.
	self t2 removeSelector: #m.
	self assert: self t5 >> #m == (self t1 >> #m).
	self should: [self c2 new m] raise: Error! !

!TraitTest methodsFor: 'testing' stamp: 'al 10/16/2005 14:55'!
testLocalMethodWithSameCodeInTrait
	"Note, takes all behaviors (classes and traits) into account"

	SystemNavigation default allBehaviorsDo: [ :each |
		each hasTraitComposition ifTrue: [
			each methodDict keys do: [ :selector |
				(each includesLocalSelector: selector) ifTrue: [
					(each traitComposition traitProvidingSelector: selector) ifNotNilDo: [ :trait |
						self deny: (trait >> selector = (each >> selector)) ] ] ] ] ].! !

!TraitTest methodsFor: 'testing' stamp: 'al 2/13/2006 17:41'!
testMarkerMethods
	"self debug: #testMarkerMethods"

	self t1 compile: 'm1 self foo bar'.
	self assert: (self t1 >> #m1) markerOrNil isNil.


	self t1 compile: 'm2 self requirement'.
	self assert: (self t1 >> #m2) markerOrNil == #requirement.
	
	self t1 compile: 'm3 ^self requirement'.
	self assert: (self t1 >> #m3) markerOrNil == #requirement.! !

!TraitTest methodsFor: 'testing' stamp: 'marcus.denker 8/15/2008 16:57'!
testPrinting
	self assertPrints: self t6 definition
		like: 'Trait named: #T6
	uses: T1 + T2 @ {#m22Alias->#m22}
	category: ''Traits-Tests'''! !

!TraitTest methodsFor: 'testing' stamp: 'md 8/15/2008 16:56'!
testPrintingClassSide
	"self run: #testPrintingClassSide"
	
	self assertPrints: self t6 classSide definition
		like: 'T6 classTrait
	uses: T1 classTrait + T2 classTrait'! !

!TraitTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:31'!
testRemoveFromSystem
	self t4 removeFromSystem.
	self deny: (Smalltalk includesKey: #T4).
	self assert: self t4 name = 'AnObsoleteT4'.
	self assert: self t4 methodDict isEmpty.
	self deny: (self t1 users includes: self t4)! !

!TraitTest methodsFor: 'testing' stamp: 'dvf 8/26/2005 14:32'!
testRequirement
	"self run: #testRequirement"

	self t1 compile: 'm self requirement'.
	self t2 compile: 'm ^true'.
	self assert: self t4 >> #m == (self t2 >> #m).
	self assert: self c2 new m.
	self t2 removeSelector: #m.
	self assert: self t5 >> #m == (self t1 >> #m).
	self should: [self c2 new m] raise: Error! !

!TraitTest methodsFor: 'testing' stamp: 'dvf 8/30/2005 14:55'!
testUsers
	self assert: self t1 users size = 3.
	self assert: (self t1 users includesAllOf: {self t4. self t5. self t6 }).
	self assert: self t3 users isEmpty.
	self assert: self t5 users size = 1.
	self assert: self t5 users anyOne = self c2.
	self c2 setTraitCompositionFrom: self t1 + self t5.
	self assert: self t5 users size = 1.
	self assert: self t5 users anyOne = self c2.
	self c2 setTraitComposition: self t2 asTraitComposition.
	self assert: self t5 users isEmpty! !

!TraitsTestCase class methodsFor: 'as yet unclassified' stamp: 'dvf 8/30/2005 11:41'!
resources
	^{TraitsResource}! !

!TraitsTestCase methodsFor: 'utility' stamp: 'al 12/16/2003 22:16'!
assertPrints: aString like: anotherString 
	self assert: (aString copyWithout: $ )
		= (anotherString copyWithout: $ )! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
c1
	^TraitsResource current c1! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
c2
	^TraitsResource current c2! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
c3
	^TraitsResource current c3! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
c4
	^TraitsResource current c4! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
c5
	^TraitsResource current c5! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
c6
	^TraitsResource current c6! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
c7
	^TraitsResource current c7! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
c8
	^TraitsResource current c8! !

!TraitsTestCase methodsFor: 'running' stamp: 'al 4/25/2004 16:31'!
categoryName
	^self class category! !

!TraitsTestCase methodsFor: 'utility' stamp: 'dvf 8/30/2005 14:44'!
createClassNamed: aSymbol superclass: aClass uses: aTraitComposition
	| class |
	class := aClass
		subclass: aSymbol
		uses: aTraitComposition
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: ''		
		category: self categoryName.
	self createdClassesAndTraits add: class.
	^class! !

!TraitsTestCase methodsFor: 'utility' stamp: 'dvf 8/30/2005 14:44'!
createTraitNamed: aSymbol uses: aTraitComposition
	| trait |
	trait := Trait
		named: aSymbol
		uses: aTraitComposition
		category: self categoryName.
	self createdClassesAndTraits add: trait.
	^trait! !

!TraitsTestCase methodsFor: 'utility' stamp: 'dvf 8/30/2005 14:44'!
createdClassesAndTraits
	createdClassesAndTraits ifNil: [
		createdClassesAndTraits := OrderedCollection new].
	^createdClassesAndTraits! !

!TraitsTestCase methodsFor: 'utility' stamp: 'dvf 8/30/2005 14:49'!
resourceClassesAndTraits
	^TraitsResource current createdClassesAndTraits! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
t1
	^TraitsResource current t1! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:44'!
t2
	^TraitsResource current t2! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:45'!
t3
	^TraitsResource current t3! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:45'!
t4
	^TraitsResource current t4! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:45'!
t5
	^TraitsResource current t5! !

!TraitsTestCase methodsFor: 'accessing' stamp: 'dvf 8/30/2005 11:45'!
t6
	^TraitsResource current t6! !

!TraitsTestCase methodsFor: 'running' stamp: 'dvf 8/30/2005 16:58'!
tearDown
	| behaviorName |
	TraitsResource resetIfDirty.
	self createdClassesAndTraits do: 
			[:aClassOrTrait | 
			behaviorName := aClassOrTrait name.
			Smalltalk at: behaviorName
				ifPresent: [:classOrTrait | classOrTrait removeFromSystem].
			ChangeSet current removeClassChanges: behaviorName].
	createdClassesAndTraits := nil! !

!TraitsTestCase methodsFor: 'testing-applying trait composition' stamp: 'al 9/19/2005 12:19'!
testChangeSuperclass
	"self run: #testChangeSuperclass"

	"Test that when the superclass of a class is changed the non-local methods
	of the class sending super are recompiled to correctly store the new superclass."

	| aC2 newSuperclass |
	aC2 := self c2 new.

	"C1 is current superclass of C2"
	self assert: aC2 m51.
	self assert: self c2 superclass == self c1.
	self deny: (self c2 localSelectors includes: #m51).

	"change superclass of C2 from C1 to X"
	newSuperclass := self createClassNamed: #X superclass: Object uses: {}.
	newSuperclass
		subclass: self c2 name
		uses: self c2 traitComposition
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: ''
		category: self c2 category.

	self assert: self c2 superclass == newSuperclass.

	newSuperclass compile: 'foo ^17'.
	self assert: aC2 m51 = 17.
	self deny: (self c2 localSelectors includes: #m51).

	self c2 compile: 'm51 ^19'.
	self assert: aC2 m51 = 19.

	self assert: self c2 >> #m52 == (self t5 >> #m52).

	
	! !

!SequenceableCollection methodsFor: '*traits' stamp: 'apb 8/22/2005 17:07'!
asTraitComposition
	"For convenience the composition {T1. T2 ...} is the same as T1 + T2 + ..."
	^self isEmpty
		ifFalse: [
			self size = 1
				ifTrue: [self first asTraitComposition]
				ifFalse: [
					self copyWithoutFirst 
						inject: self first
						into: [:left :right | left + right]]]
		ifTrue: [
			TraitComposition new]! !

Object subclass: #LocatedMethod
	instanceVariableNames: 'location selector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!LocatedMethod commentStamp: '<historical>' prior: 0!
I identify a method in the system by its selector and location (class or trait) where it is defined.!

!LocatedMethod class methodsFor: 'instance creation' stamp: 'al 1/23/2004 10:07'!
location: aPureBehavior selector: aSymbol
	^self new
		location: aPureBehavior selector: aSymbol;
		yourself! !

!LocatedMethod methodsFor: 'comparing' stamp: 'G.C 10/23/2008 10:12'!
= aLocatedMethod
	^ self method == aLocatedMethod method! !

!LocatedMethod methodsFor: 'comparing' stamp: 'al 3/6/2004 18:53'!
argumentNames
	"Return an array with the argument names of the method's selector"

	| keywords stream argumentNames argumentName delimiters |
	delimiters := {Character space. Character cr}.
	keywords := self selector keywords.
	stream := self source readStream.
	argumentNames := OrderedCollection new.
	keywords do: [ :each |
		stream match: each.
		[stream peekFor: Character space] whileTrue.
		argumentName := ReadWriteStream on: String new.
		[(delimiters includes: stream peek) or: [stream peek isNil]]
			whileFalse: [argumentName nextPut: stream next].
		argumentName isEmpty ifFalse: [
			argumentNames add: argumentName contents withBlanksTrimmed]].
	^(argumentNames copyFrom: 1 to: self method numArgs) asArray! !

!LocatedMethod methodsFor: 'convenience' stamp: 'al 3/9/2004 23:30'!
category
	^self location
		whichCategoryIncludesSelector: self selector! !

!LocatedMethod methodsFor: 'comparing' stamp: 'NS 3/30/2004 16:12'!
hash
	^ self method identityHash! !

!LocatedMethod methodsFor: 'testing' stamp: 'al 2/25/2004 11:20'!
isBinarySelector
	^self selector
		allSatisfy: [:each | each isSpecial]! !

!LocatedMethod methodsFor: 'accessing' stamp: 'al 1/23/2004 10:18'!
location
	^location! !

!LocatedMethod methodsFor: 'accessing' stamp: 'al 1/23/2004 10:07'!
location: aPureBehavior selector: aSymbol
	location := aPureBehavior.
	selector := aSymbol! !

!LocatedMethod methodsFor: 'convenience' stamp: 'al 1/23/2004 10:17'!
method
	^self location >> self selector! !

!LocatedMethod methodsFor: 'accessing' stamp: 'al 1/23/2004 10:17'!
selector
	^selector! !

!LocatedMethod methodsFor: 'convenience' stamp: 'al 2/16/2004 17:08'!
source
	^(self method
		getSourceFor: self selector
		in: self location) asString! !

Object subclass: #ModelExtension
	instanceVariableNames: 'interests lock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

ModelExtension class
	instanceVariableNames: 'current'!

ModelExtension class
	instanceVariableNames: 'current'!

ModelExtension subclass: #CodeModelExtension
	instanceVariableNames: 'perClassCache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

!CodeModelExtension classSide methodsFor: 'initialization' stamp: 'dvf 9/2/2005 12:20'!
isAbstract
	^self == CodeModelExtension! !

!CodeModelExtension methodsFor: 'access to cache' stamp: 'dvf 9/6/2005 15:29'!
cacheFor: aClass 
	^perClassCache at: aClass ifAbsentPut: [self newCacheFor: aClass]! !

!CodeModelExtension methodsFor: 'invalidation' stamp: 'dvf 1/31/2006 23:38'!
classChanged: modificationEvent 
	"We dont want to provide an out of date reply"
	modificationEvent itemClass ifNil: [self].
	self clearOut: modificationEvent itemClass
! !

!CodeModelExtension methodsFor: 'access to cache' stamp: 'dvf 9/1/2005 21:17'!
clearOut: aClass 
	^perClassCache removeKey: aClass ifAbsent: []! !

!CodeModelExtension methodsFor: 'access to cache' stamp: 'dvf 9/5/2005 14:20'!
for: aClass 
	| newSendCache |
	^perClassCache at: aClass
		ifAbsent: 
			[newSendCache := self newCacheFor: aClass.
			(self haveInterestsIn: aClass) 
				ifTrue: [perClassCache at: aClass put: newSendCache].
			newSendCache]! !

!CodeModelExtension methodsFor: 'access to cache' stamp: 'dvf 9/1/2005 21:18'!
initialize
	super initialize.
	perClassCache := IdentityDictionary new.! !

CodeModelExtension subclass: #LocalSends
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

!LocalSends commentStamp: '<historical>' prior: 0!
This class provide the model extension describing local sends for a given class. These are described in the comment for SendInfo, which is the class that actually computes this information.!

!LocalSends methodsFor: 'as yet unclassified' stamp: 'dvf 9/1/2005 18:48'!
newCacheFor: aClass 
	"Creates an instance of SendCaches, assigns it to the instance variable sendCaches and fills it with all the self-sends class-sends and super-sends that occur in methods defined in this class (or by used traits)."

	| localSendCache info |
	localSendCache := SendCaches new.
	aClass selectorsAndMethodsDo: 
			[:sender :m | 
			info := (SendInfo on: m) collectSends.
			info selfSentSelectors 
				do: [:sentSelector | localSendCache addSelfSender: sender of: sentSelector].
			info superSentSelectors 
				do: [:sentSelector | localSendCache addSuperSender: sender of: sentSelector].
			info classSentSelectors 
				do: [:sentSelector | localSendCache addClassSender: sender of: sentSelector]].
	^localSendCache! !

CodeModelExtension subclass: #ProvidedSelectors
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Requires'!

!ProvidedSelectors methodsFor: 'as yet unclassified' stamp: 'dvf 9/5/2005 14:20'!
isSelector: selector providedIn: aClass
	^(self haveInterestsIn: aClass) 
		ifFalse: [aClass classAndMethodFor: selector do: [:c :m | m isProvided] ifAbsent: [false]]
		ifTrue: [(self for: aClass) includes: selector]
		! !

!ProvidedSelectors methodsFor: 'as yet unclassified' stamp: 'dvf 9/1/2005 20:56'!
newCacheFor: aClass 

	| cache |
	aClass ifNil: [^IdentitySet new].
	cache := self for: aClass superclass copy.
	aClass selectorsAndMethodsDo: [:s :m | 
		m isProvided 
			ifTrue: [cache add: s]
			ifFalse: [cache remove: s ifAbsent: []]].
	^cache! !

CodeModelExtension subclass: #RequiredSelectors
	instanceVariableNames: 'dirty dirtyClasses newlyInterestingClasses'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Requires'!

!RequiredSelectors methodsFor: 'access to cache' stamp: 'dvf 9/18/2006 22:04'!
calculateForClass: aClass 
	| rscc |
	self clearOut: aClass.
	rscc := RequiredSelectorsChangesCalculator onModificationOf: { aClass }
				withTargets: { aClass }.
	rscc doWork! !

!RequiredSelectors methodsFor: 'invalidation' stamp: 'dvf 1/31/2006 23:39'!
classChanged: modificationEvent 
	self dirtyWithChange: modificationEvent! !

!RequiredSelectors methodsFor: 'as yet unclassified' stamp: 'dvf 9/5/2005 14:21'!
classesOfInterest
	^interests asIdentitySet! !

!RequiredSelectors methodsFor: 'accessing' stamp: 'dvf 9/2/2005 14:22'!
dirtyClasses
	dirtyClasses ifNil: [dirtyClasses := WeakSet new].
	^dirtyClasses! !

!RequiredSelectors methodsFor: 'as yet unclassified' stamp: 'dvf 9/2/2005 14:22'!
dirtyWithChange: anEvent 
	dirty := true.
	self dirtyClasses add: anEvent itemClass! !

!RequiredSelectors methodsFor: 'as yet unclassified' stamp: 'dvf 1/31/2006 23:38'!
ensureClean
	| rscc |
	dirty 
		ifTrue: 
			[rscc := RequiredSelectorsChangesCalculator 
						onModificationOf: self dirtyClasses
						withTargets: self classesOfInterest.
			rscc doWork.
			dirtyClasses := nil].
	dirty := false! !

!RequiredSelectors methodsFor: 'access to cache' stamp: 'dvf 9/5/2005 14:21'!
for: aClass 
	"Somewhat weird control flow, and populates the dictionary even with non-interesting things, which it probably shouldn't"
	perClassCache at: aClass ifAbsentPut: [RequirementsCache new].
	(self haveInterestsIn: aClass) 
		ifTrue: [self ensureClean]
		ifFalse: [self calculateForClass: aClass].
	^(perClassCache at: aClass) requirements! !

!RequiredSelectors methodsFor: 'access to cache' stamp: 'dvf 9/6/2005 15:40'!
lostInterest: client inAll: classes
	ProvidedSelectors current 
		lostInterest: self 
		inAll: (classes gather: [:cl | cl withAllSuperclasses]).
	LocalSends current 
		lostInterest: self 
		inAll: (classes gather: [:cl | cl withAllSuperclasses]).
	super lostInterest: client inAll: classes
! !

!RequiredSelectors methodsFor: 'access to cache' stamp: 'dvf 9/14/2005 15:32'!
lostOneInterestIn: aClass
	self lostInterest: nil in: aClass.
! !

!RequiredSelectors methodsFor: 'access to cache' stamp: 'dvf 9/6/2005 15:33'!
newCacheFor: aClass 
	^RequirementsCache new! !

!RequiredSelectors methodsFor: 'as yet unclassified' stamp: 'dvf 9/8/2005 20:05'!
newlyInteresting: aClass 
	dirty := true.
	self dirtyClasses add: aClass! !

!RequiredSelectors methodsFor: 'accessing' stamp: 'dvf 9/8/2005 18:46'!
newlyInterestingClasses
	newlyInterestingClasses ifNil: [newlyInterestingClasses := IdentitySet new].
	^newlyInterestingClasses! !

!RequiredSelectors methodsFor: 'accessing' stamp: 'dvf 9/8/2005 18:41'!
newlyInterestingClasses: anObject
	newlyInterestingClasses := anObject! !

!RequiredSelectors methodsFor: 'access to cache' stamp: 'dvf 9/14/2005 17:02'!
noteInterestOf: client inAll: classes 
	| newlyInteresting |
	LocalSends current noteInterestOf: self
		inAll: (classes gather: [:cl | cl withAllSuperclasses]).
	ProvidedSelectors current noteInterestOf: self
		inAll: (classes gather: [:cl | cl withAllSuperclasses]).
	newlyInteresting := classes copyWithoutAll: self classesOfInterest.
	super noteInterestOf: client inAll: classes.
	newlyInteresting do: [:cl | self newlyInteresting: cl]! !

!RequiredSelectors methodsFor: 'access to cache' stamp: 'dvf 9/14/2005 16:50'!
registerLifelongInterestOf: client inAll: classes 
	self noteInterestOf: client inAll: classes.
	classes do: [:cl | client toFinalizeSend: #lostOneInterestIn: to: self with: cl].! !

!ModelExtension classSide methodsFor: 'accessing' stamp: 'dvf 9/1/2005 21:16'!
current
	^current! !

!ModelExtension classSide methodsFor: 'accessing' stamp: 'dvf 9/1/2005 21:16'!
current: anObject
	^current := anObject! !

!ModelExtension classSide methodsFor: 'instance creation' stamp: 'dvf 9/1/2005 21:16'!
doWithTemporaryInstance: aBlock 
	| singleton |
	singleton := self current.
	
	[self current: self new.
	aBlock value] ensure: [self current: singleton]! !

!ModelExtension classSide methodsFor: 'initialization' stamp: 'dvf 9/2/2005 12:20'!
initialize
	self isAbstract not ifTrue:
		[self current: self new]! !

!ModelExtension classSide methodsFor: 'initialization' stamp: 'dvf 9/2/2005 12:20'!
isAbstract
	^self == ModelExtension! !

!ModelExtension methodsFor: 'access to cache' stamp: 'dvf 9/14/2005 11:27'!
haveInterestsIn: aClass 
	lock critical: [^interests includes: aClass]
! !

!ModelExtension methodsFor: 'invalidation' stamp: 'dvf 9/14/2005 11:11'!
initialize
	lock := Semaphore forMutualExclusion.  
	interests := IdentityBag new.
	SystemChangeNotifier uniqueInstance 
		notify: self
		ofSystemChangesOfItem: #class
		using: #classChanged:.
	SystemChangeNotifier uniqueInstance 
		notify: self
		ofSystemChangesOfItem: #method
		using: #classChanged:.
! !

!ModelExtension methodsFor: 'interests' stamp: 'dvf 9/2/2005 11:43'!
lostInterest: client in: class
	self lostInterest: client inAll: {class}! !

!ModelExtension methodsFor: 'interests' stamp: 'dvf 9/14/2005 11:11'!
lostInterest: client inAll: classes
	lock critical: [interests removeAll: classes]! !

!ModelExtension methodsFor: 'interests' stamp: 'dvf 9/2/2005 11:44'!
noteInterestOf: client in: class
	self noteInterestOf: client inAll: {class}! !

!ModelExtension methodsFor: 'interests' stamp: 'dvf 9/14/2005 11:28'!
noteInterestOf: client inAll: classes
	lock critical: [interests addAll: classes].! !

Object subclass: #RequiredSelectorsChangesCalculator
	instanceVariableNames: 'targetBehaviors classesToUpdate traitsToUpdate rootClasses modifiedBehaviors possiblyAffectedPerRoot originalSinsPerSelector targetClasses targetTraits'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Requires'!

!RequiredSelectorsChangesCalculator commentStamp: 'dvf 9/22/2005 14:20' prior: 0!
Nathanael implemented an efficient algorithm for updating the requirement status of a single selector for an inheritance subtree. However, the algorithm is not efficient enough to use as is for displaying abstractness of all the classes in a few class categories. 

To get to that performance level:
1. The RequiredSelectors class coordinates recalculation requests and tracks what classes have changed, and what classes are interesting. 
2. The current class handles a such request, by running the algorithm only on classes and selectors that may potentially be requirements.!

!RequiredSelectorsChangesCalculator classSide methodsFor: 'as yet unclassified' stamp: 'dvf 9/1/2005 20:44'!
onModificationOf: behaviors withTargets: targetBehaviors 
	| i |
	i := self new.
	i
		targetBehaviors: targetBehaviors;
		modifiedBehaviors: behaviors;
		decideParameters.
	^i! !

!RequiredSelectorsChangesCalculator methodsFor: 'as yet unclassified' stamp: 'dvf 9/9/2005 19:25'!
addUpdatePathTo: aClass from: highRoot 
	aClass withAllSuperclassesDo: [:sc | classesToUpdate add: sc. highRoot = sc ifTrue: [^self]]! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 8/9/2005 16:07'!
classesToUpdate
	^classesToUpdate! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 8/9/2005 16:07'!
classesToUpdate: anObject
	classesToUpdate := anObject! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/9/2005 19:23'!
decideParameters
	"decide whos who"

	targetClasses := IdentitySet new.
	targetTraits := IdentitySet new.
	self targetBehaviors 
		do: [:b | b isTrait ifTrue: [targetTraits add: b] ifFalse: [targetClasses add: b]].
	self findAffectedTraitsFrom: targetTraits.
	self findRootsAndRoutes.
	self findOriginalSins! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/12/2005 20:27'!
doWork
	| requiredSelectorsByClass oldRequiredSelectorsByClass classWithOldRequirementsRecorded rootsHandledBySel rootsHandled |
	requiredSelectorsByClass := IdentityDictionary new.
	oldRequiredSelectorsByClass := IdentityDictionary new.
	classWithOldRequirementsRecorded := IdentitySet new.
	rootsHandledBySel := IdentityDictionary new.
	originalSinsPerSelector keysAndValuesDo: 
			[:selector :sinners | 
			rootsHandled := rootsHandledBySel at: selector put: IdentitySet new.
			rootClasses do: 
					[:rc | 
					(self shouldProcess: rc forSinsIn: sinners) 
						ifTrue: 
							[rootsHandled add: rc.
							self 
								storeOldRequirementsUnder: rc
								into: oldRequiredSelectorsByClass
								ignoreSet: classWithOldRequirementsRecorded.
							self 
								storeRequirementsUnder: rc
								for: selector
								in: requiredSelectorsByClass]]].
	self 
		removeRequirements: oldRequiredSelectorsByClass
		thatAreNotIn: requiredSelectorsByClass
		ifIn: rootsHandledBySel.
	self setFoundRequirements: requiredSelectorsByClass! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'al 2/17/2006 08:54'!
findAffectedTraitsFrom: targetTraitsCollection
	traitsToUpdate := targetTraitsCollection 
				select: [:t | modifiedBehaviors anySatisfy: [:mb | t traitCompositionIncludes: mb]]! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/9/2005 20:48'!
findOriginalSins
	| sinnedSelectors sinners checkedClasses |
	checkedClasses := IdentitySet new.
	originalSinsPerSelector := IdentityDictionary new.
	rootClasses do: 
			[:rootClass | 
			rootClass withAllSuperclassesDo: [:superClass | 
				(checkedClasses includes: superClass) ifFalse: [
					checkedClasses add: superClass.
					sinnedSelectors := self sinsIn: superClass.
					sinnedSelectors do: 
							[:sinSel | 
							sinners := originalSinsPerSelector at: sinSel
										ifAbsentPut: [IdentitySet new].
							sinners add: superClass]]]]! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/12/2005 16:52'!
findRootsAndRoutes
	"Based on the 
	1. target classes (ones considered interesting by our clients) and the 
	2. modifiedBehaviors (ones we are told might have changed), 
	decide the 
	A. rootClasses (superclasses of target classes that include methods from modifiedBehaviors) 
	B. classesToUpdate (classes that may have been affected AND are on an inheritance path between a root class and a target class, will be updated by the algorithm. This includes the every target class that may have been affected).
	C. mapping from root classes to its classesToUpdate."

	| highestSuperclassOfCurrentTarget modifiedClasses |
	classesToUpdate := IdentitySet new.
	rootClasses := IdentitySet new.
	modifiedClasses := (modifiedBehaviors gather: [:mb | mb classesComposedWithMe]) asIdentitySet.
	targetClasses do: [:currentTargetClass | 
		highestSuperclassOfCurrentTarget := nil.
		currentTargetClass withAllSuperclassesDo: [:sc | 
			(modifiedClasses includes: sc) ifTrue: 
				[highestSuperclassOfCurrentTarget := sc.
				self noteRoot: sc possiblyAffected: currentTargetClass]].
			highestSuperclassOfCurrentTarget ifNotNilDo: [:highestRoot | 
				self addUpdatePathTo: currentTargetClass from: highestRoot]]! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/8/2005 16:14'!
initialize
	possiblyAffectedPerRoot := IdentityDictionary new.! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 9/1/2005 20:44'!
modifiedBehaviors
	^modifiedBehaviors! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 9/1/2005 20:44'!
modifiedBehaviors: anObject 
	modifiedBehaviors := anObject! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/12/2005 17:46'!
noteRoot: rootClass possiblyAffected: targetClass
	rootClasses add: rootClass.
	targetClass withAllSuperclassesDo: [:sc | 
		(self possiblyAffectedForRoot: rootClass) add: sc. rootClass = sc ifTrue: [^self]]
! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/8/2005 16:12'!
possiblyAffectedForRoot: rootClass
	^possiblyAffectedPerRoot at: rootClass ifAbsentPut: [IdentitySet new].! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/12/2005 20:15'!
removeRequirements: oldRequiredSelectorsByClass thatAreNotIn: requiredSelectorsByClass
	| cache newRequirements |
	oldRequiredSelectorsByClass keysAndValuesDo: 
			[:class :requirements | 
			newRequirements := requiredSelectorsByClass at: class
						ifAbsent: 
							[#()].
			cache := class requiredSelectorsCache.
			requirements 
				do: [:sel | (newRequirements includes: sel) ifFalse: [cache removeRequirement: sel]]]! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/12/2005 20:37'!
removeRequirements: oldRequiredSelectorsByClass thatAreNotIn: requiredSelectorsByClass ifIn: rootsHandledBySel
	| cache newRequirements unconfirmedRequirements roots affected |
	oldRequiredSelectorsByClass keysAndValuesDo: 
			[:class :oldRequirements | 
			newRequirements := requiredSelectorsByClass at: class
						ifAbsent: [#()].
			cache := class requiredSelectorsCache.
			unconfirmedRequirements := oldRequirements copyWithoutAll: newRequirements.
			unconfirmedRequirements do: [:sel | 
				roots := rootsHandledBySel at: sel ifAbsent: [#()].
				(roots anySatisfy: [:rc | 
					affected := possiblyAffectedPerRoot at: rc ifAbsent: #().
					(affected includes: class)]) ifTrue: [cache removeRequirement: sel]]]! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 8/9/2005 16:07'!
rootClasses
	^rootClasses! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 8/9/2005 16:07'!
rootClasses: anObject
	rootClasses := anObject! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/9/2005 18:22'!
selectorsToUpdateIn: aClass 
	^originalSinsPerSelector keys
! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/9/2005 23:25'!
setFoundRequirements: requiredSelectorsByClass 
	| cache |
	requiredSelectorsByClass keysAndValuesDo: 
			[:class :requirements | 
			cache := class requiredSelectorsCache.
			requirements do: [:sel | cache addRequirement: sel]].
	^cache! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/9/2005 23:36'!
shouldProcess: rc forSinsIn: sinners 
	rc withAllSuperclassesDo: [:rootSuperClass |
		(sinners includes: rootSuperClass) ifTrue: [^true].
		"theres a rootClass closer to the sin, we don't need to do it again."
		(rc ~= rootSuperClass and: [(rootClasses includes: rootSuperClass)]) ifTrue: [^false]].
	^false.! !

!RequiredSelectorsChangesCalculator methodsFor: 'as yet unclassified' stamp: 'dvf 9/9/2005 18:49'!
sinsIn: aClass 
	| negativeDefined selfSent sins |
	negativeDefined := IdentitySet new.
	aClass selectorsAndMethodsDo: [:s :m | m isProvided ifFalse: [negativeDefined add: s]].
	selfSent := aClass sendCaches selfSenders ifNil: [^negativeDefined] ifNotNilDo: [:dict | dict keys].
	sins := negativeDefined union: (selfSent copyWithoutAll: aClass providedSelectors).
	^sins! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/12/2005 17:35'!
storeOldRequirementsUnder: rc into: oldRequiredSelectorsByClass ignoreSet: classWithOldRequirementsRecorded 
	classesToUpdate do: 
			[:someClass | 
			(rc == someClass or: [(someClass inheritsFrom: rc)])
				ifTrue: 
					[(classWithOldRequirementsRecorded includes: someClass) 
						ifFalse: 
							[oldRequiredSelectorsByClass at: someClass put: someClass requirements]]]! !

!RequiredSelectorsChangesCalculator methodsFor: 'calculating' stamp: 'dvf 9/9/2005 23:27'!
storeRequirementsUnder: rc for: selector in: requiredSelectorsByClass 
	| requiringClasses selectorsForClass |
	requiringClasses := rc updateRequiredStatusFor: selector
				inSubclasses: (self possiblyAffectedForRoot: rc).
	^requiringClasses do: 
			[:requiringClass | 
			selectorsForClass := requiredSelectorsByClass at: requiringClass
						ifAbsentPut: [IdentitySet new].
			selectorsForClass add: selector]! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 8/9/2005 15:58'!
targetBehaviors
	^targetBehaviors! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 8/9/2005 15:58'!
targetBehaviors: anObject
	targetBehaviors := anObject! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 8/9/2005 16:07'!
traitsToUpdate
	^traitsToUpdate! !

!RequiredSelectorsChangesCalculator methodsFor: 'accessing' stamp: 'dvf 8/9/2005 16:07'!
traitsToUpdate: anObject
	traitsToUpdate := anObject! !

Object subclass: #RequirementsCache
	instanceVariableNames: 'requirements superRequirements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

!RequirementsCache methodsFor: 'updates' stamp: 'dvf 9/5/2005 11:34'!
addRequirement: selector
	requirements ifNil: [requirements := self newRequirementsObject].
	requirements add: selector.! !

!RequirementsCache methodsFor: 'accessing' stamp: 'dvf 9/5/2005 11:37'!
newRequirementsObject
	^ IdentitySet new.! !

!RequirementsCache methodsFor: 'updates' stamp: 'dvf 9/5/2005 11:37'!
removeRequirement: selector
	requirements ifNil: [^ self].
	requirements remove: selector ifAbsent: [].! !

!RequirementsCache methodsFor: 'accessing' stamp: 'dvf 9/5/2005 11:37'!
requirements
	^ requirements isNil
		ifTrue: [self newRequirementsObject]
		ifFalse: [requirements].! !

!RequirementsCache methodsFor: 'accessing' stamp: 'dvf 9/5/2005 11:37'!
superRequirements
	"Answer the value of superRequirements"

	^ superRequirements isNil
		ifTrue: [IdentitySet new]
		ifFalse: [superRequirements].! !

Object subclass: #SendCaches
	instanceVariableNames: 'selfSenders superSenders classSenders'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-LocalSends'!

!SendCaches commentStamp: 'NS 5/27/2005 15:13' prior: 0!
Instances of this class are used to keep track of selectors that are sent to self, sent to super, and sent to the class-side in any of the methods of a certain class. It also keeps track of the requirements of a class.

It's important to understand that this class just serves as storage for these sets of selectors. It does not contain any logic to actually compute them. In particular, it cannot compute the requirements.!

!SendCaches classSide methodsFor: 'fixup' stamp: 'apb 3/2/2006 23:24'!
initializeAllInstances
	self allSubInstancesDo: [ : each | each properlyInitialize ]! !

!SendCaches methodsFor: 'updates' stamp: 'apb 3/2/2006 23:29'!
addClassSender: sendingSelector of: sentSelector
	| senders |
	senders := classSenders at: sentSelector ifAbsent: [#()].
	classSenders at: sentSelector put: (senders copyWith: sendingSelector).! !

!SendCaches methodsFor: 'updates' stamp: 'apb 3/2/2006 23:27'!
addSelfSender: sendingSelector of: sentSelector
	| senders |
	senders := selfSenders at: sentSelector ifAbsent: [#()].
	selfSenders at: sentSelector put: (senders copyWith: sendingSelector).! !

!SendCaches methodsFor: 'updates' stamp: 'apb 3/2/2006 23:28'!
addSuperSender: sendingSelector of: sentSelector
	| senders |
	senders := superSenders at: sentSelector ifAbsent: [#()].
	superSenders at: sentSelector put: (senders copyWith: sendingSelector).! !

!SendCaches methodsFor: 'accessing-specific' stamp: 'NS 5/24/2005 20:58'!
allSentSelectorsAndSendersDo: aBlock
	self selfSentSelectorsAndSendersDo: aBlock.
	self superSentSelectorsAndSendersDo: aBlock.
	self classSentSelectorsAndSendersDo: aBlock.! !

!SendCaches methodsFor: 'accessing' stamp: 'apb 3/2/2006 23:28'!
classSendersOf: selector
	^ classSenders at: selector ifAbsent: [#()].! !

!SendCaches methodsFor: 'accessing-specific' stamp: 'apb 3/2/2006 23:29'!
classSentSelectorsAndSendersDo: aBlock
	classSenders keysAndValuesDo: aBlock! !

!SendCaches methodsFor: 'accessing' stamp: 'apb 3/2/2006 22:59'!
initialize
	selfSenders := IdentityDictionary new.
	superSenders := IdentityDictionary new.
	classSenders := IdentityDictionary new.! !

!SendCaches methodsFor: 'fixup' stamp: 'apb 3/2/2006 23:16'!
printOn: aStream 
	super printOn: aStream.
	aStream nextPut: $[.
	selfSenders printOn: aStream.
	aStream nextPut: $|.
	superSenders printOn: aStream.
	aStream nextPut: $|.
	classSenders printOn: aStream.
	aStream nextPut: $]! !

!SendCaches methodsFor: 'fixup' stamp: 'apb 3/2/2006 23:18'!
properlyInitialize
	selfSenders isEmptyOrNil ifTrue: [selfSenders := IdentityDictionary new].
	superSenders isEmptyOrNil ifTrue: [superSenders := IdentityDictionary new].
	classSenders isEmptyOrNil ifTrue: [classSenders := IdentityDictionary new].
	
! !

!SendCaches methodsFor: 'accessing' stamp: 'dvf 8/8/2005 11:20'!
selfSenders
	^selfSenders! !

!SendCaches methodsFor: 'accessing' stamp: 'dvf 8/8/2005 11:20'!
selfSenders: anObject
	^selfSenders := anObject! !

!SendCaches methodsFor: 'accessing-specific' stamp: 'apb 3/2/2006 23:25'!
selfSendersOf: selector
	^ selfSenders at: selector ifAbsent: [#()].! !

!SendCaches methodsFor: 'accessing-specific' stamp: 'apb 3/2/2006 23:26'!
selfSentSelectorsAndSendersDo: aBlock
	selfSenders keysAndValuesDo: aBlock! !

!SendCaches methodsFor: 'accessing' stamp: 'dvf 8/8/2005 11:24'!
superSenders
	^superSenders! !

!SendCaches methodsFor: 'accessing' stamp: 'dvf 8/8/2005 11:24'!
superSenders: anObject
	^superSenders := anObject! !

!SendCaches methodsFor: 'accessing-specific' stamp: 'apb 3/2/2006 23:28'!
superSendersOf: selector
	^ superSenders at: selector ifAbsent: [#()].! !

!SendCaches methodsFor: 'accessing-specific' stamp: 'apb 3/2/2006 23:27'!
superSentSelectorsAndSendersDo: aBlock
	superSenders keysAndValuesDo: aBlock! !

Object subclass: #TraitBehavior
	uses: TPureBehavior @ {#pureRemoveSelector:->#removeSelector:. #pureAddSelectorSilently:withMethod:->#addSelectorSilently:withMethod:}
	instanceVariableNames: 'methodDict traitComposition localSelectors users'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!TraitBehavior commentStamp: 'apb 3/2/2006 17:49' prior: 0!
I declare the instance variables methodDict, traitComposition and localSelectors that normally would be defined on PureBehavior (see class comment of PureBehavior). Furthermore I keep track of where I am used and I provide compatibility methods to make me useable in tools where originally a class was expected.!

TraitBehavior class
	uses: TPureBehavior classTrait
	instanceVariableNames: ''!

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/20/2005 21:24'!
addSelectorSilently: selector withMethod: compiledMethod
	self pureAddSelectorSilently: selector withMethod: compiledMethod.
	self notifyUsersOfChangedSelector: selector.! !

!TraitBehavior methodsFor: 'traits' stamp: 'NS 4/12/2004 14:16'!
addUser: aClassOrTrait
	users add: aClassOrTrait! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'dc 5/8/2007 13:51'!
allClassVarNames
	^#()! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:36'!
allInstVarNames
	^ #()! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 1/6/2004 12:24'!
allSelectors
	^ self selectors! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:34'!
allSubclassesDo: aBlock! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'al 12/31/2005 13:41'!
allSuperclasses
	^ OrderedCollection new! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:34'!
allSuperclassesDo: aBlock! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 1/6/2004 09:02'!
basicLocalSelectors
	"Direct accessor for the instance variable localSelectors.
	Since localSelectors is lazily initialized, this may 
	return nil, which means that all selectors are local."

	^ localSelectors! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 3/30/2004 13:26'!
basicLocalSelectors: aSetOrNil
	localSelectors := aSetOrNil! !

!TraitBehavior methodsFor: 'remove me later' stamp: 'apb 8/22/2005 15:28'!
classPool
	^ Dictionary new! !

!TraitBehavior methodsFor: 'traits' stamp: 'dvf 9/9/2005 19:39'!
classesComposedWithMe
	^users gather: [:u | u classesComposedWithMe]
! !

!TraitBehavior methodsFor: 'initialize-release' stamp: 'al 4/21/2004 21:56'!
forgetDoIts
	"get rid of old DoIt methods"
	self 
		basicRemoveSelector: #DoIt;
		basicRemoveSelector: #DoItIn:! !

!TraitBehavior methodsFor: 'traits' stamp: 'al 7/21/2004 20:57'!
hasTraitComposition
	^traitComposition notNil! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'al 7/29/2004 15:03'!
inheritsFrom: aClass
	"Used by RB"
	
	^false! !

!TraitBehavior methodsFor: 'initialize-release' stamp: 'NS 4/12/2004 14:28'!
initialize
	self methodDict: MethodDictionary new.
	self traitComposition: nil.
	users := IdentitySet new.! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:33'!
instSize
	^0! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:36'!
instVarNames
	^#()! !

!TraitBehavior methodsFor: 'testing' stamp: 'al 2/9/2004 22:29'!
isTrait
	^true! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 12/10/2003 10:02'!
lookupSelector: selector
	^(self includesSelector: selector)
		ifTrue: [self compiledMethodAt: selector]
		ifFalse: [nil]! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 12/9/2003 10:51'!
methodDict
	^ methodDict! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'NS 12/9/2003 10:50'!
methodDict: aDictionary
	methodDict := aDictionary! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 7/17/2004 14:31'!
precodeCommentOrInheritedCommentFor: aSelector
	^self firstPrecodeCommentFor: aSelector
	! !

!TraitBehavior methodsFor: 'traits' stamp: 'al 4/20/2004 09:33'!
removeFromTraitCompositionOfUsers
	self users do: [:each |
		each removeFromComposition: self ]! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 9/20/2005 21:24'!
removeSelector: selector
	self pureRemoveSelector: selector.
	self notifyUsersOfChangedSelector: selector.! !

!TraitBehavior methodsFor: 'traits' stamp: 'NS 4/12/2004 14:17'!
removeUser: aClassOrTrait
	users remove: aClassOrTrait ifAbsent: []! !

!TraitBehavior methodsFor: 'send caches' stamp: 'dvf 9/6/2005 14:12'!
requiredSelectors
	| sss selfSentNotProvided otherRequired |
	sss := self selfSentSelectorsFromSelectors: self allSelectors.
	selfSentNotProvided := sss copyWithoutAll: (self allSelectors select: [:e | (self >> e) isProvided]).
	otherRequired := self allSelectors select: [:e | (self >> e) isRequired].
	^(selfSentNotProvided, otherRequired) asSet
! !

!TraitBehavior methodsFor: 'send caches' stamp: 'dvf 9/1/2005 16:40'!
sendCaches
	^LocalSends current for: self! !

!TraitBehavior methodsFor: 'remove me later' stamp: 'apb 8/22/2005 15:29'!
sharedPools
	^ Dictionary new! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'NS 4/12/2004 14:34'!
subclassDefinerClass
	^nil subclassDefinerClass ! !

!TraitBehavior methodsFor: 'testing method dictionary' stamp: 'G.C 10/22/2008 09:59'!
thoroughWhichSelectorsReferTo: literal special: specialFlag byte: specialByte 
	"Answer a set of selectors whose methods access the argument as a 
	literal. Dives into the compact literal notation, making it slow but 
	thorough "
	| selectors |
	selectors := IdentitySet new.
	self selectorsAndMethodsDo: 
		[ :sel :method | 
		((method refersToLiteral: literal) or: [ specialFlag and: [ method scanFor: specialByte ] ]) ifTrue: [ selectors add: sel ] ].
	^ selectors! !

!TraitBehavior methodsFor: 'traits' stamp: 'al 12/12/2003 21:42'!
traitComposition
	traitComposition ifNil: [traitComposition := TraitComposition new].
	^traitComposition! !

!TraitBehavior methodsFor: 'traits' stamp: 'NS 12/11/2003 16:35'!
traitComposition: aTraitComposition
	traitComposition := aTraitComposition! !

!TraitBehavior methodsFor: 'send caches' stamp: 'dvf 8/9/2005 17:22'!
updateRequires
	| sss aTrait |
	sss := self selfSentSelectorsInTrait: aTrait.
	^sss copyWithoutAll: aTrait allSelectors.! !

!TraitBehavior methodsFor: 'traits' stamp: 'NS 4/12/2004 14:17'!
users
	^users! !

!TraitBehavior methodsFor: 'class compatibility' stamp: 'al 7/17/2004 14:35'!
whichClassIncludesSelector: aSymbol 
	"Traits compatibile implementation for:
	
	Answer the class on the receiver's superclass chain where the 
	argument, aSymbol (a message selector), will be found. Answer nil if none found."
	
	^(self includesSelector: aSymbol)
		ifTrue: [self]
		ifFalse: [nil]! !

!TraitBehavior methodsFor: 'accessing method dictionary' stamp: 'al 2/8/2004 19:08'!
zapAllMethods
	"Remove all methods in this trait which is assumed to be obsolete"

	methodDict := MethodDictionary new.
	self hasClassTrait ifTrue: [self classTrait zapAllMethods]! !

TraitBehavior subclass: #TraitDescription
	uses: TClassAndTraitDescription + TComposingDescription + TTransformationCompatibility
	instanceVariableNames: 'organization'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!TraitDescription commentStamp: '<historical>' prior: 0!
I add a number of facilities (most defined by traits which are also used in ClassDescription):
	Category organization for methods
	The maintenance of a ChangeSet, and logging changes on a file
	Most of the mechanism for fileOut.
	Copying of methods to other traits/classes
	Operators to create trait compositions!

TraitDescription class
	uses: TComposingDescription classTrait + TTransformationCompatibility classTrait + TClassAndTraitDescription classTrait
	instanceVariableNames: ''!

TraitDescription subclass: #ClassTrait
	uses: TApplyingOnClassSide
	instanceVariableNames: 'baseTrait'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!ClassTrait commentStamp: '<historical>' prior: 0!
While every class has an associated metaclass, a trait can have an associated classtrait, an instance of me. To preserve metaclass compatibility, the associated classtrait (if there is one) is automatically applied to the metaclass, whenever a trait is applied to a class. Consequently, a trait with an associated classtrait can only be applied to classes, whereas a trait without a classtrait can be applied to both classes and metaclasses.!

ClassTrait class
	uses: TApplyingOnClassSide classTrait
	instanceVariableNames: ''!

!ClassTrait class methodsFor: 'instance creation' stamp: 'al 3/23/2004 19:41'!
for: aTrait
	^self new
		initializeWithBaseTrait: aTrait;
		yourself! !

!ClassTrait methodsFor: 'initialize' stamp: 'al 7/18/2004 12:11'!
baseClass: aTrait traitComposition: aComposition methodDict: aMethodDict localSelectors: aSet organization: aClassOrganization

	self baseTrait: aTrait.
	self	
		traitComposition: aComposition
		methodDict: aMethodDict
		localSelectors: aSet
		organization: aClassOrganization
! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:38'!
baseTrait
	^baseTrait! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:48'!
baseTrait: aTrait
	self assert: aTrait isBaseTrait.
	baseTrait := aTrait
	
	! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:38'!
classTrait
	^self! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 4/20/2004 09:44'!
classTrait: aClassTrait
	self error: 'Trait is already a class trait!!'
	
	! !

!ClassTrait methodsFor: 'compiling' stamp: 'al 4/7/2004 14:54'!
compile: text classified: category withStamp: changeStamp notifying: requestor logSource: logSource
	
	| classSideUsersOfBaseTrait message |
	classSideUsersOfBaseTrait := self baseTrait users select: [:each | each isClassSide].
	classSideUsersOfBaseTrait isEmpty ifFalse: [
		message := String streamContents: [:stream |
			stream nextPutAll: 'The instance side of this trait is used on '; cr.
			classSideUsersOfBaseTrait
				do: [:each | stream nextPutAll: each name]
				separatedBy: [ stream nextPutAll: ', '].
			stream cr; nextPutAll: ' You can not add methods to the class side of this trait!!'].
		^TraitException signal:  message].
	
	^super
		compile: text
		classified: category
		withStamp: changeStamp
		notifying: requestor
		logSource: logSource! !

!ClassTrait methodsFor: 'copying' stamp: 'dvf 8/30/2005 16:51'!
copy
	"Make a copy of the receiver. Share the 
	reference to the base trait."

	^(self class new)
		baseTrait: self baseTrait;
		initializeFrom: self;
		yourself! !

!ClassTrait methodsFor: 'filein/out' stamp: 'al 3/26/2006 21:31'!
definitionST80
	^String streamContents: [:stream |
		stream
			nextPutAll: self name;
			crtab;
			nextPutAll: 'uses: ';
			nextPutAll: self traitCompositionString]! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:41'!
hasClassTrait
	^false! !

!ClassTrait methodsFor: 'initialize' stamp: 'dvf 8/30/2005 16:48'!
initializeFrom: anotherClassTrait
	traitComposition := self traitComposition copyTraitExpression.
	methodDict := self methodDict copy.
	localSelectors := self localSelectors copy.
	organization := self organization copy.! !

!ClassTrait methodsFor: 'initialize' stamp: 'al 3/24/2004 20:37'!
initializeWithBaseTrait: aTrait
	self baseTrait: aTrait.
	self noteNewBaseTraitCompositionApplied: aTrait traitComposition.
	aTrait users do: [:each | self addUser: each classSide].
	! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:48'!
isBaseTrait
	^false! !

!ClassTrait methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:48'!
isClassTrait
	^true! !

!ClassTrait methodsFor: 'accessing' stamp: 'al 4/21/2004 09:38'!
name
	^self baseTrait name , ' classTrait'! !

!ClassTrait methodsFor: 'initialize' stamp: 'al 7/17/2004 22:56'!
traitComposition: aComposition methodDict: aMethodDict localSelectors: aSet organization: aClassOrganization

	"Used by copy of Trait"

	localSelectors := aSet.
	methodDict := aMethodDict.
	traitComposition := aComposition.
	self organization: aClassOrganization! !

!ClassTrait methodsFor: 'composition' stamp: 'al 7/18/2004 14:02'!
uses: aTraitCompositionOrArray
	| copyOfOldTrait newComposition |
	copyOfOldTrait := self copy.
	newComposition := aTraitCompositionOrArray asTraitComposition.
	self assertConsistantCompositionsForNew: newComposition.
	self setTraitComposition: newComposition.
	SystemChangeNotifier uniqueInstance
		traitDefinitionChangedFrom: copyOfOldTrait to: self.! !

TraitDescription subclass: #Trait
	uses: TBehaviorCategorization
	instanceVariableNames: 'name environment classTrait category'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Kernel'!

!Trait commentStamp: '<historical>' prior: 0!
Each trait in the system is represented as an instance of me. Like Class, I concretize my superclass by providing instance variables for the name and the environment. Since traits do not define variables, I not provide facilities for pool variables. However, I declare an instance variable to hold the associated classtrait, which is an instance of the class ClassTrait. !

Trait class
	uses: TBehaviorCategorization classTrait
	instanceVariableNames: ''!

!Trait class methodsFor: 'instance creation' stamp: 'al 12/7/2003 16:00'!
defaultEnvironment
	^Smalltalk! !

!Trait class methodsFor: 'instance creation' stamp: 'md 3/3/2006 11:05'!
named: aSymbol uses: aTraitCompositionOrCollection category: aString
	| env |
	env := self environment.
	^self
		named: aSymbol
		uses: aTraitCompositionOrCollection
		category: aString
		env: env! !

!Trait class methodsFor: 'instance creation' stamp: 'al 3/18/2006 13:48'!
named: aSymbol uses: aTraitCompositionOrCollection category: aString env: anEnvironment
	| trait oldTrait systemCategory |
	systemCategory := aString asSymbol.
	trait := anEnvironment
		at: aSymbol
		ifAbsent: [nil].
	oldTrait := trait copy.
	trait := trait ifNil: [super new].
	
	(trait isKindOf: Trait) ifFalse: [
		^self error: trait name , ' is not a Trait'].
	trait
		setName: aSymbol
		andRegisterInCategory: systemCategory
		environment: anEnvironment.
		
	trait setTraitComposition: aTraitCompositionOrCollection asTraitComposition.
	
	"... notify interested clients ..."
	oldTrait isNil ifTrue: [
		SystemChangeNotifier uniqueInstance classAdded: trait inCategory: systemCategory.
		^ trait].
	SystemChangeNotifier uniqueInstance traitDefinitionChangedFrom: oldTrait to: trait.
	systemCategory ~= oldTrait category 
		ifTrue: [SystemChangeNotifier uniqueInstance class: trait recategorizedFrom: oldTrait category to: systemCategory].
		
	^ trait! !

!Trait class methodsFor: 'instance creation' stamp: 'al 12/19/2003 15:03'!
new
	self shouldNotImplement! !

!Trait class methodsFor: 'printing' stamp: 'al 2/9/2004 22:23'!
newTemplateIn: categoryString
	^String streamContents: [:stream |
		stream
			nextPutAll: self name;
			nextPutAll: ' named: #NameOfTrait';
			cr; tab;
			nextPutAll: 'uses: {}';
			cr; tab;
			nextPutAll: 'category: ';
			nextPut: $';
			nextPutAll: categoryString;
			nextPut: $' ]! !

!Trait methodsFor: 'private' stamp: 'al 7/17/2004 14:16'!
applyChangesOfNewTraitCompositionReplacing: oldComposition
	"Duplicated on Class"
	
	| changedSelectors |
	changedSelectors := super applyChangesOfNewTraitCompositionReplacing: oldComposition.
	self classSide
		noteNewBaseTraitCompositionApplied: self traitComposition.
	^ changedSelectors! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
baseTrait
	^self! !

!Trait methodsFor: 'accessing' stamp: 'al 3/18/2006 13:23'!
basicCategory
	^category! !

!Trait methodsFor: 'accessing' stamp: 'al 3/18/2006 13:22'!
basicCategory: aSymbol
	category := aSymbol! !

!Trait methodsFor: 'compiling' stamp: 'md 3/5/2006 23:47'!
binding

	^ Smalltalk associationAt: name ifAbsent: [nil -> self]
! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
classTrait
	^classTrait! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
classTrait: aTrait
	"Assigns the class trait associated with the receiver."
	
	self assert: aTrait isClassTrait.
	classTrait := aTrait! !

!Trait methodsFor: 'copying' stamp: 'dvf 8/30/2005 16:49'!
copy 
	| newTrait |
	newTrait := self class basicNew initialize
		name: self name
		traitComposition: self traitComposition copyTraitExpression 
		methodDict: self methodDict copy
		localSelectors: self localSelectors copy
		organization: self organization copy.
		
	newTrait classTrait initializeFrom: self classTrait.
	^newTrait! !

!Trait methodsFor: 'accessing' stamp: 'al 7/17/2004 14:16'!
environment
	^environment! !

!Trait methodsFor: 'accessing' stamp: 'al 7/17/2004 14:16'!
environment: anObject
	environment := anObject! !

!Trait methodsFor: 'filein/out' stamp: 'al 12/1/2005 16:29'!
fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex 
	"This is just copied from Class.. the whole fileout is a mess."
	^self fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex initializing: true! !

!Trait methodsFor: 'filein/out' stamp: 'al 12/1/2005 16:28'!
fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex initializing: aBool
	"File a description of the receiver on aFileStream. If the boolean argument,
	moveSource, is true, then set the trailing bytes to the position of aFileStream and
	to fileIndex in order to indicate where to find the source code."

	Transcript cr; show: name.
	super
		fileOutOn: aFileStream
		moveSource: moveSource
		toFile: fileIndex.
	self hasClassTrait ifTrue: [
		aFileStream cr; nextPutAll: '"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!!'; cr; cr.
		self classTrait
			fileOutOn: aFileStream
			moveSource: moveSource
			toFile: fileIndex]! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
hasClassTrait
	^classTrait notNil! !

!Trait methodsFor: 'initialization' stamp: 'al 7/17/2004 14:16'!
initialize
	super initialize.
	classTrait := ClassTrait for: self! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
isBaseTrait
	^true! !

!Trait methodsFor: 'accessing parallel hierarchy' stamp: 'al 7/17/2004 14:16'!
isClassTrait
	^false! !

!Trait methodsFor: 'testing' stamp: 'al 7/17/2004 14:16'!
isObsolete
	"Return true if the receiver is obsolete."
	^(self environment at: name ifAbsent: [nil]) ~~ self! !

!Trait methodsFor: 'private' stamp: 'al 7/17/2004 14:16'!
isValidTraitName: aSymbol
	^(aSymbol isEmptyOrNil
		or: [aSymbol first isLetter not]
		or: [aSymbol anySatisfy: [:character | character isAlphaNumeric not]]) not! !

!Trait methodsFor: 'accessing' stamp: 'al 7/17/2004 14:16'!
name
	^name! !

!Trait methodsFor: 'accessing' stamp: 'al 7/17/2004 14:16'!
name: aSymbol
	name := aSymbol! !

!Trait methodsFor: 'initialization' stamp: 'al 7/17/2004 23:00'!
name: aString traitComposition: aComposition methodDict: aMethodDict localSelectors: aSet organization: aClassOrganization

	"Used by copy"
	
	self name: aString.
	localSelectors := aSet.
	methodDict := aMethodDict.
	traitComposition := aComposition.
	self organization: aClassOrganization
	
	! !

!Trait methodsFor: 'initialization' stamp: 'al 7/29/2004 16:48'!
obsolete
	self name: ('AnObsolete' , self name) asSymbol.
	self hasClassTrait ifTrue: [
		self classTrait obsolete].
	super obsolete! !

!Trait methodsFor: 'initialization' stamp: 'al 7/17/2004 14:16'!
removeFromSystem
	self removeFromSystem: true! !

!Trait methodsFor: 'initialization' stamp: 'al 7/17/2004 14:16'!
removeFromSystem: logged
	self environment forgetClass: self logged: logged.
	self obsolete! !

!Trait methodsFor: 'private' stamp: 'al 7/17/2004 14:16'!
rename: aString 
	"The new name of the receiver is the argument, aString."

	| newName |
	(newName := aString asSymbol) ~= self name
		ifFalse: [^ self].
	(self environment includesKey: newName)
		ifTrue: [^ self error: newName , ' already exists'].
	(Undeclared includesKey: newName)
		ifTrue: [self inform: 'There are references to, ' , aString printString , '
from Undeclared. Check them after this change.'].
	self environment renameClass: self as: newName.
	name := newName! !

!Trait methodsFor: 'accessing' stamp: 'dvf 9/6/2005 13:11'!
requirements
	^self requiredSelectors! !

!Trait methodsFor: 'private' stamp: 'al 7/17/2004 14:16'!
setName: aSymbol andRegisterInCategory: categorySymbol environment: aSystemDictionary
	(self isValidTraitName: aSymbol) ifFalse: [TraitException signal: 'Invalid trait name'].
	
	(self environment == aSystemDictionary
		and: [self name = aSymbol
			and: [self category = categorySymbol]]) ifTrue: [^self].
		
	((aSystemDictionary includes: aSymbol) and: [(aSystemDictionary at: aSymbol) ~~ self])
		ifTrue: [TraitException signal: 'The name ''' , aSymbol , ''' is already used'].

	(self environment notNil and: [self name notNil and: [self name ~= aSymbol]]) ifTrue: [
		self environment renameClass: self as: aSymbol].
	
	self name: aSymbol.
	self environment: aSystemDictionary.	
	self environment at: self name put: self.
	self environment organization classify: self name under: categorySymbol.
	^ true! !

!TraitDescription methodsFor: 'composition' stamp: 'al 12/7/2005 21:00'!
addExclusionOf: aSymbol
	^self - {aSymbol}! !

!TraitDescription methodsFor: 'accessing method dictionary' stamp: 'al 6/29/2004 21:42'!
allMethodsInCategory: aName 
	"Answer a list of all the method categories of the receiver"
	
	| aColl |
	aColl := aName = ClassOrganizer allCategory
		ifTrue: [self organization allMethodSelectors]
		ifFalse: [self organization listAtCategoryNamed: aName].
	^aColl asSet asSortedArray

	"TileMorph allMethodsInCategory: #initialization"! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 4/21/2004 10:01'!
baseTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'organization' stamp: 'al 3/18/2006 14:10'!
category
	self subclassResponsibility! !

!TraitDescription methodsFor: 'organization' stamp: 'al 3/18/2006 14:10'!
category: aString
	self subclassResponsibility! !

!TraitDescription methodsFor: 'fileIn/Out' stamp: 'marcus.denker 9/14/2008 19:07'!
classComment: aString stamp: aStamp
	"Store the comment, aString or Text or RemoteString, associated with the class we are organizing.  Empty string gets stored only if had a non-empty one before."

	| ptr header file oldCommentRemoteStr |
	(aString isKindOf: RemoteString) ifTrue:
		[SystemChangeNotifier uniqueInstance classCommented: self.
		^ self organization classComment: aString stamp: aStamp].

	oldCommentRemoteStr := self organization commentRemoteStr.
	(aString size = 0) & (oldCommentRemoteStr isNil) ifTrue: [^ self organization classComment: nil].
		"never had a class comment, no need to write empty string out"

	ptr := oldCommentRemoteStr ifNil: [0] ifNotNil: [oldCommentRemoteStr sourcePointer].
	SourceFiles ifNotNil: [(file := SourceFiles at: 2) ifNotNil:
		[file setToEnd; cr; nextPut: $!!.	"directly"
		"Should be saying (file command: 'H3') for HTML, but ignoring it here"
		header := String streamContents: [:strm | strm nextPutAll: self name;
			nextPutAll: ' commentStamp: '.
			aStamp storeOn: strm.
			strm nextPutAll: ' prior: '; nextPutAll: ptr printString].
		file nextChunkPut: header]].
	self organization classComment: (RemoteString newString: aString onFileNumber: 2) stamp: aStamp.
	SystemChangeNotifier uniqueInstance classCommented: self.
! !

!TraitDescription methodsFor: 'accessing comment' stamp: 'al 10/17/2005 10:45'!
classCommentBlank

	| existingComment stream |
	existingComment := self instanceSide organization classComment.
	existingComment isEmpty
		ifFalse: [^existingComment].

	stream := WriteStream on: (String new: 100).
	stream
		nextPutAll: 'A';
		nextPutAll: (self name first isVowel ifTrue: ['n '] ifFalse: [' ']);
		nextPutAll: self name;
		nextPutAll: ' is xxxxxxxxx.'.
	stream cr.
	^stream contents! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'NS 4/12/2004 15:05'!
classSide
	^self classTrait! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 4/21/2004 10:01'!
classTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'copying' stamp: 'marcus.denker 9/14/2008 19:06'!
copy: sel from: class classified: cat 
	"Install the method associated with the first arugment, sel, a message 
	selector, found in the method dictionary of the second argument, class, 
	as one of the receiver's methods. Classify the message under the third 
	argument, cat."

	| code category |
	"Useful when modifying an existing class"
	code := class sourceMethodAt: sel.
	code ifNotNil: 
			[cat 
				ifNil: [category := class organization categoryOfElement: sel]
				ifNotNil: [category := cat].
			(self includesSelector: sel)
				ifTrue: [code asString = (self sourceMethodAt: sel) asString 
							ifFalse: [self error: self name 
										, ' ' 
										, sel 
										, ' will be redefined if you proceed.']].
			self compile: code classified: category]! !

!TraitDescription methodsFor: 'copying' stamp: 'dvf 9/22/2005 17:54'!
copyTraitExpression
	"When recursively copying a trait expression, the primitive traits should NOT be copied
because they are globally named 'well-known' objects"
	^ self ! !

!TraitDescription methodsFor: 'fileIn/Out' stamp: 'al 3/26/2006 21:50'!
definitionST80
	^String streamContents: [:stream |
		stream nextPutAll: self class name.
		stream nextPutAll: ' named: ';
				store: self name.
		stream cr; tab; nextPutAll: 'uses: ';
				nextPutAll: self traitCompositionString.
		stream cr; tab; nextPutAll: 'category: ';
				store: self category asString].! !

!TraitDescription methodsFor: 'fileIn/Out' stamp: 'al 8/20/2008 10:54'!
fileOut
	"Create a file whose name is the name of the receiver with '.st' as the 
	extension, and file a description of the receiver onto it."
	
	| internalStream |
	internalStream := WriteStream on: (String new: 100).
	internalStream header; timeStamp.

	self fileOutOn: internalStream moveSource: false toFile: 0.
	internalStream trailer.

	FileStream writeSourceCodeFrom: internalStream baseName: self name isSt: true! !

!TraitDescription methodsFor: 'initialization' stamp: 'NS 4/8/2004 10:59'!
forgetDoIts
	"get rid of old DoIt methods and bogus entries in the ClassOrganizer."
	SystemChangeNotifier uniqueInstance doSilently: [
		self organization
			removeElement: #DoIt;
			removeElement: #DoItIn:.
	].
	super forgetDoIts.! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 4/21/2004 10:02'!
hasClassTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'NS 4/12/2004 15:05'!
instanceSide
	^self baseTrait! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:47'!
isBaseTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'accessing parallel hierarchy' stamp: 'al 3/16/2004 10:47'!
isClassTrait
	self subclassResponsibility! !

!TraitDescription methodsFor: 'testing' stamp: 'al 4/9/2004 15:28'!
isTestCase
	^false! !

!TraitDescription methodsFor: 'private' stamp: 'marcus.denker 8/25/2008 12:05'!
linesOfCode 
	"An approximate measure of lines of code.
	Includes comments, but excludes blank lines."
	| lines |
	lines := self localMethods inject: 0 into: [:sum :each | sum + each linesOfCode]. 
	self isMeta 
		ifTrue: [^ lines]
		ifFalse: [^ lines + self class linesOfCode]! !

!TraitDescription methodsFor: 'private' stamp: 'marcus.denker 9/14/2008 19:04'!
logMethodSource: aText forMethodWithNode: aCompiledMethodWithNode inCategory: category withStamp: changeStamp notifying: requestor
	| priorMethodOrNil newText |
	priorMethodOrNil := self compiledMethodAt: aCompiledMethodWithNode selector ifAbsent: [].
	newText := ((requestor isNil or: [requestor isKindOf: SyntaxError]) not
						and: [Preferences confirmFirstUseOfStyle])
			ifTrue: [aText askIfAddStyle: priorMethodOrNil req: requestor]
			ifFalse: [aText].
	aCompiledMethodWithNode method putSource: newText
		fromParseNode: aCompiledMethodWithNode node
		class: self category: category withStamp: changeStamp 
		inFile: 2 priorMethod: priorMethodOrNil.! !

!TraitDescription methodsFor: 'organization updating' stamp: 'al 12/1/2005 15:50'!
notifyOfRecategorizedSelector: element from: oldCategory to: newCategory
	SystemChangeNotifier uniqueInstance selector: element recategorizedFrom: oldCategory to: newCategory inClass: self.
	SystemChangeNotifier uniqueInstance 
		doSilently: [self notifyUsersOfRecategorizedSelector: element from: oldCategory to: newCategory].! !

!TraitDescription methodsFor: 'users notification' stamp: 'NS 4/12/2004 15:45'!
notifyUsersOfChangedSelectors: aCollection
	self users do: [:each |
		each noteChangedSelectors: aCollection]! !

!TraitDescription methodsFor: 'users notification' stamp: 'NS 4/12/2004 21:46'!
notifyUsersOfRecategorizedSelector: element from: oldCategory to: newCategory
	self users do: [:each |
		each noteRecategorizedSelector: element from: oldCategory to: newCategory]! !

!TraitDescription methodsFor: 'private' stamp: 'md 8/11/2006 14:38'!
numberOfMethods
	"cound all methods that are local (not comming from a trait)"
	| num |
	num := (methodDict values select: [:each | self includesLocalSelector: each selector]) size.
	self isMeta 
		ifTrue: [^ num] 
		ifFalse: [^ num + self class numberOfMethods]! !

!TraitDescription methodsFor: 'initialization' stamp: 'NS 4/8/2004 11:01'!
obsolete
	"Make the receiver obsolete."
	self organization: nil.
	super obsolete! !

!TraitDescription methodsFor: 'organization' stamp: 'NS 4/8/2004 11:03'!
organization
	"Answer the instance of ClassOrganizer that represents the organization 
	of the messages of the receiver."

	organization ifNil:
		[self organization: (ClassOrganizer defaultList: self methodDict keys asSortedCollection asArray)].
	(organization isMemberOf: Array) ifTrue:
		[self recoverFromMDFaultWithTrace].
	
	"Making sure that subject is set correctly. It should not be necessary."
	organization ifNotNil: [organization setSubject: self].
	^ organization! !

!TraitDescription methodsFor: 'organization' stamp: 'NS 4/8/2004 11:04'!
organization: aClassOrg
	"Install an instance of ClassOrganizer that represents the organization of the messages of the receiver."

	aClassOrg ifNotNil: [aClassOrg setSubject: self].
	organization := aClassOrg! !

!TraitDescription methodsFor: 'fileIn/Out' stamp: 'marcus.denker 9/14/2008 19:04'!
printMethodChunk: selector withPreamble: doPreamble on: outStream
		moveSource: moveSource toFile: fileIndex
	"Copy the source code for the method associated with selector onto the fileStream.  If moveSource true, then also set the source code pointer of the method."
	| preamble method oldPos newPos sourceFile endPos |
	doPreamble 
		ifTrue: [preamble := self name , ' methodsFor: ' ,
					(self organization categoryOfElement: selector) asString printString]
		ifFalse: [preamble := ''].
	method := self methodDict at: selector ifAbsent:
		[outStream nextPutAll: selector; cr.
		outStream tab; nextPutAll: '** ERROR!!  THIS SCRIPT IS MISSING ** ' translated; cr; cr.
		outStream nextPutAll: '  '.
		^ outStream].

	((method fileIndex = 0
		or: [(SourceFiles at: method fileIndex) isNil])
		or: [(oldPos := method filePosition) = 0])
		ifTrue:
		["The source code is not accessible.  We must decompile..."
		preamble size > 0 ifTrue: [outStream cr; nextPut: $!!; nextChunkPut: preamble; cr].
		outStream nextChunkPut: method decompileString]
		ifFalse:
		[sourceFile := SourceFiles at: method fileIndex.
		preamble size > 0
			ifTrue:    "Copy the preamble"
				[outStream copyPreamble: preamble from: sourceFile at: oldPos]
			ifFalse:
				[sourceFile position: oldPos].
		"Copy the method chunk"
		newPos := outStream position.
		outStream copyMethodChunkFrom: sourceFile.
		sourceFile skipSeparators.      "The following chunk may have ]style["
		sourceFile peek == $] ifTrue: [
			outStream cr; copyMethodChunkFrom: sourceFile].
		moveSource ifTrue:    "Set the new method source pointer"
			[endPos := outStream position.
			method checkOKToAdd: endPos - newPos at: newPos.
			method setSourcePosition: newPos inFile: fileIndex]].
	preamble size > 0 ifTrue: [outStream nextChunkPut: ' '].
	^ outStream cr! !

!TraitDescription methodsFor: 'private' stamp: 'al 7/30/2004 11:49'!
spaceUsed
	^super spaceUsed + (self hasClassTrait
		ifTrue: [self classTrait spaceUsed] 
		ifFalse: [0])! !

!TraitDescription methodsFor: 'accessing class hierarchy' stamp: 'NS 4/12/2004 14:38'!
subclasses
	^ Array new! !

!TraitDescription methodsFor: 'class compatibility' stamp: 'NS 12/9/2003 11:18'!
theMetaClass
	^ self classTrait! !

!TraitDescription methodsFor: 'class compatibility' stamp: 'al 12/10/2003 00:07'!
theNonMetaClass
	^ self baseTrait! !

!TraitDescription methodsFor: 'accessing' stamp: 'NS 12/9/2003 11:00'!
traitVersion
	"Default.  Any class may return a later version to inform readers that use ReferenceStream.  8/17/96 tk"
	"This method allows you to distinguish between class versions when the shape of the class 
	hasn't changed (when there's no change in the instVar names).
	In the conversion methods you usually can tell by the inst var names 
	what old version you have. In a few cases, though, the same inst var 
	names were kept but their interpretation changed (like in the layoutFrame).
	By changing the class version when you keep the same instVars you can 
	warn older and newer images that they have to convert."
	^ 0! !

!TraitDescription methodsFor: 'accessing' stamp: 'NS 12/9/2003 15:12'!
version
		"Allows polymoprhism with ClassDescription>>version"
		
	^ self traitVersion! !

Object subclass: #TraitComposition
	instanceVariableNames: 'transformations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitComposition commentStamp: '<historical>' prior: 0!
I hold a collection of trait transformations and provide important facilities to query the trait composition. For each trait in the composition clause there exists exactly one transformation in the collection.

Note, that directly manipulating the composition of a class or trait does not effect changes automatically. Use PureBehavior>>setTraitComposition: to do this. You have to make a copy of the old trait composition before changing it because only the difference between the new and the old composition is updated!!!

!TraitComposition class methodsFor: 'instance creation' stamp: 'al 12/12/2003 18:24'!
with: aTraitTransformation
	^self new
		add: aTraitTransformation;
		yourself! !

!TraitComposition class methodsFor: 'instance creation' stamp: 'al 12/12/2003 18:24'!
with: aTraitTransformation with: anotherTraitTransformation
	^self new
		add: aTraitTransformation;
		add: anotherTraitTransformation;
		yourself! !

!TraitComposition methodsFor: 'composition' stamp: 'apb 8/22/2005 17:52'!
+ aTraitExpression
	^ aTraitExpression addCompositionOnLeft: self.
! !

!TraitComposition methodsFor: 'composition' stamp: 'al 12/15/2003 20:56'!
- anArray
	"the modifier operators #@ and #- bind stronger than +.
	Thus, #@ or #- sent to a sum will only affect the most right summand"
	
	self transformations
		addLast: (self transformations removeLast - anArray)! !

!TraitComposition methodsFor: 'composition' stamp: 'al 12/15/2003 20:56'!
@ anArrayOfAssociations
	"the modifier operators #@ and #- bind stronger than +.
	Thus, #@ or #- sent to a sum will only affect the most right summand"

	self transformations
		addLast: (self transformations removeLast @ anArrayOfAssociations)! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 3/18/2004 14:10'!
add: aTraitTransformation
	self errorIfNotAddable: aTraitTransformation.
	self transformations addLast: aTraitTransformation! !

!TraitComposition methodsFor: 'private' stamp: 'apb 8/22/2005 17:44'!
addCompositionOnLeft: aTraitComposition
	self transformations do: [ : each | aTraitComposition add: each ].
	^ aTraitComposition! !

!TraitComposition methodsFor: 'composition' stamp: 'apb 8/22/2005 17:28'!
addOnTheLeft: aTrait 
	self errorIfNotAddable: aTrait.
	self transformations addFirst: aTrait! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 3/26/2006 22:03'!
allTraits
	^self traits gather: [:trait |
		trait hasTraitComposition
			ifTrue: [trait traitComposition allTraits copyWith: trait]
			ifFalse: [Array with: trait]]! !

!TraitComposition methodsFor: 'converting' stamp: 'al 1/18/2004 19:35'!
asTraitComposition
	^self! !

!TraitComposition methodsFor: 'error-handling' stamp: 'al 3/18/2004 14:19'!
assertValidUser: aBehavior
	"Assert that this trait composition set for aBehavior
	does not introduce a cycle."
	
	(self allTraits includes: aBehavior) ifTrue: [
		TraitCompositionException signal: 'Cycle in compositions:  The composition (in)directly includes this trait!!']! !

!TraitComposition methodsFor: 'enquiries' stamp: 'apb 8/22/2005 14:48'!
changedSelectorsComparedTo: oldComposition
	| changedSelectors oldTransformation traits newTransformation |
	changedSelectors := IdentitySet new.
	traits := self traits asIdentitySet addAll: oldComposition traits asIdentitySet; yourself.
	traits do: [:each |
		newTransformation := self transformationOfTrait: each.
		oldTransformation := oldComposition transformationOfTrait: each.
		(newTransformation isNil or: [oldTransformation isNil])
			ifTrue: [
				changedSelectors addAll: each selectors]
			ifFalse: [
				changedSelectors addAll: 
					(newTransformation changedSelectorsComparedTo: oldTransformation)]].
	^changedSelectors! !

!TraitComposition methodsFor: 'copying' stamp: 'apb 8/24/2005 17:13'!
copy
	self error: 'should not be called'.
	^super copy
		transformations: (self transformations collect: [:each | each copy]);
		yourself! !

!TraitComposition methodsFor: 'copying' stamp: 'apb 8/24/2005 13:49'!
copyTraitExpression
	| newCopy |
	newCopy := self shallowCopy.
	newCopy transformations: (self transformations collect: [ : each | each copyTraitExpression ]).
	^ newCopy
! !

!TraitComposition methodsFor: 'copying' stamp: 'al 12/7/2005 20:49'!
copyWithExclusionOf: aSymbol to: aTrait
	| composition transformation |
	composition := self copyTraitExpression.
	transformation := (composition transformationOfTrait: aTrait).
	^composition
		remove: transformation;
		add: (transformation addExclusionOf: aSymbol);
		yourself! !

!TraitComposition methodsFor: 'copying' stamp: 'al 12/7/2005 20:50'!
copyWithoutAlias: aSymbol of: aTrait
	| composition transformation |
	composition := self copyTraitExpression.
	transformation := (composition transformationOfTrait: aTrait).
	^composition
		remove: transformation;
		add: (transformation removeAlias: aSymbol);
		normalizeTransformations;
		yourself! !

!TraitComposition methodsFor: 'error-handling' stamp: 'al 3/17/2004 15:22'!
errorIfNotAddable: aTraitTransformation
	(self includesTrait: aTraitTransformation trait) ifTrue: [
		^TraitCompositionException
			signal: 'Trait ' , aTraitTransformation trait asString, ' already in composition']! !

!TraitComposition methodsFor: 'testing' stamp: 'al 7/30/2004 15:50'!
includesMethod: aSelector
	^(self methodDescriptionForSelector: aSelector) isEmpty not! !

!TraitComposition methodsFor: 'testing' stamp: 'al 1/18/2004 19:32'!
includesTrait: aTrait
	^self traits includes: aTrait! !

!TraitComposition methodsFor: 'initialization' stamp: 'al 3/12/2004 17:06'!
initialize
	super initialize.
	transformations := OrderedCollection new! !

!TraitComposition methodsFor: 'testing' stamp: 'al 4/10/2004 23:25'!
isAliasSelector: aSymbol
	"Return true if the selector aSymbol is an alias defined
	in this or in another composition somewhere deeper in 
	the tree of traits compositions."

	| methodDescription |
	methodDescription := (self methodDescriptionsForSelector: aSymbol)
		detect: [:each | each selector = aSymbol].
	^methodDescription isAliasSelector! !

!TraitComposition methodsFor: 'testing' stamp: 'al 3/6/2004 19:05'!
isEmpty
	^self transformations isEmpty! !

!TraitComposition methodsFor: 'testing' stamp: 'al 4/10/2004 23:34'!
isLocalAliasSelector: aSymbol
	"Return true if the selector aSymbol is an alias defined
	in this composition."

	| methodDescription |
	methodDescription := (self methodDescriptionsForSelector: aSymbol)
		detect: [:each | each selector = aSymbol].
	^methodDescription isLocalAliasSelector! !

!TraitComposition methodsFor: 'enquiries' stamp: 'al 7/21/2004 21:03'!
methodDescriptionForSelector: aSymbol
	"Return a TraitMethodDescription for the selector aSymbol."

	| description |
	description := TraitMethodDescription selector: aSymbol.
	self transformations do: [:each |
		each collectMethodsFor: aSymbol into: description].
	^description! !

!TraitComposition methodsFor: 'enquiries' stamp: 'al 7/21/2004 21:03'!
methodDescriptionsForSelector: aSymbol
	"Return a collection of TraitMethodDescriptions for aSymbol and all the 
	aliases of aSymbol."

	| selectors collection |
	selectors := IdentitySet with: aSymbol.
	self transformations do: [:each |
		selectors addAll: (each aliasesForSelector: aSymbol)].
	collection := OrderedCollection new: selectors size.
	selectors do: [:each |
		collection add: (self methodDescriptionForSelector: each)].
	^collection! !

!TraitComposition methodsFor: 'composition' stamp: 'al 4/10/2004 00:08'!
normalizeTransformations
	self transformations: (
		self transformations collect: [:each |
			each normalized])! !

!TraitComposition methodsFor: 'testing' stamp: 'al 3/6/2004 19:05'!
notEmpty
	^self isEmpty not! !

!TraitComposition methodsFor: 'printing' stamp: 'al 12/16/2003 22:06'!
printOn: aStream
	self transformations isEmptyOrNil
		ifFalse: [
			self transformations
				do: [:each | aStream print: each]
				separatedBy: [aStream nextPutAll: ' + '] ]
		ifTrue: [aStream nextPutAll: '{}']
		! !

!TraitComposition methodsFor: 'printing' stamp: 'al 12/16/2003 22:16'!
printString
	^String streamContents: [:stream |
		self printOn: stream]! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 4/9/2004 17:12'!
remove: aTransformation
	self transformations
		remove: aTransformation! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 4/9/2004 17:12'!
removeFromComposition: aTrait
	self remove:
		(self transformationOfTrait: aTrait)! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 1/18/2004 19:29'!
size
	^transformations size! !

!TraitComposition methodsFor: 'enquiries' stamp: 'al 7/22/2004 15:35'!
traitProvidingSelector: aSymbol
	"Return the trait which originally provides the method aSymbol or return nil
	if trait composition does not provide this selector or there is a conflict.
	Take aliases into account. Return the trait which the aliased method is defined in."

	| methodDescription locatedMethod |
	methodDescription := self methodDescriptionForSelector: aSymbol.
	(methodDescription isProvided not or: [methodDescription isConflict])	
		ifTrue: [^nil].
	locatedMethod := methodDescription providedLocatedMethod.
	^locatedMethod location traitOrClassOfSelector: locatedMethod selector! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 4/13/2004 15:30'!
traits
	^self transformations collect: [:each |
		each trait]! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 3/7/2004 22:31'!
transformationOfTrait: aTrait
	"Return the transformation which holds aTrait
	or nil if this composition doesn't include aTrait."
	
	^self transformations
		detect: [:each | each trait = aTrait]
		ifNone: [nil]! !

!TraitComposition methodsFor: 'accessing' stamp: 'al 3/17/2004 19:34'!
transformations
	^transformations! !

!TraitComposition methodsFor: 'private' stamp: 'al 3/19/2004 10:30'!
transformations: aCollection
	transformations := aCollection! !

Object subclass: #TraitMethodDescription
	instanceVariableNames: 'selector locatedMethods'
	classVariableNames: 'ConflictMethods RequiredMethods'
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitMethodDescription commentStamp: '<historical>' prior: 0!
Used by TraitComposition to encapsulates a collection of methods for one particular selector when querying for changes. According to the number and kind of those methods a provided method exists, there is a conflict or there are no provided nor conflicting methods at all. I provide the interface to query for those situations, e.g., effectiveMethod returns the provided method or the conflict marker method.
!

!TraitMethodDescription class methodsFor: 'initialization' stamp: 'NS 4/19/2004 16:26'!
initialize
	"	self initialize	"
	ConflictMethods := Array new: self maxArguments + 2.
	RequiredMethods := Array new: self maxArguments + 2.! !

!TraitMethodDescription class methodsFor: 'private' stamp: 'NS 4/19/2004 16:26'!
maxArguments
	^30! !

!TraitMethodDescription class methodsFor: 'instance creation' stamp: 'al 1/23/2004 12:52'!
new
	^super new
		initialize;
		yourself! !

!TraitMethodDescription class methodsFor: 'instance creation' stamp: 'al 1/23/2004 13:04'!
selector: aSymbol
	^self new
		selector: aSymbol
		yourself! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 10:20'!
addLocatedMethod: aLocatedMethod
	locatedMethods add: aLocatedMethod! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 7/25/2004 21:21'!
conflictMethod
	| templateMethod argumentNames binary numberOfArguments |
	self isConflict ifFalse: [^nil].
	argumentNames := self getArgumentNames.
	binary := self isBinarySelector.
	numberOfArguments := binary
		ifTrue: [1]
		ifFalse: [argumentNames size + 2].
	templateMethod := self conflictMethodForArguments: numberOfArguments ifAbsentPut: [
		self
			generateTemplateMethodWithMarker: CompiledMethod conflictMarker
			forArgs: argumentNames size
			binary: binary].
	^templateMethod copyWithTempNames: argumentNames
	
	
 ! !

!TraitMethodDescription methodsFor: 'private' stamp: 'al 7/25/2004 21:20'!
conflictMethodForArguments: aNumber ifAbsentPut: aBlock
	"ConflictMethods is an array that caches the generated conflict
	methods. At position 1: binary method; 2: unary method;
	n+2: keywordmethod with n arguments." 

	^(ConflictMethods at: aNumber)
		ifNil: [ConflictMethods at: aNumber put: aBlock value]! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'NS 4/19/2004 16:13'!
effectiveMethod
	"Return the effective compiled method of this method description." 

	| locatedMethod method |
	method := self providedMethod.
	method isNil ifFalse: [^ method].
	method := self conflictMethod.
	method isNil ifFalse: [^ method].
	^ self requiredMethod.! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'NS 4/16/2004 15:06'!
effectiveMethodCategory
	^ self effectiveMethodCategoryCurrent: nil new: nil! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 7/22/2004 14:50'!
effectiveMethodCategoryCurrent: currentCategoryOrNil new: newCategoryOrNil
	| isCurrent result cat size isConflict |
	size := self size.
	size = 0 ifTrue: [^ nil].
	result := self locatedMethods anyOne category.
	size = 1 ifTrue: [^ result].
	
	isCurrent := currentCategoryOrNil isNil.
	isConflict := false.
	self locatedMethods do: [:each |
		cat := each category.
		isCurrent := isCurrent or: [cat == currentCategoryOrNil].
		isConflict := isConflict or: [cat ~~ result]].
	isConflict ifFalse: [^ result].
	(isCurrent not and: [newCategoryOrNil notNil]) ifTrue: [^ newCategoryOrNil].
	^ ClassOrganizer ambiguous.! !

!TraitMethodDescription methodsFor: 'private' stamp: 'md 2/13/2006 14:36'!
generateTemplateMethodWithMarker: aSymbol forArgs: aNumber binary: aBoolean
	| source node |
	source := String streamContents: [:stream |
		aNumber < 1
			ifTrue: [stream nextPutAll: 'selector']
			ifFalse: [aBoolean
				ifTrue: [
					stream nextPutAll: '* anObject']
				ifFalse: [
					1 to: aNumber do: [:argumentNumber |
						stream
							nextPutAll: 'with:'; space;
							nextPutAll: 'arg'; nextPutAll: argumentNumber asString; space]]].
		stream cr; tab; nextPutAll: 'self '; nextPutAll: aSymbol].
	node := self class compilerClass new
		compile: source
		in: self class
		notifying: nil
		ifFail: [].
	^node generate.! !

!TraitMethodDescription methodsFor: 'private' stamp: 'al 3/12/2004 17:04'!
getArgumentNames
	| argumentNamesCollection names defaultName |
	defaultName := 'arg'.
	argumentNamesCollection := self locatedMethods
		collect: [:each | each argumentNames ].
	names := Array new: argumentNamesCollection anyOne size.
	argumentNamesCollection do: [:collection |
		1 to: names size do: [:index |
			(names at: index) isNil
				ifTrue: [names at: index put: (collection at: index)]
				ifFalse: [(names at: index) ~= (collection at: index)
					ifTrue: [names at: index put: defaultName, index asString]]]].
	^names
		! !

!TraitMethodDescription methodsFor: 'initialization' stamp: 'NS 3/30/2004 16:12'!
initialize
	super initialize.
	locatedMethods := Set new! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'al 4/10/2004 23:29'!
isAliasSelector
	"Return true if the selector is an alias (if it is different
	from the original selector) or already an aliased method
	in the original location (recursively search the compositions).
	Return false, if not or if we have a conflict."
	
	| locatedMethod |
	^self size = 1 and: [
		locatedMethod := self locatedMethods anyOne.
		(locatedMethod selector ~= self selector) or: [
			locatedMethod location isAliasSelector: self selector]]! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'al 2/20/2004 22:18'!
isBinarySelector
	^self locatedMethods anyOne
		isBinarySelector! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'al 1/23/2004 10:30'!
isConflict
	| count |
	count := 0.
	self methodsDo: [:each |
		each isProvided ifTrue: [
			count := count + 1.
			count > 1 ifTrue: [^true]]].
	^false! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'al 1/23/2004 10:31'!
isEmpty
	^self size = 0! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'al 4/10/2004 23:34'!
isLocalAliasSelector
	"Return true if the selector is an alias (if it is different
	from the original selector). Return false, if not or if we
	have a conflict."
	
	^self size = 1 and: [
		(self locatedMethods anyOne selector ~= self selector)]! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'NS 4/19/2004 16:10'!
isProvided
	^ self providedMethod notNil! !

!TraitMethodDescription methodsFor: 'testing' stamp: 'NS 4/21/2004 12:29'!
isRequired
	self isEmpty ifTrue: [^ false].
	^ self locatedMethods allSatisfy: [:each | each method isRequired]! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 10:20'!
locatedMethods
	^locatedMethods! !

!TraitMethodDescription methodsFor: 'enumeration' stamp: 'al 7/22/2004 14:50'!
methodsDo: aBlock
	self locatedMethods do: [:each |
		aBlock value: each method]! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 7/21/2004 21:17'!
providedLocatedMethod
	| locatedMethod |
	locatedMethod := nil.
	self locatedMethods do: [:each |
		each method isProvided ifTrue: [
			locatedMethod isNil ifFalse: [^nil].
			locatedMethod := each]].
	^locatedMethod! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 7/21/2004 21:26'!
providedMethod
	^self providedLocatedMethod ifNotNilDo: [:locatedMethod | locatedMethod method]! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 7/25/2004 21:39'!
requiredMethod
	| templateMethod argumentNames numberOfArguments binary |
	self isRequired ifFalse: [^nil].
	self size = 1 ifTrue: [^self locatedMethods anyOne method].
	
	argumentNames := self getArgumentNames.
	binary := self isBinarySelector.
	numberOfArguments := binary
		ifTrue: [1]
		ifFalse: [argumentNames size + 2].
	templateMethod := self requiredMethodForArguments: numberOfArguments ifAbsentPut: [
		self
			generateTemplateMethodWithMarker: CompiledMethod implicitRequirementMarker
			forArgs: argumentNames size
			binary: binary].
	^templateMethod copyWithTempNames: argumentNames
	
	
 ! !

!TraitMethodDescription methodsFor: 'private' stamp: 'al 7/25/2004 21:37'!
requiredMethodForArguments: aNumber ifAbsentPut: aBlock
	"ConflictMethods is an array that caches the generated conflict
	methods. At position 1: binary method; 2: unary method;
	n+2: keywordmethod with n arguments." 

	^(RequiredMethods at: aNumber)
		ifNil: [ConflictMethods at: aNumber put: aBlock value]! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 10:19'!
selector
	^selector! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 10:19'!
selector: aSymbol 
	selector := aSymbol! !

!TraitMethodDescription methodsFor: 'accessing' stamp: 'al 1/23/2004 12:54'!
size
	^self locatedMethods size! !

Object subclass: #TraitTransformation
	uses: TComposingDescription
	instanceVariableNames: 'subject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitTransformation commentStamp: 'apb 8/25/2005 11:54' prior: 0!
A trait transformation is an instance of one of my concrete subclasses, TraitAlias or TraitExclusion. These represent a transformation of a trait, specified by the alias and exclusion operators. 

I define an instance variable named subject which holds the object that is transformed.  Thus, an alias transformation has as its subject a trait, and a trait exclusion has as its subject either a trait alias or a trait. Each of the concrete transformation classes implement the method allSelectors according to the transformation it represents. 

(There was formerly a subclass called TraitHolder, which was the identity transformation and which did not modify the trait.  This was clearly redundant, and was removed.)!

TraitTransformation class
	uses: TComposingDescription classTrait
	instanceVariableNames: ''!

TraitTransformation subclass: #TraitAlias
	instanceVariableNames: 'aliases'
	classVariableNames: 'AliasMethodCache'
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitAlias commentStamp: '<historical>' prior: 0!
See comment of my superclass TraitTransformation.!

!TraitAlias class methodsFor: 'instance creation' stamp: 'al 3/17/2004 15:22'!
assertValidAliasDefinition: anArrayOfAssociations
	"Throw an exceptions if the alias definition is not valid.
	It is expected to be a collection of associations and
	the number of arguments of the alias selector has to
	be the same as the original selector."

	((anArrayOfAssociations isKindOf: Collection) and: [
		anArrayOfAssociations allSatisfy: [:each |
			each isKindOf: Association]]) ifFalse: [
		TraitCompositionException signal: 'Invalid alias definition: Not a collection of associations.'].
	
	(anArrayOfAssociations allSatisfy: [:association |
		(association key numArgs = association value numArgs and: [
			(association key numArgs = -1) not])]) ifFalse: [
		TraitCompositionException signal: 'Invalid alias definition: Alias and original selector have to have the same number of arguments.']! !

!TraitAlias class methodsFor: 'instance creation' stamp: 'al 3/12/2004 16:24'!
with: aTraitComposition aliases: anArrayOfAssociations
	self assertValidAliasDefinition: anArrayOfAssociations.
	^self new
		subject: aTraitComposition;
		aliases: anArrayOfAssociations;
		yourself! !

!TraitAlias methodsFor: 'composition' stamp: 'al 3/3/2004 19:49'!
- anArrayOfSelectors
	^TraitExclusion
		with: self
		exclusions: anArrayOfSelectors! !

!TraitAlias methodsFor: 'enumeration' stamp: 'al 4/7/2004 21:14'!
aliasNamed: aSymbol ifAbsent: aBlock
	^self aliases
		detect: [:association |  association key = aSymbol]
		ifNone: aBlock! !

!TraitAlias methodsFor: 'accessing' stamp: 'al 3/12/2004 16:43'!
aliases
	"Collection of associations where key is the
	alias and value the original selector."
	
	^aliases! !

!TraitAlias methodsFor: 'accessing' stamp: 'al 3/17/2004 15:22'!
aliases: anArrayOfAssociations
	| newNames |
	newNames := (anArrayOfAssociations collect: [:each | each key]) asIdentitySet.
	newNames size < anArrayOfAssociations size ifTrue: [
		TraitCompositionException signal: 'Cannot use the same alias name twice'].
	anArrayOfAssociations do: [:each |
		(newNames includes: each value) ifTrue: [
			TraitCompositionException signal: 'Cannot define an alias for an alias']].
	aliases := anArrayOfAssociations! !

!TraitAlias methodsFor: 'enquiries' stamp: 'al 3/12/2004 17:10'!
aliasesForSelector: aSymbol
	| selectors |
	selectors := self aliases
		select: [:association | association value = aSymbol]
		thenCollect: [:association | association key].
	^(super aliasesForSelector: aSymbol)
		addAll: selectors;
		yourself
		 ! !

!TraitAlias methodsFor: 'enquiries' stamp: 'al 7/13/2004 16:53'!
allAliasesDict
	| dict |
	dict := super allAliasesDict.
	self aliases do: [:assoc |
		dict at: assoc key put: assoc value].
	^dict! !

!TraitAlias methodsFor: 'enquiries' stamp: 'al 3/7/2004 23:26'!
allSelectors
	^self subject allSelectors
		addAll: (self aliases collect: [:each | each key]) asSet;
		yourself! !

!TraitAlias methodsFor: 'enquiries' stamp: 'al 4/7/2004 21:16'!
collectMethodsFor: aSelector into: methodDescription
	| originalSelector association |
	self subject
		collectMethodsFor: aSelector
		into: methodDescription.			
	association := self aliasNamed: aSelector ifAbsent: [nil].
	association notNil ifTrue: [
		originalSelector := association value.
		self subject
			collectMethodsFor: originalSelector
			into: methodDescription]! !

!TraitAlias methodsFor: 'copying' stamp: 'al 4/9/2004 21:27'!
copy
	^super copy
		aliases: self aliases copy;
		yourself! !

!TraitAlias methodsFor: 'copying' stamp: 'dvf 8/25/2005 16:07'!
copyTraitExpression
	^super copyTraitExpression
		aliases: self aliases deepCopy;
		yourself! !

!TraitAlias methodsFor: 'testing' stamp: 'al 4/10/2004 23:57'!
isEmpty
	^self aliases isEmpty! !

!TraitAlias methodsFor: 'printing' stamp: 'al 2/22/2004 21:45'!
printOn: aStream
	super printOn: aStream.
	aStream
		space;
		nextPut: $@;
		space;
		nextPut: ${.
	self aliases do: [:each | aStream print: each]
		separatedBy: [aStream nextPutAll: '. '].
	aStream nextPut: $}! !

!TraitAlias methodsFor: 'composition' stamp: 'al 4/10/2004 23:51'!
removeAlias: aSymbol
	self aliases: (self aliases
		reject: [:each | each key = aSymbol])! !

TraitTransformation subclass: #TraitExclusion
	instanceVariableNames: 'exclusions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Traits-Composition'!

!TraitExclusion commentStamp: '<historical>' prior: 0!
See comment of my superclass TraitTransformation.!

!TraitExclusion class methodsFor: 'instance creation' stamp: 'dvf 8/18/2005 17:36'!
with: aTraitComposition exclusions: anArrayOfSelectors
	^self new
		subject: aTraitComposition;
		exclusions: anArrayOfSelectors;
		yourself
! !

!TraitExclusion methodsFor: 'composition' stamp: 'al 4/9/2004 16:55'!
addExclusionOf: aSymbol
	self exclusions: (self exclusions copyWith: aSymbol)! !

!TraitExclusion methodsFor: 'enquiries' stamp: 'al 3/7/2004 22:48'!
allSelectors
	| selectors |
	selectors := self subject allSelectors.
	self exclusions do: [:each |
		selectors remove: each ifAbsent: []].
	^selectors! !

!TraitExclusion methodsFor: 'enquiries' stamp: 'al 3/4/2004 16:35'!
collectMethodsFor: aSelector into: methodDescription
	(self exclusions includes: aSelector) ifFalse: [
		self subject
			collectMethodsFor: aSelector
			into: methodDescription]! !

!TraitExclusion methodsFor: 'copying' stamp: 'al 4/9/2004 19:12'!
copy
	^super copy
		exclusions: self exclusions copy;
		yourself! !

!TraitExclusion methodsFor: 'copying' stamp: 'dvf 8/25/2005 16:07'!
copyTraitExpression
	^super copyTraitExpression 
		exclusions: self exclusions deepCopy;
		yourself! !

!TraitExclusion methodsFor: 'accessing' stamp: 'al 12/7/2003 11:58'!
exclusions
	^exclusions! !

!TraitExclusion methodsFor: 'accessing' stamp: 'al 12/7/2003 11:58'!
exclusions: aCollection
	exclusions := aCollection! !

!TraitExclusion methodsFor: 'testing' stamp: 'al 4/10/2004 23:56'!
isEmpty
	^self exclusions isEmpty! !

!TraitExclusion methodsFor: 'accessing' stamp: 'dvf 8/19/2005 12:25'!
methodReferencesInCategory: aCategoryName
	^(self organization listAtCategoryNamed: aCategoryName)
		collect: [:ea | MethodReference new
						setClassSymbol: self theNonMetaClass name
						classIsMeta: self isMeta
						methodSymbol: ea
						stringVersion: '']
! !

!TraitExclusion methodsFor: 'printing' stamp: 'al 2/22/2004 21:45'!
printOn: aStream
	super printOn: aStream. 
	aStream
		space;
		nextPut: $-;
		space;
		nextPut: ${.
	self exclusions do: [:each | aStream print: each]
		separatedBy: [aStream nextPutAll: '. '].
	aStream nextPut: $}! !

!TraitTransformation methodsFor: 'composition' stamp: 'al 12/7/2005 21:19'!
- anArray
	TraitCompositionException signal: 'Invalid trait exclusion. Exclusions have to be specified after aliases.'! !

!TraitTransformation methodsFor: 'composition' stamp: 'al 12/7/2005 21:20'!
@ anArrayOfAssociations
	TraitCompositionException signal: 'Invalid trait exclusion. Aliases have to be specified before exclusions.'! !

!TraitTransformation methodsFor: 'composition' stamp: 'al 4/9/2004 17:01'!
addExclusionOf: aSymbol
	^self - {aSymbol}! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'al 3/11/2004 21:49'!
aliasesForSelector: aSymbol
	"Return a collection of alias selectors that are defined in this transformation."
	
	^self subject aliasesForSelector: aSymbol! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'NS 4/16/2004 13:40'!
allAliasesDict
	"Return a dictionary with all alias associations that are defined in this transformation."
	
	^self subject allAliasesDict! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'al 3/7/2004 22:36'!
allSelectors
	^self subclassResponsibility! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'NS 4/16/2004 14:20'!
changedSelectorsComparedTo: aTraitTransformation
	| selectors otherSelectors changedSelectors aliases otherAliases |
	selectors := self allSelectors asIdentitySet.
	otherSelectors := aTraitTransformation allSelectors asIdentitySet.
	changedSelectors := IdentitySet withAll: (
		(selectors difference: otherSelectors) union: (otherSelectors difference: selectors)).
	aliases := self allAliasesDict.
	otherAliases := aTraitTransformation allAliasesDict.
	aliases keysAndValuesDo: [:key :value |
		(value ~~ (otherAliases at: key ifAbsent: [nil])) ifTrue: [changedSelectors add: key]].
	otherAliases keysAndValuesDo: [:key :value |
		(value ~~ (aliases at: key ifAbsent: [nil])) ifTrue: [changedSelectors add: key]].
	^ changedSelectors.! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'al 3/6/2004 18:51'!
collectMethodsFor: aSelector into: methodDescription
	"Collect instances of LocatedMethod into methodDescription
	for each method that has the selector aSelector and is not excluded
	or for which aSelector is an alias."

	self subclassResponsibility! !

!TraitTransformation methodsFor: 'copying' stamp: 'apb 8/24/2005 17:13'!
copy
	self error: 'should not be called'.
	^super copy
		subject: self subject copy;
		yourself! !

!TraitTransformation methodsFor: 'copying' stamp: 'dvf 8/25/2005 16:06'!
copyTraitExpression
	^self shallowCopy 
		subject: self subject copyTraitExpression;
		yourself! !

!TraitTransformation methodsFor: 'testing' stamp: 'al 4/10/2004 23:56'!
isEmpty
	self subclassResponsibility! !

!TraitTransformation methodsFor: 'accessing parallel hierarchy' stamp: 'dvf 8/18/2005 18:29'!
isMeta
	^self subject isMeta! !

!TraitTransformation methodsFor: 'accessing' stamp: 'al 4/10/2004 00:06'!
normalized
	^self isEmpty
		ifFalse: [
			self subject: self subject normalized.
			self]
		ifTrue: [self subject normalized]
		
		! !

!TraitTransformation methodsFor: 'printing' stamp: 'al 12/12/2003 18:29'!
printOn: aStream
	aStream print: self subject! !

!TraitTransformation methodsFor: 'composition' stamp: 'al 4/10/2004 23:48'!
removeAlias: aSymbol
	self subject removeAlias: aSymbol! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'dvf 8/18/2005 18:39'!
selectors
	^self allSelectors! !

!TraitTransformation methodsFor: 'browser support' stamp: 'dvf 8/18/2005 18:31'!
sourceCodeTemplate
	^ self subject sourceCodeTemplate! !

!TraitTransformation methodsFor: 'accessing' stamp: 'NS 12/11/2003 15:13'!
subject
	^subject! !

!TraitTransformation methodsFor: 'accessing' stamp: 'al 4/9/2004 18:23'!
subject: aTraitTransformation
	subject := aTraitTransformation! !

!TraitTransformation methodsFor: 'accessing parallel hierarchy' stamp: 'dvf 8/18/2005 18:28'!
theNonMetaClass 
	^ self subject theNonMetaClass  ! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'al 12/12/2003 18:00'!
trait
	^self subject trait! !

!TraitTransformation methodsFor: 'enquiries' stamp: 'dvf 8/16/2005 17:35'!
traitTransformations 
	^ { subject }! !
TraitMethodDescription initialize!
ModelExtension initialize!
